// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1997 Asger Alstrup
 *           and the LyX Team.
 *
 *======================================================*/

#include <config.h>
#include <cmath>	// fabs()
#include <stdlib.h>	// atoi()

#ifdef __GNUG__
#pragma implementation
#endif

#include "FontInfo.h"
#include "debug.h"
#include FORMS_H_LOCATION
#include "lyxrc.h"	// lyxrc.use_scalable_fonts
#include "lstrings.h"

extern LyXRC * lyxrc;


/// Load font close to this size
string FontInfo::getFontname(int size)
{
	if (!exist()) return string();

	lyxerr[Debug::INFO] << "Requesting font size: " << size << '\n';
	// First check if we have a perfect match
	FontList::const_iterator fc = fontList.find(size);
	if (fc != fontList.end()) { // Perfect match
		lyxerr[Debug::INFO] << "Perfect font match with:\n"
				    << (*fc).second << '\n';
		return (*fc).second;
	} else if (scalable() && lyxrc->use_scalable_fonts) {
		// We can use scalable
		string font = fontList[0];
		lyxerr[Debug::INFO] << "Using scalable font to get\n"
				    << font << '\n';
		return resize(font, size);
	} else { // Could not find perfect match nor a scalable
		// Try to find a close match instead.
		double error = 100000.0;
		int close_size = -1;
		for (FontList::const_iterator fci = fontList.begin();
		     fci != fontList.end(); ++fci) {
			// Scalable font should not be considered close
			if ((*fci).first != 0 &&
			    abs((*fci).first - size - 0.1) < error) {
				error = abs((*fci).first - size - 0.1);
				close_size = (*fci).first;
			}
		}
		// Did any fonts get close?
		if (close_size > 0) {
			return fontList[close_size];
		}
	}
	return string();
}


/// Build newly sized font string 
string FontInfo::resize(string const & font, int size)
{
	// Find the position of the size spec
	// find the 7th and 8th '-'
	string::size_type seven = 0;
	string::size_type eight = 0;
	string::size_type count = 0;
	for (string::size_type i = 0; i < font.length(); ++i) {
		if (font[i] == '-') {
			++count;
			if (count == 7) {
				seven = i + 1; // Note we step the '-'
			}  else if (count == 8) {
				eight = i;
			}
		}
	}

	lyxerr[Debug::INFO] << "Font to resize: " << font << '\n';
	string ret_font(font);
	ret_font.replace(seven, eight - seven, tostr(size));
	lyxerr[Debug::INFO] << "Resizeed font: " << ret_font << '\n';
	return ret_font;
}


/// Set new pattern
void FontInfo::setPattern(string const & pat)
{
	release();
	init();
	pattern = pat;
}


/// Query font in X11
void FontInfo::query()
{
	if (queried)
		return;

	if (pattern.empty()) {
		lyxerr << "Can not use empty font name for font query.\n";
		queried = true;
		return;
	}

	int matches = 0;
	char ** list = XListFonts(fl_display, pattern.c_str(), 100, &matches);

	if (matches) {
		// We have matches. Run them through
		lyxerr[Debug::INFO] << "We have " << matches 
				    << " font matches.\n";
		for(int i = 0; i < matches; ++i) {
			// Only insert the first entry of a given size
			int size = atoi(token(list[i], '-', 7).c_str());
			lyxerr[Debug::INFO] << "Checking size "
					    << size << '\n';
			if (fontList.find(size) == fontList.end()) {
				lyxerr[Debug::INFO] << "Inserting \""
						    << list[i] << "\"\n";
				fontList[size] = list[i];
			}
		}
		XFreeFontNames(list);
	}
	queried = true;

	//List all font found:
	if (lyxerr.debugging()) {
		lyxerr << "Font List:\n";
		FontList::const_iterator ci = fontList.begin();
		for (; ci != fontList.end(); ++ci) {
			lyxerr << (*ci).first
			       << " -> " + (*ci).second << '\n';
		}
	}
}


/// Release allocated stuff
void FontInfo::release()
{
	fontList.erase(fontList.begin(), fontList.end());
}
