/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor 	 
 *	     Copyright (C) 1995 Matthias Ettrich
 *           Copyright (C) 1995-1999 The LyX Team.
 *
 *           This file is Copyright (C) 1996-1999
 *           Lars Gullik Bjnnes
 *
 *======================================================
 */

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "filetools.h"
#include "LaTeX.h"
#include "lyxlex.h"
#include "FileInfo.h"
#include "debug.h"
#include "lyxlib.h"
#include "syscall.h"
#include "syscontr.h"
#include "path.h"
#include "bufferlist.h"
#include "minibuffer.h"
#include "gettext.h"
#include "DepTable.h"
#include "lstrings.h"

// 	$Id: LaTeX.C,v 1.8 1998/10/22 16:56:05 larsbj Exp $	

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: LaTeX.C,v 1.8 1998/10/22 16:56:05 larsbj Exp $";
#endif /* lint */

extern BufferList bufferlist;

struct texfile_struct {
	LaTeX::TEX_FILES file;
	char const *extension;
};

static
const texfile_struct all_files[] = {
	{ LaTeX::AUX, ".aux"},
	{ LaTeX::BBL, ".bbl"},
	{ LaTeX::DVI, ".dvi"},
	{ LaTeX::GLO, ".glo"},
	{ LaTeX::IDX, ".idx"},
	{ LaTeX::IND, ".ind"},
	{ LaTeX::LOF, ".lof"},
	{ LaTeX::LOA, ".loa"},
	{ LaTeX::LOG, ".log"},
	{ LaTeX::LOT, ".lot"},
	{ LaTeX::TOC, ".toc"},
	{ LaTeX::LTX, ".ltx"},
	{ LaTeX::TEX, ".tex"}
};


// This should perhaps be placed in LyXLex
static
string readLine(FILE *file)
{
	if (feof(file))
		return string();

	int i = 0;
	char s[512];

	do {
		s[i] = fgetc(file);
		++i;
	} while (!feof(file) && s[i-1] != '\n' && i<510);
	s[i] = '\0';
	string tmp;
	if (i == 1 && feof(file))
		;
	else
		tmp = s;

	return tmp;
}



/*
 * CLASS TEXERRORS
 */

void TeXErrors::scanError(LyXLex &lex)
{
	string token = lex.GetString();
	// Sometimes the error string goes over more than one
	// line, and we need to get them all.
	string tmp;
	string errstr;
	tmp = frontStrip(readLine(lex.getFile()));
	if (tmp == "\n" || tmp.empty()) {
		tmp = frontStrip(readLine(lex.getFile()));
		if (contains(tmp, "See the LaTeX manual")) {
			do {
				tmp = readLine(lex.getFile());
			} while (!tmp.empty() && !contains(tmp, "..."));
		}
		tmp =frontStrip( readLine(lex.getFile()));
	}

	while ((tmp != "\n" || !contains(errstr, "l."))
		&& !prefixIs(tmp, "! ")
		&& !contains(tmp, "(job aborted")
		&& !tmp.empty()) {
		errstr += tmp;
		tmp = frontStrip(readLine(lex.getFile()));
	}
	lyxerr[Debug::INFO] << "tmp: " << errstr << '\n';
	int line = 0;
	// unfortunately the error line is not always given
	// by "l.###" in the beginning of the error string
	// therefore we must search for "l.###" in the error
	// msg.
	if (contains(errstr, "l.")) {
		// We make a const copy to make [] fast. (Asger)
		string const es = errstr;
		for (unsigned int i = 2; i < es.length(); ++i) {
			if (es[i-2] == 'l' && es[i-1] == '.' &&
			    (es[i] >= '0' && es[i]<= '9')) {
				line = atoi(es.c_str() + i);
				break;
			}
		}
	}
	insertError(line, token, errstr);

	if (prefixIs(tmp, "! ")) {
		scanError(lex);
	}
}


bool TeXErrors::getFirstError(int *line, string *text)
{
	error_iter = errorlist.begin();
	if(error_iter != errorlist.end()) {
		*line = (*error_iter).error_in_line;
		*text = (*error_iter).error_desc + "\n" +
			(*error_iter).error_text;
		++error_iter;
		return true;
	}
	return false;
}


bool TeXErrors::getNextError(int *line, string *text)
{
	if(error_iter != errorlist.end()) {
		*line = (*error_iter).error_in_line;
		*text = (*error_iter).error_desc + "\n" +
			(*error_iter).error_text;
		++error_iter;
		return true;
	}
	return false;
}


void TeXErrors::insertError(int line, string const &error_desc,
			    string const &error_text)
{
        Error newerr(line, error_desc, error_text);
	errorlist.push_back(newerr);
}


void TeXErrors::printErrors()
{
        lyxerr << "Printing errors.\n";
	ErrorList::const_iterator elc = errorlist.begin();
	for (; elc != errorlist.end(); ++elc) {
		lyxerr << "Error in line "
		       << (*elc).error_in_line << ": "
		       << (*elc).error_desc << '\n'
		       << (*elc).error_text << '\n';
	}
}


/*
 * CLASS LaTeX
 */

LaTeX::LaTeX(string const & latex, string const & f, string const & p)
		: cmd(latex), file(f), path(p)
{
	tex_files = NO_FILES;
	file_count = sizeof(all_files) / sizeof(texfile_struct);
	num_errors = 0;
	depfile = file + ".dep";
}


int LaTeX::run(TeXErrors & terr, MiniBuffer * minib)
	// We know that this function will only be run if the lyx buffer
	// has been changed. We also know that a newly written .tex file
	// is always different from the previous one because of the date
	// in it. However it seems safe to run latex (at least) one time each
	// time the .tex file changes.
{
	int scanres = LaTeX::NO_ERRORS;
	unsigned int count = 0; // number of times run
	num_errors = 0; // just to make sure.
	const unsigned int MAX_RUN = 6;
	DepTable dep;
	bool rerun = false; // rerun requested
	
	// The class LaTeX does not know the temp path.
#warning This is almost certainly the wrong place for this.
	bufferlist.updateIncludedTeXfiles(GetCWD());
	
	// Never write the depfile if an error was encountered.
	
	// 0
	// first check if the file dependencies exist:
	//     ->If it does exist
	//             check if any of the files mentioned in it have
	//             changed (done using a checksum).
	//                 -> if changed:
	//                        run latex once and
	//                        remake the dependency file
	//                 -> if not changed:
	//                        just return there is nothing to do for us.
	//     ->if it doesn't exist
	//             make it and
	//             run latex once (we need to run latex once anyway) and
	//             remake the dependency file.
	//
	FileInfo fi(depfile);
	if (fi.exist()) {
		lyxerr[Debug::LATEX] << "Dependency file exists.\n";
		// Read the dep file:
		dep.read(depfile);
		// Update the checksums
		dep.update();
		
		if (dep.sumchange()) {
			lyxerr[Debug::LATEX] <<"Dependency file has changed\n";
			lyxerr[Debug::LATEX] << _("Run #") << ++count << '\n';
			minib->Set(string(_("LaTeX run number ")) + 
				   tostr(count));
			minib->Store();
			this->operator()(); // running latex
			scanres = scanLogFile(terr);
			if (scanres & LaTeX::ERRORS) return scanres; // return on error
		} else {
			lyxerr[Debug::LATEX] 
				<< "Dependency file has NOT changed\n";
			return LaTeX::NO_CHANGE;
		}
	} else {
		lyxerr[Debug::LATEX] << "Dependency file does not exist\n";
		lyxerr[Debug::LATEX] << _("Run #") << ++count << '\n';
		deptex(dep);
		dep.update();
		minib->Set(string(_("LaTeX run number ")) + tostr(count));
		minib->Store();
		this->operator()(); // running latex
		scanres = scanLogFile(terr);
		if (scanres & LaTeX::ERRORS) return scanres; // return on error
	}

	// update the dependencies.
	deplog(dep); // reads the latex log
	deptex(dep); // checks for latex files
	dep.update();

	// 0.5
	// At this point we must run external programs if needed.
	// makeindex will be run if a .idx file changed or was generated.
	// And if there were undefined citations or changes in references
	// the .aux file is checked for signs of bibtex. Bibtex is then run
	// if needed.
	
	// run makeindex
	if (dep.haschanged(ChangeExtension(file, ".idx", true))) {
		// no checks for now
		minib->Set(_("Running MakeIndex."));
		minib->Store();
		rerun = runMakeIndex(ChangeExtension(file, ".idx", true));
	}

	// run bibtex
	if (scanres & LaTeX::UNDEF_CIT || scanres & LaTeX::RERUN) {
		// Here we must scan the .aux file and look for
		// "\bibdata" and/or "\bibstyle". If one of those
		// tags is found -> run bibtex and set rerun = true;
		// no checks for now
		minib->Set(_("Running BibTeX."));
		minib->Store();
		rerun = runBibTeX(ChangeExtension(file, ".aux", true));
	}
	
	// 1
	// we know on this point that latex has been run once (or we just
	// returned) and the question now is to decide if we need to run
	// it any more. This is done by asking if any of the files in the
	// dependency file has changed. (remember that the checksum for
	// a given file is reported to have changed if it just was created)
	//     -> if changed or rerun == true:
	//             run latex once more and
	//             update the dependency structure
	//     -> if not changed:
	//             we do nothing at this point
	//
	if (rerun || dep.sumchange()) {
		rerun = false;
		lyxerr[Debug::LATEX] 
			<< "Dep. file has changed or rerun requested\n";
		lyxerr[Debug::LATEX] << "Run #" << ++count << '\n';
		minib->Set(string(_("LaTeX run number ")) + tostr(count));
		minib->Store();
		this->operator()();
		scanres = scanLogFile(terr);
		if (scanres & LaTeX::ERRORS) return scanres; // return on error
		// update the depedencies
		deplog(dep); // reads the latex log
		dep.update();
	} else {
		lyxerr[Debug::LATEX] << "Dep. file has NOT changed\n";
	}

	// 1.5
	// The inclusion of files generated by external programs like
	// makeindex or bibtex might have done changes to pagenumbereing,
	// etc. And because of this we must run the external programs
	// again to make sure everything is redone correctly.
	// Also there should be no need to run the external programs any
	// more after this.
	
	// run makeindex if the <file>.idx has changed or was generated.
	if (dep.haschanged(ChangeExtension(file, ".idx", true))) {
		// no checks for now
		minib->Set(_("Running MakeIndex."));
		minib->Store();
		rerun = runMakeIndex(ChangeExtension(file, ".idx", true));
	}
	
	// 2
	// we will only run latex more if the log file asks for it.
	//     -> rerun asked for:
	//             run latex and
	//             remake the dependency file
	//             goto 2 or return if max runs are reached.
	//     -> rerun not asked for:
	//             just return (fall out of bottom of func)
	//
	while ((rerun || (scanres & LaTeX::RERUN)) && count < MAX_RUN) {
		// Yes rerun until message goes away, or until
		// MAX_RUNS are reached.
		rerun = false;
		lyxerr[Debug::LATEX] << _("Run #") << ++count << '\n';
		minib->Set(string(_("LaTeX run number ")) + tostr(count));
		minib->Store();
		this->operator()();
		scanres = scanLogFile(terr);
		if (scanres & LaTeX::ERRORS) return scanres; // return on error
		// keep this updated
		dep.update();
	}

	// Write the dependencies to file.
	dep.write(depfile);
	lyxerr[Debug::LATEX] << "Done.\n";
	return scanres;
}


int LaTeX::operator()()
{
#ifndef __EMX__
	string tmp = cmd + ' ' + file + " > /dev/null";
#else // cmd.exe (OS/2) causes SYS0003 error at "/dev/null"
	string tmp = cmd + ' ' + file + " > nul";
#endif
        Systemcalls one;
	return one.Startscript(Systemcalls::System, tmp);
}


bool LaTeX::runMakeIndex(string const &file)
{
	lyxerr[Debug::LATEX] 
		<< "idx file has been made,"
		" running makeindex on file "
		<< file << '\n';

	// It should be possible to set the switches for makeindex
	// sorting style and such. It would also be very convenient
	// to be able to make style files from within LyX. This has
	// to come for a later time. (0.13 perhaps?)
	string cmd = "makeindex -c -q ";
	cmd += file;
	Systemcalls one;
	one.Startscript(Systemcalls::System, cmd);
	return true;
}


bool LaTeX::runBibTeX(string const &file)
{
	LyXLex lex(0, 0);
	string token;
	if (!lex.setFile(file)) {
		// unable to open .aux file
		// return at once
		return false;
	}

	while (lex.IsOK()) {
		if (lex.EatLine())
			token=lex.GetString();
		else // blank line in the file being read
			continue;

		if (contains(token, "\\bibdata{")) {
			// run bibtex and
			string cmd="bibtex ";
			cmd += ChangeExtension(file, string(), true);
			Systemcalls one;
			one.Startscript(Systemcalls::System, cmd);
			return true;
		}
		
	}
	// bibtex was not run.
	return false;
}


int LaTeX::scanLogFile(TeXErrors &terr)
{
	string token;
	int retval = NO_ERRORS;
	
	LyXLex lex(0, 0);

	string tmp = ChangeExtension(file, ".log", true);
	
	if (!lex.setFile(tmp)) {
		// unable to open file
		// return at once
		retval |= NO_LOGFILE;
		return retval;
	}
	
	while (lex.IsOK()) {
		if (lex.EatLine())
			token = lex.GetString();
		else // blank line in the file being read
			continue;

		lyxerr[Debug::LATEX] << token << '\n';
		
		if (prefixIs(token, "LaTeX Warning:")) {
			// Here shall we handle different
			// types of warnings
			retval |= LATEX_WARNING;
			lyxerr[Debug::LATEX] << "LaTeX Warning.\n";
			if (contains(token, "Rerun to get cross-references")) {
				retval |= RERUN;
				lyxerr[Debug::LATEX] << "We should rerun.\n";
			} else if (contains(token, "Citation")
				   && contains(token, "on page")
				   && contains(token, "undefined")) {
				retval |= UNDEF_CIT;
			}
		} else if (prefixIs(token, "Package")) {
			// Package warnings
			retval |= PACKAGE_WARNING;
			if (contains(token, "natbib Warning:")) {
				// Natbib warnings
				if (contains(token, "Citation")
				    && contains(token, "on page")
				    && contains(token, "undefined")) {
					retval |= UNDEF_CIT;
				}
			} else if (contains(token, "Rerun LaTeX.")) {
				// at least longtable.sty might use this.
				retval |= RERUN;
			}
		} else if (prefixIs(token, "! LaTeX Error:")) {
			// Here shall we handle different
			// types of errors
			retval |= LATEX_ERROR;
			lyxerr[Debug::LATEX] << "LaTeX Error.\n";
			// this is not correct yet
			terr.scanError(lex);
			++num_errors;
		} else if (prefixIs(token, "! ")) {
			// TeX Error
			retval |=  TEX_ERROR;
			lyxerr[Debug::LATEX] << "TeX Error.\n";
			terr.scanError(lex);
			++num_errors;
		} else {
			// information messages, TeX warnings and other
			// warnings we have not caught earlier.
			if (prefixIs(token, "Overfull ")) {
				retval |= TEX_WARNING;
			} else if (prefixIs(token, "Underfull ")) {
				retval |= TEX_WARNING;
			} else if (contains(token, "Rerun to get citations")) {
				// Natbib seems to use this.
				retval |= RERUN;
			}
		}
	}	
	return retval;
}


void LaTeX::deplog(DepTable &dep)
{
	string tmp = ChangeExtension(file, ".log", true);
	LyXLex lex(0, 0);
	lex.setFile(tmp);
	while (lex.IsOK()) {
		if (!lex.EatLine()) continue;
		
		string token = lex.GetString();

		if (token.empty()) continue;
		
		// skip forward in the file until we reach
		// " *File List*"
		if (contains(token, "*File List*")) {
			readFileList(lex, dep);
		}
		// Here either the log file has been read or not,
		// we don't have to care either way.
	}
}


void LaTeX::readFileList(LyXLex &lex, DepTable &dep)
{
	lyxerr[Debug::LATEX] << "We got \"*File List*\"!\n";
	string token;
	while (lex.IsOK()) {
		// now read line by line until we reach
		// " ***********"
		
		if (!lex.EatLine()) continue;
		
		token = frontStrip(lex.GetString());
		
		if (contains(token, "**********")) {
			lyxerr[Debug::LATEX] << "End of file section.\n";
			return;
		}
		
		string foundfile;
		token = split(token, foundfile, ' ');
		lyxerr[Debug::LATEX] << "examining file: " 
				     << foundfile << '\n';
		// Ok now we found a file.
		// Now we should make sure that this is a file
		// that we can access through the normal paths:
		// (1) Foundfile is an absolute path and should
		//     be inserted.
		if (AbsolutePath(foundfile)) {
			dep.insert(foundfile);
			lyxerr[Debug::LATEX] << "Found file:"
					     << foundfile << '\n';
			continue;
		}
		// (2) The foundfile can be found in the same dir
		//     as the .lyx file and should be inserted.
		Path p(path);
		if (FileInfo(foundfile).exist()) {
			dep.insert(MakeAbsPath(foundfile));
			lyxerr[Debug::LATEX] << "Found file:" 
					     << foundfile << '\n';
			p.pop();
			continue;
		}
		p.pop();
		
		// (3) foundfile is in the tmpdir
		//     insert it into dep
		if (FileInfo(OnlyFilename(foundfile)).exist()) {
			dep.insert(OnlyFilename(foundfile));
			lyxerr[Debug::LATEX] 
				<< "Found file:" 
				<< OnlyFilename(foundfile) << '\n';
			continue;
		}
	}
}


void LaTeX::deptex(DepTable &dep)
{
	int except = AUX|LOG|DVI|BBL|IND|GLO; 
	string tmp;
	FileInfo fi;
	for (int i = 0; i < file_count; ++i) {
		if (!(all_files[i].file & except)) {
			tmp = ChangeExtension(file,
					      all_files[i].extension,
					      true);
			lyxerr[Debug::LATEX] << "deptex: " << tmp << '\n';
			if (fi.newFile(tmp).exist())
				dep.insert(tmp);
		}
	}
}
