// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1998 The LyX Team
 *
 *======================================================*/

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "TextFormat.h"

TextFormat::TextFormat(Encoding & inenc_, Encoding & outenc_, 
		       LString eol_, int width_)
	: FileFormat(inenc_, outenc_), eol(eol_), width(width_)
{ 
	right = 0;
	left = 0;
	parindent = 0;
	column = 0;
	lineno = 0;
	wordwrap = true;
	breakwords = true;
}

TextFormat::~TextFormat() {
	close();
}

void TextFormat::close() {
	if (!line.empty()) {
		print(line.c_str());
		line.erase();
		linebreak();
	}
	FileFormat::close();
}

void TextFormat::write(LString const & str, bool doconv) {
	LString output;
	if (doconv) {
		// Convert the string to the target format
		output = outenc->convert(*(inenc), str);
	} else {
		output = str;
	}

	char const * out = output.c_str();
	int length = strlen(out);

	// Should we word wrap?
	if (!wordwrap) {
		line = line + output;
		column += length;
		return;
	}

	// Ok, we are word wrapping
	// Add one character at a time, and flush when needed
	for (int i=0 ; i < length; ++i) {
		char c = out[i];
		if (available() <= 0) {
			if (c==' ') {
				flushline();
				// Eat the space
				continue;
			}
			splitline();
		} 
		if (c==' ') {
			spaces.push_back(column);
		}
		line = line + c;
		++column;
	}
}

void TextFormat::write(fmt::Symbol symb) {
	switch (symb) {
	case fmt::br:        linebreak(); break;
	case fmt::par:       parstart(); break;
	case fmt::nbsp:      nbsp(); break;
	default: break;
	}
}

void TextFormat::splitline() {
	// Do we have an earlier break-point?
	if (spaces.size() != 0) {
		// Yes, use the latest
		int brcol = *(spaces.rbegin());
		LString head = line;
		LString tail = line;
		if (brcol - 1 >= 0) {
			head.substring(0, brcol - 1);
		} else {
			head.erase();
		}
		if (brcol + 1 < tail.length()) {
			tail.substring(brcol + 1, tail.length() - 1);
		} else {
			tail.erase();
		}
		line = head;
		flushline();
		line = tail;
	} else if (breakwords) {
		// No, but we are asked to force a break
		flushline();
	}
}

void TextFormat::linebreak() {
	flushline();
	print(eol.c_str());
	column = 0;
	leftmargin = left;
	++lineno;
}

void TextFormat::parstart() {
	leftmargin = parindent + left;
}

void TextFormat::nbsp() {
	splitline();
	line += " ";
	++column;
}

int TextFormat::available() {
	return width - (leftmargin + right + column);
}

void TextFormat::flushline() {
	// Do we have anything pending?
	if (!line.empty()) {
		// Yes, first do the margin
		line = doindent(leftmargin) + line;
		// Let the derivate classes have a look at it
		processLine(line);
		// and then print the line
		print(line.c_str());
		line.erase();
	}
	spaces.clear();
}

LString TextFormat::doindent(int c) {
	int tabs = 0;
	int rem = c;
	if (usetabs()) {
		tabs = c / 8;
		rem = c % 8;
	}
	LString res;
	for (int t=0; t < tabs; ++t) {
		res += "\t";
	}
	for (int r=0; r < rem; ++r) {
		res += " ";
	}
	return res.c_str();
}
