/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 *
 *	     Copyright (C) 1995 Matthias Ettrich
 *           Copyright (C) 1995-1999 The LyX Team.
 *
 *           This file is Copyright (C) 1996-1999
 *           Lars Gullik Bjnnes
 *
 * ======================================================
 */

#include <config.h>

#include <stdlib.h>
#include <unistd.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "definitions.h"
#include "buffer.h"
#include "bufferlist.h"
#include "lyx_main.h"
#include "lyx_gui_misc.h"
#include "lyxrc.h"
#include "lyxlex.h"
#include "tex-strings.h"
#include "linuxdoc.h"
#include "textclasslist.h"
#include "lyx_cb.h"
#include "minibuffer.h"
#include "lyxfont.h"
#include "formulamacro.h"
#include "lyxinset.h"
#include "inseterror.h"
#include "insetlabel.h"
#include "insetref.h"
#include "insetinfo.h"
#include "insetquotes.h"
#include "insetlatexaccent.h"
#include "insetbibkey.h"
#include "insetbibtex.h"
#include "insetcitation.h"
#include "insetindex.h" 
#include "insetinclude.h"
#include "insettoc.h"
#include "insetlof.h"
#include "insetlot.h"
#include "insetloa.h"
#include "insetparent.h"
#include "insetspecialchar.h"
#include "insetfig.h"
#include "insettabular.h"
#include "insetert.h"
#include "insetfoot.h"
#include "insetnote.h"
#include "filetools.h"
#include "path.h"
#include "LaTeX.h"
#include "Chktex.h"
#include "LyXView.h"
#include "debug.h"
#include "LaTeXFeatures.h"
#include "syscall.h"
#include "lyxlib.h"
#include "FileInfo.h"
#include "gettext.h"
#include "LAssert.h"
#include "lyxtext.h"

// 	$Id: buffer.C,v 1.32 1998/10/22 16:56:07 larsbj Exp $	

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: buffer.C,v 1.32 1998/10/22 16:56:07 larsbj Exp $";
#endif /* lint */

// all these externs should eventually be removed.
extern BufferList bufferlist;

static const float LYX_FORMAT = 3.0;

extern int tex_code_break_column;


Buffer::Buffer()
{
	// find out how often this is run
	lyxerr << "Running Buffer::Buffer()\n";
	lyxvc.setBuffer(this);
	paragraph = 0;
	//the_locking_inset = 0;
	lyx_clean = true;
	bak_clean = true;
	dvi_clean = false;
	dep_clean = 0;
	users = 0;
}


Buffer::Buffer(string const & file, LyXRC *lyxrc, bool ronly)
{
	// find out how often this is run
	lyxerr << "Running Buffer::Buffer(string const & file,"
		" LyXRC *lyxrc, bool ronly)\n";
	lyxvc.setBuffer(this);
	filename = file;
	filepath = OnlyPath(file);
	paragraph = 0;
	//the_locking_inset = 0;
	lyx_clean = true;
	bak_clean = true;
	dvi_clean = false;
	dep_clean = 0;
	read_only = ronly;
	users = 0;
	if (read_only || (lyxrc && lyxrc->use_tempdir)) {
		tmppath = CreateBufferTmpDir();
	} else {
		tmppath = filepath;
	}
}


Buffer::Buffer(const Buffer &buf)
{
	// find out how often this is run
	lyxerr << "Running Buffer::Buffer(const Buffer &buf)\n";
	// We are in deep sh** if we try to initialize from a buffer in use
	Assert(buf.paragraph == 0);

	paragraph = buf.paragraph;
	filename = buf.filename;
	filepath = buf.filepath;
	tmppath = buf.tmppath;
	buf.tmppath.erase(); // make the old buffer forget the dir, then we
	// take resposibility for it.
	lyx_clean = buf.lyx_clean;
	bak_clean = buf.bak_clean;
	dvi_clean = buf.dvi_clean;
	dep_clean = buf.dep_clean;
	read_only = buf.read_only;
	format = buf.format;
	users = buf.users;
	//the_locking_inset = buf.the_locking_inset;
	lyxvc.setBuffer(this);
}


Buffer::~Buffer()
{
	lyxerr[Debug::INFO] << "Running Buffer::~Buffer()\n";
	// here the buffer should take care that it is
	// saved properly, before it goes into the void.

	// make sure that views using this buffer
	// forgets it.
	if (users)
		users->setBuffer(0);
	
	if (!tmppath.empty()) {
		DestroyBufferTmpDir(tmppath);
	}
	
	LyXParagraph *par = paragraph;
	LyXParagraph *tmppar;
	while (par) {
		tmppar = par->next;
		delete par;
		par = tmppar;
	}
	paragraph = 0;
}


bool Buffer::saveParamsAsDefaults()
{
	string fname = AddName(AddPath(user_lyxdir, "templates/"),
				"defaults.lyx");
	Buffer defaults = Buffer(fname);
	
	// Use the current buffer's parameters as default
	defaults.params = params;
	
	// add an empty paragraph. Is this enough?
	defaults.paragraph = new LyXParagraph();

	return defaults.writeFile(defaults.filename,false);
}


/// Update window titles of all users
// Should work on a list
/* This func should be removed. the buffer should usually never initiate
   operations in the UI. we should rather use a signal like "titleChanged"
   and "statusChanged", we should perhaps make a getStatusString for the UI
   to use */
void Buffer::updateTitles()
{
	if (users) users->getOwner()->updateWindowTitle();
}


/// Reset autosave timer of all users
// Should work on a list
void Buffer::resetAutosaveTimers()
{
	if (users) users->getOwner()->resetAutosaveTimer();
}


void Buffer::setFileName(string const & newfile)
{
	filename = MakeAbsPath(newfile);
	filepath = OnlyPath(filename);
	setReadonly(IsFileWriteable(filename) == 0);
	updateTitles();
}

#if 0
void Buffer::InsetUnlock()
{
	if (the_locking_inset) {
		if (!inset_slept) the_locking_inset->InsetUnlock();
		the_locking_inset = 0;
		users->text->FinishUndo();
		inset_slept = false;
	}
}
#endif

#warning should probably be moved to BufferView (parts of it)
// the rest of it should be here, and modified to take a FILE* as 
// argument.

// Inserts a file into current document
bool Buffer::insertLyXFile(string const & filen)
	//
	// (c) CHT Software Service GmbH
	// Uwe C. Schroeder
	//
	// Insert a Lyxformat - file into current buffer
	//
	// Moved from lyx_cb.C (Lgb)
{
	if (filen.empty()) return false;

	string filename = MakeAbsPath(filen);

	// check if file exist
	FileInfo fi(filename);

	if (!fi.exist() || !fi.readable()) {
		WriteAlert(_("Error!"),
			   _("Cannot open specified file:"),
			   MakeDisplayPath(filename,50));
		return false;
	}
	
	users->beforeChange();

	FilePtr myfile(filename, FilePtr::read);
	if (!myfile()) {
		WriteAlert(_("Error!"),
			   _("Cannot open specified file:"),
			   MakeDisplayPath(filename,50));
		return false;
	}
	LyXLex lex(0, 0);
	lex.setFile(myfile);
	int c = fgetc(myfile());
	ungetc(c, myfile);

	bool res = true;

	if (c=='#') {
		lyxerr[Debug::INFO] << "Will insert file with header\n";
		res = readFile(lex, users->text->cursor.par);
	} else {
		lyxerr[Debug::INFO] << "Will insert file without header\n";
		res = readLyXformat2(lex, users->text->cursor.par);
	}
	resize();
	return res;
}


bool Buffer::readLyXformat2(LyXLex &lex, LyXParagraph *par)
{
	string tmptok;
	int pos = 0;
	int tmpret, tmpret2;
	char depth = 0; // signed or unsigned?
	LyXParagraph::footnote_flag footnoteflag = LyXParagraph::NO_FOOTNOTE;
        LyXParagraph::footnote_kind footnotekind = LyXParagraph::FOOTNOTE;
        bool the_end_read = false;
	
        LyXParagraph *return_par = 0;
        LyXFont font = LyXFont(LyXFont::ALL_INHERIT);
	
        // If we are inserting, we cheat and get a token in advance
        bool has_token = false;
        string pretoken;
	
        if(!par) {
                par = new LyXParagraph();
        } else {
                users->text->BreakParagraph();
                return_par = users->text->FirstParagraph();
                pos=0;
                markDirty();
                // We don't want to adopt the parameters from the
                // document we insert, so we skip until the text begins:
                while (lex.IsOK()) {
                        lex.nextToken();
                        pretoken = lex.GetString();
                        if (pretoken == "\\layout") {
                                has_token = true;
                                break;
                        }
                }
        }
	
        while (lex.IsOK()) {
                if (has_token) {
                        has_token = false;
                } else {
                        lex.nextToken();
                        pretoken = lex.GetString();
                }
		
                // Profiling show this should give a lot: (Asger)
                string const token = pretoken;
		
                if (token.empty())
                        continue;
                else if (token[0] != '\\') {
                        int n = token.length();
                        for (int i=0; i < n; ++i) {
                                par->InsertChar(pos, token[i]);
                                par->SetFont(pos, font);
                                ++pos;
                        }
                } else if (token == "\\i") {
                        Inset *inset = new InsetLatexAccent;
                        inset->Read(lex);
                        par->InsertChar(pos, LYX_META_INSET); 
                        
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\layout") {
                        if (!return_par) 
                                return_par = par;
                        else 
                                par = new LyXParagraph(par);
                        pos = 0;
                        lex.EatLine();
                        par->layout =
                                textclasslist.numberOfLayout(params.textclass,
                                                        lex.GetString());
                        if (par->layout == -1) // layout not found
                                // use default layout "Standard" (0)
                                par->layout = 0;
                        // Test whether the layout is obsolete.
                        LyXLayout const * layout = textclasslist.style(params.textclass,
                                                           par->layout); 
                        if (!layout->obsoleted_by.empty())
                                par->layout = 
                                        textclasslist.numberOfLayout(params.textclass,
								layout->obsoleted_by);
                        par->footnoteflag = footnoteflag;
                        par->footnotekind = footnotekind;
                        par->depth = depth;
                        font = LyXFont(LyXFont::ALL_INHERIT);
                } else if (token == "\\end_float") {
                        if (!return_par) 
                                return_par = par;
                        else 
                                par = new LyXParagraph(par);
                        footnotekind = LyXParagraph::FOOTNOTE;
                        footnoteflag = LyXParagraph::NO_FOOTNOTE;
                        pos = 0;
                        lex.EatLine();
                        par->layout = LYX_DUMMY_LAYOUT;
                        font = LyXFont(LyXFont::ALL_INHERIT);
                } else if (token == "\\begin_float") {
                        tmpret = lex.FindToken(string_footnotekinds);
                        if (tmpret == -1) ++tmpret;
                        if (tmpret != LYX_LAYOUT_DEFAULT) 
                                footnotekind = (LyXParagraph::footnote_kind)tmpret; // bad
                        if (footnotekind == LyXParagraph::FOOTNOTE
                            || footnotekind == LyXParagraph::MARGIN)
                                footnoteflag = LyXParagraph::CLOSED_FOOTNOTE;
                        else 
                                footnoteflag = LyXParagraph::OPEN_FOOTNOTE;
                } else if (token == "\\begin_deeper") {
                        ++depth;
                } else if (token == "\\end_deeper") {
                        if (!depth) {
                                lex.printError("\\end_deeper: "
                                               "depth is already 0");
                        }
                        else
                                --depth;
                } else if (token == "\\begin_preamble") {
                        params.readPreamble(lex);
                } else if (token == "\\textclass") {
                        lex.EatLine();
                        params.textclass = textclasslist.numberOfClass(lex.GetString(
));
                        if (params.textclass == -1) {
                                lex.printError("Unknown textclass `$$Token'");
                                params.textclass = 0;
                        }
                        if (!textclasslist.load(params.textclass)) {
                                // if the textclass wasn't loaded properly
                                // we need to either substitute another
                                // or stop loading the file.
                                // I can substitute but I don't see how I can
                                // stop loading... ideas??  ARRae980418
                                WriteAlert(_("Textclass Loading Error!"),
                                           string(_("Can't load textclass ")) +
                                           textclasslist.nameOfClass(params.textclass),
                                           _("-- substituting default"));
                                params.textclass = 0;
                        }
                } else if (token == "\\options") {
                        lex.EatLine();
                        params.options = lex.GetString();
                } else if (token == "\\language") {
                        params.readLanguage(lex);           
                } else if (token == "\\fontencoding") {
                        lex.EatLine();
                } else if (token == "\\inputencoding") {
                        lex.EatLine();
                        params.inputenc = lex.GetString();
                } else if (token == "\\graphics") {
                        params.readGraphicsDriver(lex);
                } else if (token == "\\fontscheme") {
                        lex.EatLine();
                        params.fonts = lex.GetString();
                } else if (token == "\\noindent") {
                        par->pparams.noindent = true;
                } else if (token == "\\fill_top") {
                        par->pparams.added_space_top = VSpace::VFILL;
                } else if (token == "\\fill_bottom") {
                        par->pparams.added_space_bottom = VSpace::VFILL;
                } else if (token == "\\line_top") {
                        par->pparams.line_top = true;
                } else if (token == "\\line_bottom") {
                        par->pparams.line_bottom = true;
                } else if (token == "\\pagebreak_top") {
                        par->pparams.pagebreak_top = true;
                } else if (token == "\\pagebreak_bottom") {
                        par->pparams.pagebreak_bottom = true;
                } else if (token == "\\start_of_appendix") {
                        par->start_of_appendix = true;
                } else if (token == "\\paragraph_separation") {
                        tmpret = lex.FindToken(string_paragraph_separation);
                        if (tmpret == -1) ++tmpret;
                        if (tmpret != LYX_LAYOUT_DEFAULT) 
                                params.paragraph_separation = tmpret;
                } else if (token == "\\defskip") {
                        lex.nextToken();
                        params.defskip = VSpace(lex.GetString());
                } else if (token == "\\no_isolatin1") {
                        lex.nextToken();
                } else if (token == "\\no_babel") {
                        lex.nextToken();
                } else if (token == "\\no_epsfig") {
                        lex.nextToken();
//              } else if (token == "\\epsfig") { // obsolete
                        // Indeed it is obsolete, but we HAVE to be backwards
                        // compatible until 0.14, because otherwise all figures
                        // in existing documents are irretrivably lost. (Asger)
//                      params.readGraphicsDriver(lex);
		} else if (token == "\\quotes_language") {
                        tmpret = lex.FindToken(string_quotes_language);
                        if (tmpret == -1) ++tmpret;
                        if (tmpret != LYX_LAYOUT_DEFAULT) {
                                InsetQuotes::quote_language tmpl =
                                        InsetQuotes::EnglishQ;
                                switch(tmpret) {
                                case 0:
                                        tmpl = InsetQuotes::EnglishQ;
                                        break;
                                case 1:
                                        tmpl = InsetQuotes::SwedishQ;
                                        break;
                                case 2:
                                        tmpl = InsetQuotes::GermanQ;
                                        break;
                                case 3:
                                        tmpl = InsetQuotes::PolishQ;
                                        break;
                                case 4:
                                        tmpl = InsetQuotes::FrenchQ;
                                        break;
                                case 5:
                                        tmpl = InsetQuotes::DanishQ;
                                        break;  
                                }
                                params.quotes_language = tmpl;
                        }
                } else if (token == "\\quotes_times") {
                        lex.nextToken();
                        switch(lex.GetInteger()) {
                        case 1: 
                                params.quotes_times = InsetQuotes::SingleQ; 
                                break;
                        case 2: 
                                params.quotes_times = InsetQuotes::DoubleQ; 
                                break;
                        }
                } else if (token == "\\papersize") {
                                tmpret = lex.FindToken(string_papersize);
                        if (tmpret == -1)
                                ++tmpret;
                        else
                                params.papersize2 = tmpret;
                } else if (token == "\\paperpackage") {
                        tmpret = lex.FindToken(string_paperpackages);
                        if (tmpret == -1) {
                                ++tmpret;
                                params.paperpackage = PACKAGE_NONE;
                        } else
                                params.paperpackage = tmpret;
                } else if (token == "\\use_geometry") {
                        lex.nextToken();
                        params.use_geometry = lex.GetInteger();
                } else if (token == "\\use_amsmath") {
                        lex.nextToken();
                        params.use_amsmath = lex.GetInteger();
                } else if (token == "\\paperorientation") {
                        tmpret = lex.FindToken(string_orientation);
                        if (tmpret == -1) ++tmpret;
                        if (tmpret != LYX_LAYOUT_DEFAULT) 
                                params.orientation = tmpret;
                } else if (token == "\\paperwidth") {
                        lex.next();
                        params.paperwidth = lex.GetString();
                } else if (token == "\\paperheight") {
                        lex.next();
                        params.paperheight = lex.GetString();
                } else if (token == "\\leftmargin") {
                        lex.next();
                        params.leftmargin = lex.GetString();
                } else if (token == "\\topmargin") {
                        lex.next();
                        params.topmargin = lex.GetString();
                } else if (token == "\\rightmargin") {
                        lex.next();
                        params.rightmargin = lex.GetString();
                } else if (token == "\\bottommargin") {
                        lex.next();
                        params.bottommargin = lex.GetString();
                } else if (token == "\\headheight") {
                        lex.next();
                        params.headheight = lex.GetString();
                } else if (token == "\\headsep") {
                        lex.next();
                        params.headsep = lex.GetString();
                } else if (token == "\\footskip") {
                        lex.next();
                        params.footskip = lex.GetString();
                } else if (token == "\\paperfontsize") {
                        lex.nextToken();
                        params.fontsize = strip(lex.GetString());
                } else if (token == "\\papercolumns") {
                        lex.nextToken();
                        params.columns = lex.GetInteger();
                } else if (token == "\\papersides") {
                        lex.nextToken();
                        params.sides = lex.GetInteger();
                } else if (token == "\\paperpagestyle") {
                        lex.nextToken();
                        params.pagestyle = strip(lex.GetString());
                } else if (token == "\\bullet") {
                        lex.nextToken();
                        int index = lex.GetInteger();
                        lex.nextToken();
                        int temp_int = lex.GetInteger();
                        params.user_defined_bullets[index].setFont(temp_int);
                        params.temp_bullets[index].setFont(temp_int);
                        lex.nextToken();
                        temp_int = lex.GetInteger();
                        params.user_defined_bullets[index].setCharacter(temp_int);
                        params.temp_bullets[index].setCharacter(temp_int);
                        lex.nextToken();
                        temp_int = lex.GetInteger();
                        params.user_defined_bullets[index].setSize(temp_int);
                        params.temp_bullets[index].setSize(temp_int);
                        lex.nextToken();
                        string temp_str = lex.GetString();
                        if (temp_str != "\\end_bullet") {
                                // this element isn't really necessary for
                                // parsing but is easier for humans
                                // to understand bullets. Put it back and
                                // set a debug message?
                                lex.printError("\\end_bullet expected, got" + temp_str);
                                //how can I put it back?
                        }
                } else if (token == "\\bulletLaTeX") {
                        lex.nextToken();
                        int index = lex.GetInteger();
                        lex.next();
                        string temp_str = lex.GetString(), sum_str;
                        while (temp_str != "\\end_bullet") {
                                // this loop structure is needed when user
                                // enters an empty string since the first
                                // thing returned will be the \\end_bullet
                                // OR
                                // if the LaTeX entry has spaces. Each element
                                // therefore needs to be read in turn
                                sum_str += temp_str;
                                lex.next();
                                temp_str = lex.GetString();
                        }
                        params.user_defined_bullets[index].setText(sum_str);
                        params.temp_bullets[index].setText(sum_str);
                } else if (token == "\\secnumdepth") {
                        lex.nextToken();
                        params.secnumdepth = lex.GetInteger();
                } else if (token == "\\tocdepth") {
                        lex.nextToken();
                        params.tocdepth = lex.GetInteger();
                } else if (token == "\\spacing") {
                        lex.next();
                        string tmp = strip(lex.GetString());
                        if (tmp == "single") {
                                params.spacing.set(Spacing::Single);
                        } else if (tmp == "onehalf") {
                                params.spacing.set(Spacing::Onehalf);
                        } else if (tmp == "double") {
                                params.spacing.set(Spacing::Double);
                        } else if (tmp == "other") {
                                lex.next();
                                params.spacing.set(Spacing::Other,
                                                   lex.GetFloat());
                        } else {
                                lex.printError("Unknown spacing token: '$$Token'");
                        }
                } else if (token == "\\float_placement") {
                        lex.nextToken();
                        params.float_placement = lex.GetString();
                } else if (token == "\\family") { 
                        lex.next();
                        font.setLyXFamily(lex.GetString());
                } else if (token == "\\series") {
                        lex.next();
                        font.setLyXSeries(lex.GetString());
                } else if (token == "\\shape") {
                        lex.next();
                        font.setLyXShape(lex.GetString());
                } else if (token == "\\size") {
                        lex.next();
                        font.setLyXSize(lex.GetString());
                } else if (token == "\\latex") {
                        lex.next();
                        string tok = lex.GetString();
                        // This is dirty, but gone with LyX3. (Asger)
                        if (tok == "no_latex")
                                font.setLatex(LyXFont::OFF);
                        else if (tok == "latex")
                                font.setLatex(LyXFont::ON);
                        else if (tok == "default")
                                font.setLatex(LyXFont::INHERIT);
                        else
                                lex.printError("Unknown LaTeX font flag "
                                               "`$$Token'");
                } else if (token == "\\emph") {
                        lex.next();
                        font.setEmph(font.setLyXMisc(lex.GetString()));
                } else if (token == "\\bar") {
                        lex.next();
                        string tok = lex.GetString();
                        // This is dirty, but gone with LyX3. (Asger)
                        if (tok == "under")
                                font.setUnderbar(LyXFont::ON);
                        else if (tok == "no")
                                font.setUnderbar(LyXFont::OFF);
                        else if (tok == "default")
                                font.setUnderbar(LyXFont::INHERIT);
                        else
                                lex.printError("Unknown bar font flag "
                                               "`$$Token'");
                } else if (token == "\\noun") {
                        lex.next();
                        font.setNoun(font.setLyXMisc(lex.GetString()));
                } else if (token == "\\color") {
                        lex.next();
                        font.setLyXColor(lex.GetString());
                } else if (token == "\\align") {
                        tmpret = lex.FindToken(string_align);
                        if (tmpret == -1) ++tmpret;
                        if (tmpret != LYX_LAYOUT_DEFAULT) {
                                tmpret2 = 1;
                                for (; tmpret>0; --tmpret)
                                        tmpret2 = tmpret2 * 2;
                                par->pparams.align = tmpret2;
                        }
                } else if (token == "\\added_space_top"){
                        lex.nextToken();
                        par->pparams.added_space_top = lex.GetString();
                } else if (token == "\\added_space_bottom") {
                        lex.nextToken();
                        par->pparams.added_space_bottom = lex.GetString();
                } else if (token == "\\pextra_type") {
                        lex.nextToken();
                        par->pextra_type = lex.GetInteger();
                } else if (token == "\\pextra_width") {
                        lex.nextToken();
                        par->pextra_width = lex.GetString();
                } else if (token == "\\pextra_widthp") {
                        lex.nextToken();
                        par->pextra_widthp = lex.GetString();
                } else if (token == "\\pextra_alignment") {
                        lex.nextToken();
                        par->pextra_alignment = lex.GetInteger();
                } else if (token == "\\pextra_hfill") {
                        lex.nextToken();
                        par->pextra_hfill = lex.GetInteger();
                } else if (token == "\\pextra_start_minipage") {
                        lex.nextToken();
                        par->pextra_start_minipage = lex.GetInteger();
                } else if (token == "\\labelwidthstring") {
                        lex.EatLine();
                        par->labelwidthstring = lex.GetString();
                        /* do not delete this token, it is still needed! */ 
                } else if (token == "\\end_inset") {
                        /* simple ignore this. The insets do not have
                         *  to read this */
                        // but insets should read it, it is a part of
                        //the inset isn't it? Lgb.
                } else if (token == "\\begin_inset") {
			Inset *inset = 0;
                        lex.next();
                        tmptok = lex.GetString();
                        /* test the different insets */ 
                        if (tmptok == "Quotes") {
                                inset = new InsetQuotes(this);
                                inset->Read(lex);
                        } else if (tmptok == "\\i") {
                                inset = new InsetLatexAccent;
                                inset->Read(lex);
                        } else if (tmptok == "FormulaMacro") {
                                inset = new InsetFormulaMacro;
                                inset->Read(lex);
                        } else if (tmptok == "Formula") {
                                inset = new InsetFormula;
                                inset->Read(lex);
                        } else if (tmptok == "Figure") {
                                inset = new InsetFig(100,100, this);
                                inset->Read(lex);
			} else if (tmptok == "Tabular") {
				inset = new InsetTabular(this);
                                inset->Read(lex);
			} else if (tmptok == "ERT") {
				inset = new InsetERT(this);
                                inset->Read(lex);
                        } else if (tmptok == "Foot") {
				inset = new InsetFoot(this);
                                inset->Read(lex);
                        } else if (tmptok == "Info") {
                                inset = new InsetInfo;
                                inset->Read(lex);
                        } else if (tmptok == "Include") {
                                inset = new InsetInclude(string(), this);
                                inset->Read(lex);
                        } else if (tmptok == "LatexCommand") {
                                InsetCommand inscmd;
                                inscmd.Read(lex);
                                if (inscmd.getCmdName()=="cite") {
                                        inset = new InsetCitation(inscmd.getContents(), inscmd.getOptions());
                                } else if (inscmd.getCmdName()=="bibitem") {
                                        lex.printError("Wrong place for bibitem");
                                        inset = inscmd.Clone();
                                } else if (inscmd.getCmdName()=="BibTeX") {
                                        inset = new InsetBibtex(inscmd.getContents(), inscmd.getOptions(), this);
                                } else if (inscmd.getCmdName()=="index") {
                                        inset = new InsetIndex(inscmd.getContents());
                                } else if (inscmd.getCmdName()=="include") {
                                        inset = new InsetInclude(inscmd.getContents(), this);
                                } else if (inscmd.getCmdName()=="label") {
                                        inset = new InsetLabel(inscmd.getCommand());
                                } else if (inscmd.getCmdName() == "ref"
                                           || inscmd.getCmdName() == "pageref") {
                                        inset = new InsetRef(inscmd, this);
                                } else if (inscmd.getCmdName()=="tableofcontents") {
                                        inset = new InsetTOC(this);
                                } else if (inscmd.getCmdName()=="listoffigures"){
                                        inset = new InsetLOF(this);
                                } else if (inscmd.getCmdName()=="listofalgorithms") {
                                        inset = new InsetLOA(this);
                                } else if (inscmd.getCmdName()=="listoftables") {
                                        inset = new InsetLOT(this);
                                } else if (inscmd.getCmdName()=="printindex") {
                                        inset = new InsetPrintIndex(this);
                                } else if (inscmd.getCmdName()=="lyxparent") {
                                        inset = new InsetParent(inscmd.getContents(),this);
                                } else 
                                        // The following three are only for compatibility
                                        if (inscmd.getCmdName()=="-") {
                                                inset = new InsetSpecialChar(InsetSpecialChar::HYPHENATION);
                                        } else if (inscmd.getCmdName()=="@.") {
                                                inset = new InsetSpecialChar(InsetSpecialChar::END_OF_SENTENCE);
                                        } else if (inscmd.getCmdName()=="ldots") {
                                                inset = new InsetSpecialChar(InsetSpecialChar::LDOTS);
                                        } else
                                                inset = inscmd.Clone();
                               
                        }
			if (inset) {
				par->InsertChar(pos, LYX_META_INSET);
				par->InsertInset(pos, inset);
				par->SetFont(pos, font);
				++pos;
			} else {
				lex.printError("Unknown inset `$$Token'. "
					       "Inserting as text.");
			}
                } else if (token == "\\InsetQuotes") {
                        Inset *inset = new InsetQuotes(this);
                        inset->Read(lex);
                        par->InsertChar(pos, LYX_META_INSET); 
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\InsetFormula") {
                        Inset *inset = new InsetFormula;
                        inset->Read(lex);
                        par->InsertChar(pos, LYX_META_INSET); 
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\SpecialChar") {
                        Inset *inset = new InsetSpecialChar;
                        inset->Read(lex);
                        par->InsertChar(pos, LYX_META_INSET); 
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\Figure") {
                        Inset *inset = new InsetFig(100,100, this);
                        inset->Read(lex);
                        par->InsertChar(pos, LYX_META_INSET); 
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\newline") {
                        par->InsertChar(pos, LYX_META_NEWLINE);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\LyXTable") {
			Inset *inset = new InsetTabular(this);
			inset->Read(lex);
			par->InsertChar(pos, LYX_META_INSET);
			par->InsertInset(pos, inset);
			par->SetFont(pos, font);
			++pos;
                } else if (token == "\\hfill") {
                        // now obsolete, but we have a back compability
                        Inset *inset = new InsetSpecialChar(InsetSpecialChar::HFILL);
                        par->InsertChar(pos, LYX_META_INSET);
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\protected_separator") {
                        // now obsolete, but we have a back compability
                        //par->InsertChar(pos, LYX_META_PROTECTED_SEPARATOR);
                        Inset *inset = new InsetSpecialChar(InsetSpecialChar::PROTECTED_SPACE);
                        par->InsertChar(pos, LYX_META_INSET);
                        par->InsertInset(pos, inset);
                        par->SetFont(pos, font);
                        ++pos;
                } else if (token == "\\bibitem") {  // ale970302
                        if (!par->bibkey)
                                par->bibkey = new InsetBibKey;
                        par->bibkey->Read(lex);                 
                }else if (token == "\\backslash") {
                        par->InsertChar(pos, '\\');
                        par->SetFont(pos, font);
                        ++pos;
                }else if (token == "\\the_end") {
                        the_end_read = true;
                } else {
                        // This should be insurance for the future: (Asger)
                        lex.printError("Unknown token `$$Token'. "
                                       "Inserting as text.");
                        int n = token.length();
                        for (int i=0; i < n; ++i) {
                                par->InsertChar(pos, token[i]);
                                par->SetFont(pos, font);
                                ++pos;
                        }
                }
        }
   
        if (!return_par)
                return_par = par;

        paragraph = return_par;
        
        return the_end_read;
}


// if par = 0 normal behavior
// else insert behavior
// Returns false if "\the_end" is not read for formats >= 2.13. (Asger)
bool Buffer::readLyXformat3(LyXLex &lex, LyXParagraph *par)
{
	bool has_token = false;
	string pretoken;

	if (par) {
		// we are inserting into the buffer, and we don't want
		// the parameters. so we skip to the beginning of the
		// body.
		while (lex.IsOK()) {
			lex.nextToken();
			pretoken = lex.GetString();
			if (pretoken == "\\begin_body") {
				has_token = true;
				break;
			}
		}
	}

	while (lex.IsOK()) {
		// instead of this strange construct we
		// should make a LyXLeX::pushToken
		if (has_token) {
			has_token = false;
		} else {
			lex.nextToken();
			pretoken = lex.GetString();
		}

		// Profiling show this should give a lot: (Asger)
		string const token = pretoken;
		
		if (token.empty())
			continue;
		else if (token == "\\begin_parameters") {
			params.read(lex);
		} else if (token == "\\end_parameters") {
			// This should be stepped in params.read(...).
			lex.printError("\\end_parameters "
				       "without \\begin_parameters");
		} else if (token == "\\begin_body") {
			readBody(lex, par);
		} else if (token == "\\end_body") {
			// This should be stepped in readBody(...).
			lex.printError("\\end_body "
				       "without \\begin_body");
		} else if (token == "\\the_end") {
			pretoken = token;
 			break;
 		} else {
 			lex.printError("Unknown token `$$Token'. "
 				       "Throwing it away.");
		}
 	}
   
	if (pretoken != "\\the_end") {
		// \the_end is missing at the end of the file
		return false;
	} else {
		return true;
	}
}


void Buffer::readBody(LyXLex &lex, LyXParagraph *par)
{
	string token;
	string tmptok;
	int pos = 0;
	unsigned char depth = 0;
	LyXParagraph::footnote_flag footnoteflag = LyXParagraph::NO_FOOTNOTE;
	LyXParagraph::footnote_kind footnotekind = LyXParagraph::FOOTNOTE;

	LyXParagraph *return_par = 0;
	LyXFont font = LyXFont(LyXFont::ALL_INHERIT);

	if(!par) {
		par = new LyXParagraph();
	} else {
		users->text->BreakParagraph();
		return_par = users->text->FirstParagraph();
		pos=0;
		markDirty();
	}

	while (lex.IsOK()) {
		lex.nextToken();
		token = lex.GetString();

		if (token.empty())
			continue;
		else if (token[0] != '\\') {
			// Regular characters
			int n = token.length();
			for (int i=0; i < n; ++i) {
				par->InsertChar(pos, token[i]);
				par->SetFont(pos, font);
				++pos;
			}
		} else if (token == "\\begin_paragraph") {
			// Ideally this should be moved into LyXParagraph,
			// and it will be moved. Perhaps not possible before
			// some of the other _real_ structure is rewritten.
			if (!return_par) 
				return_par = par;
			else 
				par = new LyXParagraph(par);
			pos = 0;
			lex.EatLine();
			par->layout =
				textclasslist.numberOfLayout(params.textclass,
							lex.GetString());
			if (par->layout == -1) // Layout not found.
				// Use default layout "Standard" (0).
				par->layout = 0;
			// Test whether the layout is obsolete.
			LyXLayout const * layout = textclasslist.style(params.textclass,
							   par->layout); 
			if (!layout->obsoleted_by.empty())
				par->layout = 
					textclasslist.numberOfLayout(params.textclass, 
								layout->obsoleted_by);
			par->footnoteflag = footnoteflag;
			par->footnotekind = footnotekind;
			par->depth = depth;
			font = LyXFont(LyXFont::ALL_INHERIT);
		} else if (token == "\\end_paragraph") {
			// So far we are not using this actively that is
			// going to change later.
		} else if (token == "\\begin_paragraph_parameters") {
			par->pparams.read(lex);
		} else if (token == "\\end_paragraph_parameters") {
			lex.printError("\\end_paragraph_parameters without \\begin_paragraph_parameters");
// #ifndef NEWFLOATCODE
// 		} else if (token == "\\begin_float") {
// 			tmpret = lex.FindToken(string_footnotekinds);
// 			if (tmpret == -1) ++tmpret;
// 			if (tmpret != LYX_LAYOUT_DEFAULT) 
// 				footnotekind = (LyXParagraph::footnote_kind)tmpret; // bad
// 			if (footnotekind == LyXParagraph::FOOTNOTE
// 			    || footnotekind == LyXParagraph::MARGIN)
// 				footnoteflag = LyXParagraph::CLOSED_FOOTNOTE;
// 			else 
// 				footnoteflag = LyXParagraph::OPEN_FOOTNOTE;
// 		} else if (token == "\\end_float") {
// 			if (!return_par) 
// 				return_par = par;
// 			else 
// 				par = new LyXParagraph(par);
// 			footnotekind = LyXParagraph::FOOTNOTE;
// 			footnoteflag = LyXParagraph::NO_FOOTNOTE;
// 			pos = 0;
// 			lex.EatLine();
// 			par->layout = LYX_DUMMY_LAYOUT;
// 			font = LyXFont(LyXFont::ALL_INHERIT);
// #else
		} else if (token == "\\begin_float") {
			Inset *inset = 0;
			lex.next();
			tmptok = lex.GetString();
			// test the different insets
			if (tmptok == "footnote") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="margin") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="fig") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="tab") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="alg") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="wide-fig") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok =="wide-tab") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} 
			if (inset) {
				par->InsertChar(pos, LYX_META_INSET);
				par->InsertInset(pos, inset);
				par->SetFont(pos, font);
				++par;
			} else {
				lex.printError(
					"Unknown float `$$Token'. "
					"Please report this to the devies");
			}
		} else if (token == "\\end_float") {
			// The \begin_float should skip this.
			lex.printError("\\end_float without \\begin_float");
//#endif			
		} else if (token == "\\begin_deeper") {
			// when the bufferstructure changes this
			// tag will be removed.
			++depth;
		} else if (token == "\\end_deeper") {
			// when the bufferstructure changes this
			// tag will be removed.
			if (!depth) {
				lex.printError("\\end_deeper: "
					       "depth is already 0");
			}
			else
				--depth;
		} else if (token == "\\start_of_appendix") {
			par->start_of_appendix = true;
		} else if (token == "\\family") { 
			lex.next();
			font.setLyXFamily(lex.GetString());
		} else if (token == "\\series") {
			lex.next();
			font.setLyXSeries(lex.GetString());
		} else if (token == "\\shape") {
			lex.next();
			font.setLyXShape(lex.GetString());
		} else if (token == "\\size") {
			lex.next();
			font.setLyXSize(lex.GetString());
		} else if (token == "\\latex") {
			lex.next();
			string tok = lex.GetString();
			// This is dirty, but gone with LyX3. (Asger) sure?
			if (tok == "no_latex")
				font.setLatex(LyXFont::OFF);
			else if (tok == "latex")
				font.setLatex(LyXFont::ON);
			else if (tok == "default")
				font.setLatex(LyXFont::INHERIT);
			else
				lex.printError("Unknown LaTeX font flag "
					       "`$$Token'");
		} else if (token == "\\emph") {
			lex.next();
			font.setEmph(font.setLyXMisc(lex.GetString()));
		} else if (token == "\\bar") {
			lex.next();
			string tok = lex.GetString();
			// This is dirty, but gone with LyX3. (Asger)
			if (tok == "under")
				font.setUnderbar(LyXFont::ON);
			else if (tok == "no")
				font.setUnderbar(LyXFont::OFF);
			else if (tok == "default")
				font.setUnderbar(LyXFont::INHERIT);
			else
				lex.printError("Unknown bar font flag "
					       "`$$Token'");
		} else if (token == "\\noun") {
			lex.next();
			font.setNoun(font.setLyXMisc(lex.GetString()));
		} else if (token == "\\color") {
			lex.next();
			font.setLyXColor(lex.GetString());
                } else if (token == "\\pextra_type") {
                        lex.nextToken();
                        par->pextra_type = lex.GetInteger();
                } else if (token == "\\pextra_width") {
                        lex.nextToken();
			par->pextra_width = lex.GetString();
                } else if (token == "\\pextra_widthp") {
                        lex.nextToken();
			par->pextra_widthp = lex.GetString();
                } else if (token == "\\pextra_alignment") {
                        lex.nextToken();
                        par->pextra_alignment = lex.GetInteger();
                } else if (token == "\\pextra_hfill") {
                        lex.nextToken();
                        par->pextra_hfill = lex.GetInteger();
                } else if (token == "\\pextra_start_minipage") {
                        lex.nextToken();
                        par->pextra_start_minipage = lex.GetInteger();
		} else if (token == "\\labelwidthstring") {
			// Do not delete this token, it is still needed!
			lex.EatLine();
			par->labelwidthstring = lex.GetString();
		} else if (token == "\\begin_inset") {
			Inset *inset = 0;
			lex.next();
			tmptok = lex.GetString();
			// test the different insets
			if (tmptok == "Quotes") {
				inset = new InsetQuotes(this);
				inset->Read(lex);
			} else if (tmptok == "SpecialChar") {
				inset = new InsetSpecialChar;
				inset->Read(lex);
			} else if (tmptok == "LaTeXAccent"){
				inset = new InsetLatexAccent;
				inset->Read(lex);
			} else if (tmptok == "FormulaMacro") {
				inset = new InsetFormulaMacro;
				inset->Read(lex);
			} else if (tmptok == "Formula") {
				inset = new InsetFormula;
				inset->Read(lex);
			} else if (tmptok == "Tabular") {
				inset = new InsetTabular(this);
                                inset->Read(lex);
			} else if (tmptok == "ERT") {
				inset = new InsetERT(this);
                                inset->Read(lex);
			} else if (tmptok == "Foot") {
				inset = new InsetFoot(this);
                                inset->Read(lex);
			} else if (tmptok == "Figure") {
				inset = new InsetFig(100,100, this);
				inset->Read(lex);
			} else if (tmptok == "Info") {
				inset = new InsetInfo;
				inset->Read(lex);
			} else if (tmptok == "Footnote") {
				inset = new InsetNote(this);
				inset->Read(lex);
			} else if (tmptok == "Include") {
				inset = new InsetInclude(string(), this);
				inset->Read(lex);
			} else if (tmptok == "LatexCommand") {
#warning this block should be moved
				// either to its own function or inside
				// Insetcommand.
				InsetCommand inscmd;
				inscmd.Read(lex);
				if (inscmd.getCmdName()=="cite") {
					inset = new InsetCitation(inscmd.getContents(), inscmd.getOptions());
				} else if (inscmd.getCmdName()=="bibitem") {
					lex.printError("Wrong place for bibitem");
					inset = inscmd.Clone();
				} else if (inscmd.getCmdName()=="BibTeX") {
					inset = new InsetBibtex(inscmd.getContents(), inscmd.getOptions(), this);
				} else if (inscmd.getCmdName()=="index") {
					inset = new InsetIndex(inscmd.getContents());
				} else if (inscmd.getCmdName()=="include") {
					inset = new InsetInclude(inscmd.getContents(), this);
				} else if (inscmd.getCmdName()=="label") {
					inset = new InsetLabel(inscmd.getCommand());
				} else if (inscmd.getCmdName() == "ref"
					   || inscmd.getCmdName() == "pageref") {
					inset = new InsetRef(inscmd, this);
				} else if (inscmd.getCmdName()=="tableofcontents") {
					inset = new InsetTOC(this);
				} else if (inscmd.getCmdName()=="listoffigures") {
					inset = new InsetLOF(this);
				} else if (inscmd.getCmdName()=="listofalgorithms") {
					inset = new InsetLOA(this);
				} else if (inscmd.getCmdName()=="listoftables") {
					inset = new InsetLOT(this);
				} else if (inscmd.getCmdName()=="printindex") {
					inset = new InsetPrintIndex(this);
				} else if (inscmd.getCmdName()=="lyxparent") {
					inset = new InsetParent(inscmd.getContents(),this);
				}
			}       
			if (inset) {
				par->InsertChar(pos, LYX_META_INSET);
				par->InsertInset(pos, inset);
				par->SetFont(pos, font);
				++pos;
			} else {
				lex.printError(
					"Unknown inset `$$Token'. "
					"Please report this to the devies");
			}
		} else if (token == "\\end_inset") {
			// the insets should step this.
			// if they don't that is a bug
			lex.printError("\\end_inset "
				     "without \\begin_inset");
		} else if (token == "\\newline") {
			par->InsertChar(pos, LYX_META_NEWLINE);
			par->SetFont(pos, font);
			++pos;
			//} else if (token == "\\LyXTable") {
			//; // do nothing anymore here
//			par->table = new LyXTable(lex);
		} else if (token == "\\bibitem") {  // ale970302
		        if (!par->bibkey)
				par->bibkey = new InsetBibKey;
		        par->bibkey->Read(lex);		        
		}else if (token == "\\backslash") {
			par->InsertChar(pos, '\\');
			par->SetFont(pos, font);
			++pos;
		}else if (token == "\\end_body") {
			break;
		} else {
			// This should be insurance for the future: (Asger)
			lex.printError("Unknown token `$$Token'. "
				       "Inserting as text.");
			int n = token.length();
			for (int i=0; i < n; ++i) {
				par->InsertChar(pos, token[i]);
				par->SetFont(pos, font);
				++pos;
			}
		}
	}

	if (!return_par)
		return_par = par;

	paragraph = return_par;
}


bool Buffer::readFile(LyXLex &lex, LyXParagraph *par)
{
	string token;

	if (!lex.IsOK()) {
		WriteAlert(_("ERROR!"), _("Unable to read file!"));
		return false;
	}
	
	lex.next();
	token = lex.GetString();
	if (token != "\\lyxformat") { // the first token _must_ be...
		WriteAlert(_("ERROR!"), _("Not a LyX file!"));
		return false;
	}
		
	lex.next();
	format = lex.GetFloat();
	bool the_end = false;
	if (format <= 1) {
		WriteAlert(_("ERROR!"),
			   _("Old LyX file format found. "
			     "Use LyX 0.10.x to read this!"));
		return false;
	} else if (format >= 3) {
		lyxerr << "reading format 3\n";
		// Hey, we can't predict what will happen in the future!
		if (format > 3) {
			lyxerr << "maybe this won't work\n";
		}
		the_end = readLyXformat3(lex, par);
	} else if (format < 3 && format > 1) {
		lyxerr << "reading format 2\n";
		the_end = readLyXformat2(lex, par);
	}
	
	setPaperStuff();
	if (!the_end)
		WriteAlert(_("Warning!"),
			   _("Reading of document is not complete"),
			   _("Maybe the document is truncated"));
	// We simulate a safe reading anyway to allow
	// users to take the chance... (Asger)
	return true;
}
	    	    

// Returns false if unsuccesful
// if flag is false writeFile will not create any GUI
// warnings, only stderr.
// Needed for autosave in background or panic save (Matthias 120496)
bool Buffer::writeFile(string const & filename, bool flag)
{
	if (read_only && (filename == this->filename)) {
		// Here we should come with a question if we should
		// perform the write anyway.
		if (flag)
			lyxerr << _("Error! Document is read-only: ")
			       << filename << '\n';
		else
			WriteAlert(_("Error! Document is read-only: "),
				   filename);
		return false;
	}

	FileInfo finfo(filename);
	if (finfo.exist() && !finfo.writable()) {
		// Here we should come with a question if we should
		// try to do the save anyway. (i.e. do a chmod first)
		if (flag)
			lyxerr << _("Error! Cannot write file: ")
			       << filename << '\n';
		else
			WriteFSAlert(_("Error! Cannot write file: "),
				     filename);
		return false;
	}

	FilePtr file(filename, FilePtr::truncate);
	if (!file()) {
		if (flag)
			lyxerr << _("Error! Cannot write file: ")
			       << filename << '\n';
		else
			WriteFSAlert(_("Error! Cannot write file: "),
				     filename);
		return false;
	}
	
	// The top of the file should not be written by params.
	// collect some very important information
	string userName(getUserName()) ;

	// write out a comment in the top of the file
	fprintf(file,
		"#This file was created by <%s> %s",
		userName.c_str(),(char*)date());
	fprintf(file,
		"#LyX 1.1 (C) 1995-1998 Matthias Ettrich"
		" and the LyX Team\n");

	// at the very beginning the used lyx format
	fprintf(file, "\\lyxformat %.2f\n", LYX_FORMAT);

	// now write out the buffer parameters.
	fprintf(file, "\\begin_parameters\n");
	params.writeFile(file);
	fprintf(file, "\\end_parameters\n");

	char footnoteflag = 0;
	char depth = 0;

	// this will write out all the paragraphs
	// using recursive descent.
	fprintf(file, "\\begin_body\n");
	paragraph->writeFile(file, params, footnoteflag, depth);
	fprintf(file, "\n\\end_body\n");
	// And in between here we could put things like localwords.
	// Write marker that shows file is complete
	fprintf(file, "\\the_end\n");
	if (file.close()) {
		if (flag)
			lyxerr << _("Error! Could not close file properly: ")
			       << filename << '\n';
		else
			WriteFSAlert(_("Error! Could not close file properly: "),
				     filename);
		return false;
	}
	return true;
}


void Buffer::writeFileAscii(string const & /*filename*/, int /*linelen*/) 
{
	// Note: I am commenting this out so that I don't have to think
	// about making this work when restructuring the kernel
        // Who is I? It would be nice to know who puts comments for
        // for others to read inside the source (Jug)

	// Who do you think? I am the only one making more than
	// one line commens without signing it. (Lgb)
#warning reimplement writeFileAscii (with tabular-insets)
#if 0	
	FilePtr	file(filename, FilePtr::write);
	LyXFont
		font1, font2;
	Inset
		*inset;
	LyXParagraph
		*par = paragraph;
	char
		c,
		footnoteflag = 0,
		depth = 0;

	string
		fname1,
		tmp;
	
	int
		i,j,h,
		ltype=0,
		ltype_depth=0,
		noparbreak=0,
		islatex=0,
		*clen=0,
		actcell=0,
		actpos=0,
		cell=0,
	        cells=0,
	        currlinelen=0;
	long
		fpos=0;
	bool
		ref_printed = false;
   
   
	if (!file()) {
		WriteFSAlert(_("Error: Cannot write file:"), filename);
		return;
	}
	fname1=TmpFileName();
	while (par) {
		noparbreak = 0;
		islatex = 0;
		if (par->footnoteflag != LyXParagraph::NO_FOOTNOTE ||
		    !par->previous ||
		    par->previous->footnoteflag == LyXParagraph::NO_FOOTNOTE){
	 
			/* begins a footnote environment ? */ 
			if (footnoteflag != par->footnoteflag) {
				footnoteflag = par->footnoteflag;
				if (footnoteflag) {
					j=strlen(string_footnotekinds[par->footnotekind])+4;
					if (currlinelen+j > linelen)
						fprintf(file,"\n");
					fprintf(file, "([%s] ", 
						string_footnotekinds[par->footnotekind]);
					currlinelen += j;
				}
			}
	 
			/* begins or ends a deeper area ?*/ 
			if (depth != par->depth) {
				if (par->depth > depth) {
					while (par->depth > depth) {
						++depth;
					}
				}
				else {
					while (par->depth < depth) {
						--depth;
					}
				}
			}
	 
			/* First write the layout */
			tmp = textclasslist.NameOfLayout(params.textclass,par->layout);
			if (tmp == "Itemize") {
				ltype = 1;
				ltype_depth = depth+1;
			} else if (tmp =="Enumerate") {
				ltype = 2;
				ltype_depth = depth+1;
			} else if (strstr(tmp.c_str(),"ection")) {
				ltype = 3;
				ltype_depth = depth+1;
			} else if (strstr(tmp.c_str(),"aragraph")) {
				ltype = 4;
				ltype_depth = depth+1;
			} else if (tmp == "Description") {
				ltype = 5;
				ltype_depth = depth+1;
			} else if (tmp == "Abstract") {
				ltype = 6;
				ltype_depth = 0;
			} else if (tmp == "Bibliography") {
				ltype = 7;
				ltype_depth = 0;
			} else {
				ltype = 0;
				ltype_depth = 0;
			}
	 
			/* maybe some vertical spaces */ 

			/* the labelwidthstring used in lists */ 
	 
			/* some lines? */ 
	 
			/* some pagebreaks? */ 
	 
			/* noindent ? */ 
	 
			/* what about the alignment */ 
		} else {
			/* dummy layout, that means a footnote ended */ 
			footnoteflag = LyXParagraph::NO_FOOTNOTE;
			fprintf(file, ") ");
			noparbreak = 1;
		}
		/* It might be a table */ 
		if (par->table){
			if (!lyxrc->ascii_roff_command.empty() &&
                            lyxrc->ascii_roff_command != "none") {
				RoffAsciiTable(file,par);
				par = par->next;
				continue;
			}
			cell = 1;
                        actcell = 0;
			cells = par->table->columns;
			clen = new int [cells];
			memset(clen,0,sizeof(int)*cells);
			for (i = 0, j = 0, h=1; i < par->last; ++i, ++h) {
				c = par->GetChar(i);
				if (c == LYX_META_INSET) {
					if ((inset = par->GetInset(i))) {
						FilePtr fp(fname1,
							   FilePtr::write);
						if (!fp()) {
							WriteFSAlert(_("Error: Cannot open temporary file:"), fname1);
							return;
						}
						inset->Latex(fp,-1);
						h += ftell(fp) - 1;
						remove(fname1.c_str());
					}
				} else if (c == LYX_META_NEWLINE) {
					if (clen[j] < h)
						clen[j] = h;
					h = 0;
					j = (++j) % par->table->NumberOfCellsInRow(actcell);
                                        ++actcell;
				}
			}
			if (clen[j] < h)
				clen[j] = h;
		}
		font1 = LyXFont(LyXFont::ALL_INHERIT);
                actcell=0;
		for (i = 0,actpos=1; i < par->last; ++i, ++actpos) {
			if (!i && !footnoteflag && !noparbreak){
				fprintf(file, "\n\n");
				for(j=0; j<depth; ++j)
					fprintf(file, "  ");
				currlinelen = depth*2;
				switch(ltype) {
				case 0: /* Standart */
				case 4: /* (Sub)Paragraph */
                                case 5: /* Description */
					break;
				case 6: /* Abstract */
					fprintf(file, "Abstract\n\n");
					break;
				case 7: /* Bibliography */
					if (!ref_printed) {
						fprintf(file, "References\n\n");
						ref_printed = true;
					}
					break;
				default:
					fprintf(file,"%s ",par->labelstring.c_str());
					break;
				}
				if (ltype_depth > depth) {
					for(j=ltype_depth-1; j>depth; --j)
						fprintf(file, "  ");
					currlinelen += (ltype_depth-depth)*2;
				}
				if (par->table) {
					for(j=0;j<cells;++j) {
						fputc('+',file);
						for(h=0; h<(clen[j]+1); ++h)
							fputc('-',file);
					}
					fprintf(file,"+\n");
					for(j=0; j<depth; ++j)
						fprintf(file, "  ");
					currlinelen = depth*2;
					if (ltype_depth > depth) {
						for(j=ltype_depth; j>depth; --j)
							fprintf(file, "  ");
						currlinelen += (ltype_depth-depth)*2;
					}
					fprintf(file,"| ");
				}
			}
			font2 = par->GetFontSettings(i);
			if (font1.latex() != font2.latex()) {
				if (font2.latex() == LyXFont::OFF)
					islatex = 0;
				else
					islatex = 1;
			}
			else {
				islatex = 0;
			}
			c = par->GetChar(i);
			if (islatex)
				continue;
			switch (c) {
			case LYX_META_INSET:
				if ((inset = par->GetInset(i))) {
					fpos = ftell(file);
					inset->Latex(file,-1);
					currlinelen += (ftell(file) - fpos);
					actpos += (ftell(file) - fpos) - 1;
				}
				break;
			case LYX_META_NEWLINE:
				if (par->table) {
					if (par->table->NumberOfCellsInRow(actcell) <= cell) {
						for(j=actpos;j<clen[cell-1];++j)
							fputc(' ',file);
						fprintf(file," |\n");
						for(j=0; j<depth; ++j)
							fprintf(file, "  ");
						currlinelen = depth*2;
						if (ltype_depth > depth) {
							for(j=ltype_depth; j>depth; --j)
								fprintf(file, "  ");
							currlinelen += (ltype_depth-depth)*2;
						}
						for(j=0;j<cells;++j) {
							fputc('+',file);
							for(h=0; h<(clen[j]+1); ++h)
								fputc('-',file);
						}
						fprintf(file,"+\n");
						for(j=0; j<depth; ++j)
							fprintf(file, "  ");
						currlinelen = depth*2;
						if (ltype_depth > depth) {
							for(j=ltype_depth; j>depth; --j)
								fprintf(file, "  ");
							currlinelen += (ltype_depth-depth)*2;
						}
						fprintf(file,"| ");
						cell = 1;
					} else {
						for(j=actpos; j<clen[cell-1]; ++j)
							fputc(' ',file);
						fprintf(file, " | ");
						++cell;
					}
                                        ++actcell;
					currlinelen = actpos = 0;
				} else {
					fprintf(file, "\n");
					for(j=0; j<depth; ++j)
						fprintf(file, "  ");
					currlinelen = depth*2;
					if (ltype_depth > depth) {
						for(j=ltype_depth; j>depth; --j)
							fprintf(file, "  ");
						currlinelen += (ltype_depth-depth)*2;
					}
				}
				break;
			case '\\': 
				fprintf(file, "\\");
				break;
			default:
				if (currlinelen > (linelen-10) \
                                    && c==' ' && (i+2 < par->last)) {
					fprintf(file, "\n");
					for(j=0; j<depth; ++j)
						fprintf(file, "  ");
					currlinelen = depth*2;
					if (ltype_depth > depth) {
						for(j=ltype_depth; j>depth; --j)
							fprintf(file, "  ");
						currlinelen += (ltype_depth-depth)*2;
					}
				} else if (c != '\0')
					fprintf(file, "%c", c);
				else if (c == '\0')
					lyxerr[Debug::INFO] 
						<< "writeAsciiFile: 0 "
						"char in structure.\n";
				++currlinelen;
				break;
			}
		}
		if (par->table) {
			for(j=actpos;j<clen[cell-1];++j)
				fputc(' ',file);
			fprintf(file," |\n");
			for(j=0; j<depth; ++j)
				fprintf(file, "  ");
			currlinelen = depth*2;
			if (ltype_depth > depth) {
				for(j=ltype_depth; j>depth; --j)
					fprintf(file, "  ");
				currlinelen += (ltype_depth-depth)*2;
			}
			for(j=0;j<cells;++j) {
				fputc('+',file);
				for(h=0; h<(clen[j]+1); ++h)
					fputc('-',file);
			}
			fprintf(file,"+\n");
			delete [] clen;    
		}      
		par = par->next;
	}
   
	fprintf(file, "\n");
#endif
}


void Buffer::makeLaTeXFile(string const & filename, 
			   string const & original_path,
			   bool tmprun, bool only_body)
{
	lyxerr[Debug::LATEX] << "makeLaTeXFile...\n";
	tex_code_break_column = lyxrc->ascii_linelen;

        LyXTextClass const *tclass = textclasslist.textClass(params.textclass);
  
	FilePtr file(filename, FilePtr::write);
	if (!file()) {
		WriteFSAlert(_("Error: Cannot write file:"),filename);
		return;
	}

	// will spawn a recursive decent of the complete buffer
	// structure.
	lyxerr[Debug::LATEX] << "  Validating buffer...\n";
	LaTeXFeatures features(tclass->numLayouts());
	validate(features);
	lyxerr[Debug::LATEX] << "  Buffer validation done.\n";
	
	texrow.reset();
	// The starting paragraph of the coming rows is the 
	// first paragraph of the document. (Asger)
	texrow.start(paragraph, 0);

	string userName(getUserName());
	string LFile;
	
	if (!only_body && !tmprun) {
		LFile += "%% This LaTeX-file was created by <";
		LFile += userName + "> " + (char*)date();
		LFile += "%% LyX 1.1 (C) 1995-1999 by Matthias Ettrich and the LyX Team\n";
		LFile += "\n%% Do not edit this file unless you know what you are doing.\n";
		texrow.newline();
		texrow.newline();
		texrow.newline();
		texrow.newline();
	}
	lyxerr[Debug::INFO] << "lyx header finished\n";
	// There are a few differences between nice LaTeX and usual files:
	// usual is \batchmode, uses \listfiles and has a 
	// special input@path to allow the including of figures
	// with either \input or \includegraphics (what figinsets do).
	// batchmode is not set if there is a tex_code_break_column.
	// In this case somebody is interested in the generated LaTeX,
	// so this is OK. input@path is set when the actual parameter
	// original_path is set. This is done for usual tex-file, but not
	// for nice-latex-file. (Matthias 250696)
	if (!only_body) {
		if (tmprun){
			// code for usual, NOT exported latex-file
			LFile += "\\batchmode\n";
			texrow.newline();
			// the latex dependency stuff needs this.
			LFile += "\\listfiles\n";
			texrow.newline();
		}
		if (!original_path.empty()) {
			LFile += "\\makeatletter\n";
			texrow.newline();
			LFile += "\\def\\input@path{{" + original_path
				 + "/}}\n";
			texrow.newline();
			LFile += "\\makeatother\n";
			texrow.newline();
		}
		
		LFile += "\\documentclass";
		
		string options; // the document class options.
		
		if (tokenPos(tclass->opt_fontsize, '|',params.fontsize) >= 0) {
			// only write if existing in list (and not default)
			options += params.fontsize;
			options += "pt,";
		}
		
		
		if (!params.use_geometry &&
		    (params.paperpackage == PACKAGE_NONE)) {
			switch (params.papersize) {
			case PAPER_A4PAPER:
				options += "a4paper,";
				break;
			case PAPER_USLETTER:
				options += "letterpaper,";
				break;
			case PAPER_A5PAPER:
				options += "a5paper,";
				break;
			case PAPER_B5PAPER:
				options += "b5paper,";
				break;
			case PAPER_EXECUTIVEPAPER:
				options += "executivepaper,";
				break;
			case PAPER_LEGALPAPER:
				options += "legalpaper,";
				break;
			}
		}

		// if needed
		if (params.sides != tclass->sides) {
			if (params.sides == 2)
				options += "twoside,";
			else
				options += "oneside,";
		}

		// if needed
		if (params.columns != tclass->columns) {
			if (params.columns == 2)
				options += "twocolumn,";
			else
				options += "onecolumn,";
		}

		if (!params.use_geometry && params.orientation == ORIENTATION_LANDSCAPE)
			options += "landscape,";
		
		// language should be a parameter to \documentclass
		if (params.language != "default") {
			options += params.language + ',';
		}
		
		// the user-defined options
		if (!params.options.empty()) {
			options += params.options + ',';
		}
		
		if (!options.empty()){
			options = strip(options, ',');
			LFile += '[';
			LFile += options;
			LFile += ']';
		}
		
		LFile += '{';
		LFile += textclasslist.latexNameOfClass(params.textclass);
		LFile += "}\n";
		texrow.newline();
		// end of \documentclass defs
		
		// this one is not per buffer
		if (lyxrc->fontenc != "default") {
			LFile += "\\usepackage[" + lyxrc->fontenc
				 + "]{fontenc}\n";
			texrow.newline();
		}
		if (params.inputenc != "default") {
			LFile += "\\usepackage[" + params.inputenc
				 + "]{inputenc}\n";
			texrow.newline();
		}
		if (params.fonts != "default") {
			LFile += "\\usepackage{" + params.fonts + "}\n";
			texrow.newline();
		}
		
		/* at the very beginning the text parameters */
		if (params.paperpackage != PACKAGE_NONE) {
			switch (params.paperpackage) {
			case PACKAGE_A4:
				LFile += "\\usepackage{a4}\n";
				texrow.newline();
				break;
			case PACKAGE_A4WIDE:
				LFile += "\\usepackage{a4wide}\n";
				texrow.newline();
				break;
			case PACKAGE_WIDEMARGINSA4:
				LFile += "\\usepackage[widemargins]{a4}\n";
				texrow.newline();
				break;
			}
		}
		if (params.use_geometry) {
			LFile += "\\usepackage{geometry}\n";
			texrow.newline();
			LFile += "\\geometry{verbose";
			if (params.orientation == ORIENTATION_LANDSCAPE)
				LFile += ",landscape";
			switch (params.papersize2) {
			case VM_PAPER_CUSTOM:
				if (!params.paperwidth.empty())
					LFile += ",paperwidth="
						 + params.paperwidth;
				if (!params.paperheight.empty())
					LFile += ",paperheight="
						 + params.paperheight;
				break;
			case VM_PAPER_USLETTER:
				LFile += ",letterpaper";
				break;
			case VM_PAPER_USLEGAL:
				LFile += ",legalpaper";
				break;
			case VM_PAPER_USEXECUTIVE:
				LFile += ",executivepaper";
				break;
			case VM_PAPER_A3:
				LFile += ",a3paper";
				break;
			case VM_PAPER_A4:
				LFile += ",a4paper";
				break;
			case VM_PAPER_A5:
				LFile += ",a5paper";
				break;
			case VM_PAPER_B3:
				LFile += ",b3paper";
				break;
			case VM_PAPER_B4:
				LFile += ",b4paper";
				break;
			case VM_PAPER_B5:
				LFile += ",b5paper";
				break;
			}
			if (!params.topmargin.empty())
				LFile += ",tmargin=" + params.topmargin;
			if (!params.bottommargin.empty())
				LFile += ",bmargin=" + params.bottommargin;
			if (!params.leftmargin.empty())
				LFile += ",lmargin=" + params.leftmargin;
			if (!params.rightmargin.empty())
				LFile += ",rmargin=" + params.rightmargin;
			if (!params.headheight.empty())
				LFile += ",headheight=" + params.headheight;
			if (!params.headsep.empty())
				LFile += ",headsep=" + params.headsep;
			if (!params.footskip.empty())
				LFile += ",footskip=" + params.footskip;
			LFile += "}\n";
			texrow.newline();
		}
		if (params.use_amsmath
		    && !prefixIs(textclasslist.latexNameOfClass(params.textclass), "ams")) {
			LFile += "\\usepackage{amsmath}\n";
		}

		if (tokenPos(tclass->opt_pagestyle, '|',params.pagestyle) >= 0) {
			if (params.pagestyle == "fancy") {
				LFile += "\\usepackage{fancyhdr}\n";
				texrow.newline();
			}
			LFile += "\\pagestyle{" + params.pagestyle + "}\n";
			texrow.newline();
		}

		// We try to load babel late, in case it interferes
		// with other packages.
		if (params.language != "default") {
			LFile += "\\usepackage{babel}\n";
			texrow.newline();
		}

		if (params.secnumdepth != tclass->secnumdepth) {
			LFile += "\\setcounter{secnumdepth}{";
			LFile += tostr(params.secnumdepth);
			LFile += "}\n";
			texrow.newline();
		}
		if (params.tocdepth != tclass->tocdepth) {
			LFile += "\\setcounter{tocdepth}{";
			LFile += tostr(params.tocdepth);
			LFile += "}\n";
			texrow.newline();
		}
		
		if (params.paragraph_separation) {
			switch (params.defskip.kind()) {
			case VSpace::SMALLSKIP: 
				LFile += "\\setlength\\parskip{\\smallskipamount}\n";
				break;
			case VSpace::MEDSKIP:
				LFile += "\\setlength\\parskip{\\medskipamount}\n";
				break;
			case VSpace::BIGSKIP:
				LFile += "\\setlength\\parskip{\\bigskipamount}\n";
				break;
			case VSpace::LENGTH:
				LFile += "\\setlength\\parskip{"
					 + params.defskip.length().asLatexString()
					 + "}\n";
				break;
			default: // should never happen // Then delete it.
				LFile += "\\setlength\\parskip{\\medskipamount}\n";
				break;
			}
			texrow.newline();
			
			LFile += "\\setlength\\parindent{0pt}\n";
			texrow.newline();
		}

		// Write out what we've generated so far...and reset LFile
		fwrite(LFile.c_str(), sizeof(char), LFile.length(), file);
		LFile.erase(); 

		// Now insert the LyX specific LaTeX commands...
		string preamble, tmppreamble;

		// The optional packages;
		preamble = features.getPackages(params);

		// this might be useful...
		preamble += "\n\\makeatletter\n\n";

		// Some macros LyX will need
		tmppreamble = features.getMacros(params);

		if (!tmppreamble.empty()) {
			preamble += "\n%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% "
				"LyX specific LaTeX commands.\n"
				+ tmppreamble + '\n';
		}

		// the text class specific preamble 
		tmppreamble = features.getTClassPreamble(params);
		if (!tmppreamble.empty()) {
			preamble += "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% "
				"Textclass specific LaTeX commands.\n"
				+ tmppreamble + '\n';
		}

		/* the user-defined preamble */
		if (!params.preamble.empty()) {
			preamble += "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% "
				"User specified LaTeX commands.\n"
				+ params.preamble + '\n';
		}

		preamble += "\\makeatother\n\n";

		// Itemize bullet settings need to be last in case the user
		// defines their own bullets that use a package included
		// in the user-defined preamble -- ARRae
		for (int i = 0; i < 4; ++i) {
			if (params.user_defined_bullets[i] != ITEMIZE_DEFAULTS[i]) {
				preamble += "\\renewcommand\\labelitemi";
				switch (i) {
					// `i' is one less than the item to modify
				case 0:
					break;
				case 1:
					preamble += 'i';
					break;
				case 2:
					preamble += "ii";
					break;
				case 3:
					preamble += 'v';
					break;
				}
				preamble += "[0]{" + params.user_defined_bullets[i].getText() + "}\n";
			}
		}

		for (int j = countChar(preamble, '\n'); --j ;) {
			texrow.newline();
		}

		// A bit faster than printing a char at a time I think.
		fwrite(preamble.c_str(), sizeof(char),
		       preamble.length(), file);

		// make the body.
		LFile += "\\begin{document}\n\n";
		texrow.newline();
		texrow.newline();
	} // only_body
	lyxerr[Debug::INFO] << "preamble finished, now the body.\n";
	
	bool was_title = false;
	bool already_title = false;
	string ftnote;
	TexRow ft_texrow;
	int ftcount = 0;
	//int loop_count = 0;

	LyXParagraph *par = paragraph;

	// if only_body
	while (par) {
		//++loop_count;
		if (par->IsDummy())
			lyxerr[Debug::LATEX] << "Error in MakeLateXFile.\n";
		LyXLayout const * layout = textclasslist.style(params.textclass,
						    par->layout);
	    
	        if (layout->intitle) {
			if (already_title) {
				lyxerr << "Error in MakeLatexFile: You"
					" should not mix title layouts"
					" with normal ones.\n";
			} else
				was_title = true;
	        } else if (was_title && !already_title) {
			LFile += "\\maketitle\n";
			texrow.newline();
			already_title = true;
			was_title = false;		    
		}
		if (layout->isEnvironment()
                    || par->pextra_type != PEXTRA_NONE) {
			par = par->TeXEnvironment(LFile, texrow,
						  ftnote, ft_texrow, ftcount);
		} else {
			// We are at depth 0 so we can just use
			// ordinary \footnote{} generation
			// flag this with ftcount
			ftcount = -1;
			par = par->TeXOnePar(LFile, texrow,
					     ftnote, ft_texrow, ftcount);
			ftcount = 0;
		}

		// Write out what we've generated...and reset LFile
		if (ftcount) {
			if (ftcount > 1) {
				LFile += "\\addtocounter{footnote}{-";
				LFile += ftcount - 1;
				LFile += '}';
			}
			LFile += ftnote;
			texrow += ft_texrow;
			ftnote.erase();
			ft_texrow.reset();
			ftcount = 0;
		}
		// removed. we can insert clever optimizations later
		// now we need code that is easy to read.
		//if (loop_count == 2) {
		// fwrite()ing every second time through the loop
		// gains a few extra % of speed; going higher than
		// 2 will slow things down again.  I'll look at
		// LFile.length() in a future revision.  ARRae
		fwrite(LFile.c_str(), sizeof(char), LFile.length(), file);
		LFile.erase();
		//loop_count = 0;
		//}
	}

	if (!only_body) {
		LFile += "\\end{document}\n";
		texrow.newline();
	
		lyxerr[Debug::LATEX] << "makeLaTeXFile...done\n";
	} else {
		lyxerr[Debug::LATEX] << "LaTeXFile for inclusion made.\n";
	}

	// Just to be sure. (Asger)
	texrow.newline();

	// Write out what we've generated...and reset LFile
	fwrite(LFile.c_str(), sizeof(char), LFile.length(), file);
	//LFile.erase(); no need nothing more happens here

	if (file.close()) {
		WriteFSAlert(_("Error! Could not close file properly:"),
			     filename);
	}
	lyxerr[Debug::INFO] << "Finished making latex file.\n";
}


/* Temporary. Sorry. */
// Why isn't these in class Buffer? (even if they are temporary?)
static int depth;              /* paragraph depth */
static int enum_flags[10];     /* enum/itemize status */
static int desc_on;            /* description mode on/off */
// not true desc_on has more values than on/off
static bool verb_on;            /* verbatim mode on/off */


void Buffer::linuxDocInitVars()
{
	depth = -1;
	enum_flags[0] = LINUXDOC_NONE;
	desc_on = 0;
}

void Buffer::linuxDocHandleFootnote(LyXParagraph* &par)
{
	while (par && par->footnoteflag != LyXParagraph::NO_FOOTNOTE) {
		/* no footnote support yet */
		par = par->next;
	}
}

void Buffer::linuxDocStartPar(FILE *file, LyXParagraph* &par)
	/* before article body : limited tag set */
{
	int     error = 0;                /* error flag */
	string    error_message;        /* error message */
	LyXFont font;                     /* font information */
        static int oldtag = -1;

	if (!par) return;
	if (par->IsDummy()){
		error = 1;
		lyxerr[Debug::INFO] << "Error in LinuxDocStartPar\n";
	}
	par->AutoDeleteInsets();

	int tag = LinuxDocTag(textclasslist.nameOfLayout(params.textclass,
						    par->GetLayout()));

	LyXLayout const *style=textclasslist.style(params.textclass, par->layout);

	// special handling for <toc> tag (\tableofcontents TeX macro)
	if (par->GetChar(0) == LYX_META_INSET) {
		Inset *inset = par->GetInset(0);
		char   lyx_code = inset->LyxCode();
		switch (lyx_code) {
		case Inset::TOC_CODE:
			fprintf(file, "\n<toc>");
			error = 1;     // fake : no futher processing required
			break;
		case Inset::HTMLURL_CODE: // Needed only to get <author> in sgml-file...
			if (tag == LINUXDOC_AUTHOR) {
				if (oldtag != tag)
					fprintf(file, "\n<%s>",
						style->latexname.c_str());
//			    else fprintf(file, "\n<and>");
				else fprintf(file, "\n");
			}
			break;
		default:
			break;
		}
	} else {
		switch(tag) {
		case LINUXDOC_SECT:
			depth = 0;
		case LINUXDOC_TITLE:
		case LINUXDOC_AUTHOR:
		case LINUXDOC_DATE:
		case LINUXDOC_ABSTRACT:
			fprintf(file, "\n");
			if ((tag !=LINUXDOC_AUTHOR) || 
			    (oldtag != tag))  // two author-lines are not allowed in LinuxDoc...
				fprintf(file, "<%s>", style->latexname.c_str());
//				else // (tag == LINUXDOC_AUTHOR) && (oldtag == tag)
//                                   fprintf(file, "<and>");
			break;
		default:
			error_message = _("Style ") + style->latexname + _(" not allowed before first Section."),
			LinuxDocError(par, 0, error_message);
			error = 1;
			break;
		}
	}
	if (!error) {
		SimpleLinuxDocOnePar(file, font, par, verb_on, desc_on);
		if (tag == LINUXDOC_SECT)
			fprintf(file, "\n<p>");
	}
	par = par->next;
	oldtag = tag;
	linuxDocHandleFootnote(par);
}


/* handle the layout constraints for a given paragraph */
void Buffer::linuxDocOnePar(FILE *file, LyXParagraph* &par) 
{

	LyXFont font;
	int i = 0;

	int error;                     /* error flag */
	string error_message;        /* error message */
	int tag;                       /* number of the layout tag */
        static string last_paragraph[10]; /* contains end of paragraph sign */
        string special_tag_string, special_tag_close; /* special tags */

	/* standard initialisation */
	verb_on = 0;
	error = 0;
	LyXLayout const *style=0;
	if (par) {
		if (par->IsDummy()){
			error = 1;
			lyxerr[Debug::INFO] << "Error in LinuxDocOnePar\n";
		}
		par->AutoDeleteInsets();
		
		tag = LinuxDocTag(textclasslist.
				  nameOfLayout(params.textclass,
					       par->GetLayout()));
	        style = textclasslist.style(params.textclass, par->layout);
	} else{
		if (depth == -1)
			return; /* No enumerate/itemize/descrip open */
		tag = -1; /* Undefined Token! */
	}


	/* handle enumerate, itemize and quote SGML tags closing */
	switch(enum_flags[depth]) {
	case LINUXDOC_ENUM:
	case LINUXDOC_ITEM:
		if (((tag != enum_flags[depth]) && (tag!=LINUXDOC_VERB) 
		     && (tag!=LINUXDOC_CODE)) ||
		    (par && (depth > par->depth))) {
			bool flag;
			do {
                                flag = false;
				if (!par || depth >=par->depth)
					fprintf(file, "\n</%s>",
						last_paragraph[depth].c_str());
				if (depth == 0)
					enum_flags[depth] = LINUXDOC_NONE;
				if (par && (depth>par->depth)) 
					{ depth --; 
					if ((tag != LINUXDOC_ENUM) &&
					    (tag != LINUXDOC_ITEM)   ) 
						flag = true;
					else flag = (depth > par->depth); 
					}
			} while (flag);
		}
		break;
	default:
		if ((tag!=LINUXDOC_ITEM) && (tag!=LINUXDOC_ENUM) 
		    && (par && (depth < par->depth))) {
			special_tag_string = "<quote>";
                        fprintf(file, "\n%s", special_tag_string.c_str());
			if (tag!=LINUXDOC_DESCR)
				special_tag_close  = "</quote>";
		}
		break;
	}

	/* handle descriptions closing */
	if ((desc_on) &&
	    ((tag != LINUXDOC_DESCR) || (!par) || (depth>par->depth))) {
                int goal_depth = (par? par->depth : 0);
		while (depth>goal_depth) 
			{ fprintf(file, "\n</descrip>");
			if (depth != 0) fprintf(file, "\n</quote>");
			--depth; 
			}
                if (tag!=LINUXDOC_DESCR) fprintf(file, "\n</descrip>");
		if ((par) && (par->depth==0) && (tag!=LINUXDOC_DESCR))
			desc_on = 0;
	}

	if (!par) return; /* End of document! */

	/* write opening SGML tags */
	switch(tag) {
	case LINUXDOC_TITLE:
	case LINUXDOC_AUTHOR:
	case LINUXDOC_DATE:
	case LINUXDOC_ABSTRACT:
		error_message = string(_("Style ")) + linuxdoc_styles[tag] + string(_(" not allowed after first Section."));
		LinuxDocError(par ,i, error_message);
		error = 1;
		break;
	case LINUXDOC_PAR:
		if (special_tag_string.empty()) fprintf(file, "\n\n");
		else fprintf(file, "%s", special_tag_string.c_str());
		break;
	case LINUXDOC_VERB:
		font = par->getFont(0);
		verb_on = 1;
		/* special type handling for tscreen */
		if ((font.family() == LyXFont::TYPEWRITER_FAMILY)
		    && (depth < par->depth)) {
			special_tag_string = "<tscreen>";
			special_tag_close  = "</tscreen>";
		}
		fprintf(file, "%s\n<verb>", special_tag_string.c_str());
		break;
	case LINUXDOC_CODE:
		fprintf(file, "%s\n<code>", special_tag_string.c_str());
		break;
	case LINUXDOC_SECT:
	case LINUXDOC_SECT1:
	case LINUXDOC_SECT2:
	case LINUXDOC_SECT3:
	case LINUXDOC_SECT4:
		if (depth!=0) {
			for (i=depth; i>=0; --i) {
				if (enum_flags[depth]!=LINUXDOC_NONE)
					fprintf(file, "\n</%s>",
						linuxdoc_tags[enum_flags[depth]]);
				enum_flags[i] = LINUXDOC_NONE;
			}
		}
		fprintf(file, "%s", special_tag_string.c_str());
		fprintf(file, "\n<%s>", style->latexname.c_str());
		break;
	case LINUXDOC_ENUM:
	case LINUXDOC_ITEM:
		if (depth < par->depth) {
			depth = par->depth;
			enum_flags[depth] = LINUXDOC_NONE;
		}
		if (enum_flags[depth] != tag) {
			last_paragraph[depth]= style->latexname;
			fprintf(file, "\n<%s>", last_paragraph[depth].c_str());
			enum_flags[depth] = tag;
		}
		fprintf(file, "\n<item>");
		depth = par->depth;
		break;
	case LINUXDOC_DESCR:
		if (!desc_on || depth<par->depth) {
			depth=par->depth;
			enum_flags[depth] = LINUXDOC_DESCR;
			fprintf(file, "\n<%s>", style->latexname.c_str());
		}
                else if (depth>par->depth) {
			fprintf(file, "\n<%s>",
				style->latexname.c_str());
			depth = par->depth;
		}
		fprintf(file, "\n<tag>");
		desc_on = 1;
		break;
	default:
		fprintf(file, "\n<%s>", style->latexname.c_str());
		break;
	} // switch(tag)

	if (!error) {
		SimpleLinuxDocOnePar(file, font, par, verb_on, desc_on);
    
		/* write closing SGML tags */
		switch(tag) {
		case LINUXDOC_ENUM:
		case LINUXDOC_ITEM:
		case LINUXDOC_DESCR:
		case LINUXDOC_PAR:
		case LINUXDOC_DATE:
		case LINUXDOC_AUTHOR:
		case LINUXDOC_TITLE:
			break;
		case LINUXDOC_SECT:
		case LINUXDOC_SECT1:
		case LINUXDOC_SECT2:
		case LINUXDOC_SECT3:
		case LINUXDOC_SECT4:
			fprintf(file, "\n<p>");
			break;
		default:
			fprintf(file, "\n</%s>", style->latexname.c_str());
			break;
		}
                if (!special_tag_close.empty())
			fprintf(file, "%s", special_tag_close.c_str());
	}

	depth = par->depth;
	par = par->next;

	linuxDocHandleFootnote(par);
}


void Buffer::makeLinuxDocFile(string const & filename, int column)
{
	LyXParagraph *par = paragraph;

	FilePtr file(filename, FilePtr::write);
	tex_code_break_column = column; 

	if (!file()) {
		WriteAlert(_("LYX_ERROR:"), _("Cannot write file"), filename);
		return;
	}
   
	texrow.reset();

        string userName(getUserName());
	fprintf(file,
		"<!-- LinuxDoc file was created by LyX 1.1 (C) 1995-1998 ");
	fprintf(file, "by <%s> %s -->\n", userName.c_str(), (char *)date());
	fprintf(file,
		"<!-- Export filter v0.6 by Pascal Andre/Bernhard Iselborn-->\n");
   
	if (params.preamble.empty()) {
		fprintf(file, "\n<!doctype linuxdoc system>\n\n");
	}
	else {
		fprintf(file, "\n<!doctype linuxdoc system \n [ %s \n]>\n\n", 
			params.preamble.c_str() );
	}
	fprintf(file, "<article>\n");   

        linuxDocInitVars();

	while (par) {
		if (par->IsDummy()) {
			lyxerr[Debug::INFO] << "Error in MakeLinuxDocFile.\n";
			par = par->next;
                } else if (depth == -1)
                        linuxDocStartPar(file, par);
                else
			linuxDocOnePar(file, par);
	}
   
	linuxDocOnePar(file, par);
	fprintf(file, "\n\n</article>\n");

	if (file.close()) {
		WriteFSAlert(_("Error! Could not close file properly:"),
			     filename);
	}
}


/* gets tag number from layout name */
int Buffer::LinuxDocTag(string const & layout)
{
	for (int i = 0; i < LINUXDOC_NONE; ++i)
		if (linuxdoc_styles[i] == layout) return i;

	return LINUXDOC_NONE;
}


/* push a tag in a style stack */
void Buffer::push_tag(FILE *file, char const *tag,
		      int& pos, char stack[5][3])
{
	int j;

	/* pop all previous tags */
	for (j = pos; j >= 0; --j)
		fprintf(file, "</%s>", stack[j]); 

	/* add new tag */
	sprintf(stack[++pos], "%s", tag);

	/* push all tags */
	for (j = 0; j <= pos; ++j)
		fprintf(file, "<%s>", stack[j]);
}


// pop a tag from a style stack
void Buffer::pop_tag(FILE *file, char const * tag,
		     int& pos, char stack[5][3])
{
	int j;

	// pop all tags till specified one
	for (j = pos; j >= 0 && compare(stack[j], tag); --j)
		fprintf(file, "</%s>", stack[j]);

	// closes the tag
	fprintf(file, "</%s>", tag);

	// push all tags, but the specified one
	for (j = j + 1; j <= pos; ++j) {
		fprintf(file, "<%s>", stack[j]);
		strcpy(stack[j - 1], stack[j]);
	}
	--pos;
}


// gets next latex parameter
// note: the convert-flag was introduced for work-arounds of the
// SGML-tool backend. But it had not the desired effekt.
// So it's "for further usage" :-)
int Buffer::GetNextParameter(LyXParagraph *par, int pos, string &param, bool convert)
{
	int i, flag;
	char c;
	Inset *inset;

	flag = 1;
        param.erase();
	for (i=pos+1; i<=par->last() && flag; ++i) {
		c = par->GetChar(i);
		if (c == LYX_META_INSET) {
			inset = par->GetInset(i);
			switch (inset->LyxCode()) {
			case Inset::QUOTE_CODE:
				c='\"';
				break;
			case Inset::SEPARATOR_CODE:
			case Inset::ENDING_CODE:
				flag = 0;
				break;
			case Inset::TOC_CODE:
				LinuxDocError(par, i, _("Error : table of contents not allowed "
							"in body.\n"));
				break;
			case Inset::IGNORE_CODE:
				// ignore
				break;
			default:
				LinuxDocError(par, i, _("Error : unhandled inset.\n"));
				break;
			}
		}
		if (flag) {
			string temp;
			if (convert) linuxDocConvertChar(c, temp);
			else         temp=c;
			param += temp;
		}
	}

	if (flag) {
		LinuxDocError(par, pos, _("Error : incomplete macro.\n"));
		return 0;
	}
	return i;
}


bool Buffer::removeAutoInsets()
{
	LyXParagraph *par = paragraph;

	bool a = false;
	while (par) {
		if (par->AutoDeleteInsets()){
			a = true;
		}
		par = par->next;
	}
	return a;
}


int Buffer::runLaTeX()
{
#warning this check is a bit wrong
	// This test if we have a typeset version of the document, but
	// is actually enough for us if we have data in the paragraphs
	if (!users->text) return 0;

	ProhibitInput();

	// get LaTeX-Filename
	string name = SpaceLess(ChangeExtension (filename, ".tex", true));

	string path = OnlyPath(filename);

	string org_path = path;
	if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)) {
		path = tmppath;	 
	}

	Path p(path);
	users->getOwner()->getMiniBuffer()->Set(_("Running LaTeX..."));   

	// Remove all error insets
	bool a = removeAutoInsets();

	// generate the LaTeX file always
	makeLaTeXFile(name, org_path, true);
	markDviDirty();

	TeXErrors terr;
	LaTeX latex(lyxrc->latex_command, name, filepath);
	int res = latex.run(terr, users->getOwner()->getMiniBuffer());

	// check return value from latex.run().
	if ((res & LaTeX::NO_LOGFILE)) {
		WriteAlert(_("LaTeX did not work!"),
			   _("Missing log file:"), name);
	} else if ((res & LaTeX::ERRORS)) {
		users->getOwner()->getMiniBuffer()->Set(_("Done"));
		// Insert all errors as errors boxes
		insertErrors(terr);
		
		// Dvi should also be kept dirty if the latex run
		// ends up with errors. However it should be possible
		// to view a dirty dvi too.
	} else {
		//no errors or any other things to think about so:
		users->getOwner()->getMiniBuffer()->Set(_("Done"));
		markDviClean();
	}

	// if we removed error insets before we ran LaTeX or if we inserted
	// error insets after we ran LaTeX this must be run:
	if (a || (res & LaTeX::ERRORS)){
		users->redraw();
		users->fitCursor();
		users->updateScrollbar();
	}
	AllowInput();

	return latex.getNumErrors();
}


// This should be enabled when the Chktex class is implemented. (Asger)
// chktex should be run with these flags disabled: 3, 22, 25, 30, 38(?)
// Other flags: -wall -v0 -x
int Buffer::runChktex()
{
	// ditto
	if (!users->text) return 0;

	ProhibitInput();

	// get LaTeX-Filename
	string name = SpaceLess(ChangeExtension (filename, ".tex", true));
	string path = OnlyPath(filename);

	string org_path = path;
	if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)) {
		path = tmppath;	 
	}

	Path p(path);
	users->getOwner()->getMiniBuffer()->Set(_("Running chktex..."));

	// Remove all error insets
	bool a = removeAutoInsets();

	// Generate the LaTeX file if neccessary
	if (!isDviClean() || a) {
		makeLaTeXFile(name, org_path, true);
		markDviDirty();
	}

	TeXErrors terr;
	Chktex chktex(lyxrc->chktex_command, name, filepath);
	int res = chktex.run(terr); // run chktex

	if (res == -1) {
		WriteAlert(_("chktex did not work!"),
			   _("Could not run with file:"), name);
	} else if (res > 0) {
		// Insert all errors as errors boxes
		insertErrors(terr);
	}

	// if we removed error insets before we ran chktex or if we inserted
	// error insets after we ran chktex, this must be run:
	if (a || res){
		users->redraw();
		users->fitCursor();
		users->updateScrollbar();
	}
	AllowInput();

	return res;
}


void Buffer::insertErrors(TeXErrors &terr)
{
	// Save the cursor position
	LyXCursor cursor = users->text->cursor;

	// Now traverse all the errors and insert them
	bool firsterror = true;
	bool more = true;

	// This is drastic, but it's the only fix, I could find. (Asger)
	users->allFloats(BufferView::OpenFootMargin);
	users->allFloats(BufferView::OpenFigTab);

	while (more) {
		string errortext;
		int errorrow = 0;

		if (firsterror) {
			more = terr.getFirstError(&errorrow, &errortext);
			firsterror = false;
		} else {
			more = terr.getNextError(&errorrow, &errortext);
		}

		if (!more)
			break;

		// Insert error string for row number
		int tmpid = -1; 
		int tmppos = -1;

		texrow.getIdFromRow(errorrow, tmpid, tmppos);

		LyXParagraph* texrowpar;

		if (tmpid == -1) {
			texrowpar = users->text->FirstParagraph();
			tmppos = 0;
		} else {
			texrowpar = users->text->GetParFromID(tmpid);
		}

		if (texrowpar == 0)
			continue;

		InsetError *new_inset = new InsetError(errortext);

		users->text->SetCursorIntern(texrowpar, tmppos);
		users->text->InsertInset(new_inset);
		users->text->FullRebreak();
	}
	// Restore the cursor position
	users->text->SetCursorIntern(cursor.par, cursor.pos);
}


/* handle internal paragraph parsing -- layout already processed */

// checks, if newcol chars should be put into this line
// writes newline, if necessary.
static
void linux_doc_line_break(FILE *file, unsigned int &colcount, const unsigned int newcol)
{
	colcount += newcol;
	if (colcount > lyxrc->ascii_linelen) {
		fprintf(file, "\n");
		colcount = newcol; // assume write after this call
	}
}


void Buffer::SimpleLinuxDocOnePar(FILE *file, LyXFont &font1, 
				  LyXParagraph *par, bool &verb_on, 
				  int &desc_on) 
{
	LyXFont font2;
	char c;
	Inset *inset;
	int main_body, j;
	LyXLayout const * style = textclasslist.style(params.textclass, par->GetLayout());

	bool is_tt = false;               // typewriter font flag 
	bool is_bf = false;               // bold font flag 
	bool is_em = false;               // emphasis (italic) font flag 
	bool is_quote = false;            // quote (paragraph indenting) flag 

	int stack_num = -1;          // style stack position 
	char stack[5][3];    	     // style stack 
        unsigned int char_line_count = 5;     // Heuristic choice ;-) 

	if (style->labeltype != LABEL_MANUAL)
		main_body = 0;
	else
		main_body = par->BeginningOfMainBody();

	/* gets paragraph main font */
	if (main_body > 0)
		font1 = style->labelfont;
	else
		font1 = style->font;

  
	/* parsing main loop */
	for (int i = 0; i < par->last(); ++i) {

		/* handle quote tag */
		if (i == main_body && !par->IsDummy()) {
			if (main_body > 0)
				font1 = style->font;
			if ((!verb_on) && (par->pparams.align == LYX_ALIGN_RIGHT)) {
				linux_doc_line_break(file, char_line_count, 7);
				fprintf(file, "<quote>");
				is_quote = true;
			}
		}

		font2 = par->getFont(i);

		/* handle <tt> tag */
		if (font1.family() != font2.family()) {
			if (verb_on)
				LinuxDocError(par, 0,
					      "Error : TypeWriter font not"
					      " allowed in Verbatim.\n");
			else {
				if (font2.family() == LyXFont::TYPEWRITER_FAMILY) {
					push_tag(file, "tt", stack_num, stack);
					is_tt = true;
				} else if (is_tt) {
					pop_tag(file, "tt", stack_num, stack);
					is_tt = false;
				}
			}
		}

		/* handle <bf> tag */
		if (font1.series() != font2.series()) {
			if (verb_on)
				LinuxDocError(par, 0,
					      "Error : Bold font not allow"
					      " in Verbatim.\n");
			else {
				if (font2.series() == LyXFont::BOLD_SERIES) {
					push_tag(file, "bf", stack_num, stack);
					is_bf = true;
				}
				else if (is_bf) {
					pop_tag(file, "bf", stack_num, stack);
					is_bf = false;
				}
			}
		}

		/* handle <em> tag */
		if ((!verb_on) && (font1.shape() != font2.shape())) {
			if (verb_on)
				LinuxDocError(par, 0,
					      "Error : Emphasis (italic font)"
					      " not allowed in Verbatim.\n");
			else {
				if (font2.shape() == LyXFont::ITALIC_SHAPE) {
					push_tag(file, "em", stack_num, stack);
					is_em = true;
				} else if (is_em) {
					pop_tag(file, "em", stack_num, stack);
					is_em = false;
				}
			}
		}
      
		c = par->GetChar(i);
      
		if (font2.latex() == LyXFont::ON) {
			// "TeX"-Mode on ==> SGML-Mode on.
			if (c!='\0')
				fprintf(file, "%c", c); // see LaTeX-Generation...
			++char_line_count;
		} else if (c == LYX_META_INSET) {
			inset = par->GetInset(i);
			switch (inset->LyxCode()) {
			case Inset::TOC_CODE:
				LinuxDocError(par, i,
					      "Error: table of contents not"
					      " allowed in body.\n");
				break;
			case Inset::QUOTE_CODE:
				fprintf(file, "\"");
				++char_line_count; ++char_line_count;
				break;
			case Inset::MARK_CODE:
			{
				int j;
				string label;

                               
				if ((j = GetNextParameter(par, i, label, true))) {
					i = j; 
				        linux_doc_line_break(file,
							     char_line_count,
							     13+label.length());
					fprintf(file,
						"<label id=\"%s\">",
						label.c_str());
				}
				++i;
			}
			break;
			case Inset::LABEL_CODE:
			{
                                const char *out =
					((InsetLabel*) inset)->getLabel(1).c_str();
				linux_doc_line_break(file, char_line_count, 13+strlen(out));
				fprintf(file, "<label id=\"%s\">", out);
			}
			break;
			case Inset::REF_CODE:
			{
				int j;
				string label, name;

				// idea: If we have a LaTeX-reference,
				//       then the name of the label is
				//       If it's a SGML-version, then
				//       we have two positional parameters
				//       We got errors, if the names are
				//       empty!
				label = ((InsetCommand*)inset)->getContents();
				if (label.empty()) {
					if ((j = GetNextParameter(par, i, label, true)) && 
				    	    (j = GetNextParameter(par, j-1, name, true))) {
						linux_doc_line_break(file, char_line_count, 
							     20+label.length()+name.length());
						i = j;
					}
					if (j == 0)
					{ // We got an error inset. We have to skip it...
						++i;
					}
				} else
					name = label;
				if (!name.empty()) {
				 // name is empty implies, that we got
				 // an error above. So we can't write
				 // anything to the file
					fprintf(file,
						"<ref id=\"%s\" name=\"%s\">",
						label.c_str(), name.c_str());
				}
			}
			break;
			case Inset::URL_CODE:
			{
				int j;
				string url, name;

				if ((j = GetNextParameter(par, i, url, true)) && 
				    (j = GetNextParameter(par, j-1, name, true))) {
					linux_doc_line_break(file, char_line_count, 
							     20+url.length()+name.length());
					i = j;
					fprintf(file,
						"<url url=\"%s\" name=\"%s\">",
						url.c_str(), name.c_str());
				}
			}
			break;
			case Inset::HTMLURL_CODE:
			{
				int j;
				string url, name;

				if ((j = GetNextParameter(par, i, url, true)) && 
				    (j = GetNextParameter(par, j-1, name, true))) {
					linux_doc_line_break(file, char_line_count, 
							     24+url.length()+name.length());
					i = j;
					fprintf(file, "<htmlurl url=\"%s\" name=\"%s\">", url.c_str(), name.c_str());
				}
			}
			break;
			case Inset::IGNORE_CODE:
				/* ignore */ 
				break;
			default:
				LinuxDocError(par, i, 
					      "Error : unhandled inset.\n");
				++i;
				break;
			}
		} else {
			string sgml_string;
			if (linuxDocConvertChar(c, sgml_string) 
			    && !style->free_spacing) { // in freespacing
				                     // mode, spaces are
				                     // non-breaking characters
				// char is ' '
				if (desc_on == 1) {
					++char_line_count;
					linux_doc_line_break(file, char_line_count, 6);
					fprintf(file, "</tag>");
					desc_on = 2;
				}
				else  {
					linux_doc_line_break(file, char_line_count, 1);
					fprintf(file, "%c", c);
				}
			}
			else {
				fprintf(file, "%s", sgml_string.c_str());
				char_line_count += sgml_string.length();
			}
		}
		font1 = font2;
	}

	/* needed if there is an optional argument but no contents */ 
	if (main_body > 0 && main_body == par->last()) {
		font1 = style->font;
	}

	/* pop all defined Styles (non verbatim) */
	if (!verb_on) 
		for (j=stack_num; j>=0; --j) {
			linux_doc_line_break(file, 
					     char_line_count, 
					     3+strlen(stack[j]));
			fprintf(file, "</%s>", stack[j]);
		}

	/* closes quote tag if necessary */
	if (is_quote) {
		linux_doc_line_break(file, char_line_count, 8);
		fprintf(file, "</quote>");
	}

	/* resets description flag correctly */
	if (desc_on == 2) {
		desc_on = 1;
	} else if (desc_on == 1) { 
		/* <tag> not closed... */
		linux_doc_line_break(file, char_line_count, 6);
		fprintf(file, "</tag>\n");
	}
}


bool Buffer::linuxDocConvertChar(char c, string &sgml_string)
{
	bool retval = false;
	switch (c) {
	case LYX_META_NEWLINE:
		sgml_string = '\n';
		break;
	case '&': 
		sgml_string = "&amp;";
		break;
	case '<': 
		sgml_string = "&lt;"; 
		break;
	case '>':
		sgml_string = "&gt;"; 
		break;
	case '$': 
		sgml_string = "&dollar;"; 
		break;
	case '#': 
		sgml_string = "&num;";
		break;
	case '%': 
		sgml_string = "&percnt;";
		break;
	case '[': 
		sgml_string = "&lsqb;";
		break;
	case ']': 
		sgml_string = "&rsqb;";
		break;
	case '{': 
		sgml_string = "&lcub;";
		break;
	case '}': 
		sgml_string = "&rcub;";
		break;
	case '~': 
		sgml_string = "&tilde;";
		break;
	case '\"': 
		/* should work says the doc, but ... fprintf(file, "&dquote;"); */
		sgml_string = "\"";
		break;
	case ' ':
		retval = true;
                sgml_string = ' ';
		break;
	case '\0': /* Ignore :-) */
                sgml_string.erase();
		break;
	default:
		sgml_string = c;
		break;
	}
	return retval;
}


/* print an error message */
void Buffer::LinuxDocError(LyXParagraph *par, int pos,
			   string const &message) 
{
	InsetError *new_inset;

	/* insert an error marker in text */
	new_inset = new InsetError(message);
	par->InsertChar(pos, LYX_META_INSET);
	par->InsertInset(pos, new_inset);
}


void Buffer::RoffAsciiTable(FILE */*file*/, LyXParagraph */*par*/)
{
	// Commented out so that I don't have to bother with this during
	// kernel rewrite.
#if 0
	LyXFont
		font1 =  LyXFont(LyXFont::ALL_INHERIT),
		font2;
	Inset
		*inset;
	int
		i,j,
		cell = 0;
	char
		c;
	string
		fname1,
		fname2;
	FILE
		*fp,*fp2;
	
	fname1 = TmpFileName(string(),"RAT1");
	fname2 = TmpFileName(string(),"RAT2");
	if (!(fp=fopen(fname1.c_str(),"w"))) {
		WriteAlert(_("LYX_ERROR:"),
			   _("Cannot open temporary file:"), fname1);
		return;
	}
	par->table->RoffEndOfCell(fp, -1);
	for (i = 0; i < par->last; ++i) {
		c = par->GetChar(i);
		if (par->table->IsContRow(cell)) {
			if (c == LYX_META_NEWLINE)
				++cell;
			continue;
		}
		font2 = par->GetFontSettings(i);
		if (font1.latex() != font2.latex()) {
			if (font2.latex() != LyXFont::OFF)
				continue;
		}
		switch (c) {
		case LYX_META_INSET:
			if ((inset = par->GetInset(i))) {
				if (!(fp2=fopen(fname2.c_str(),"w+"))) {
					WriteAlert(_("LYX_ERROR:"),
						   _("Cannot open temporary file:"), fname2);
					fclose(fp);
					remove(fname1.c_str());
					return;
				}
				inset->Latex(fp2,-1);
				rewind(fp2);
				c = fgetc(fp2);
				while(!feof(fp2)) {
					if (c == '\\')
						fprintf(fp,"\\\\");
					else
						fputc(c,fp);
					c = fgetc(fp2);
				}
				fclose(fp2);
			}
			break;
		case LYX_META_NEWLINE:
			if (par->table->CellHasContRow(cell)>=0)
				par->RoffContTableRows(fp, i+1,cell);
			par->table->RoffEndOfCell(fp, cell);
			++cell;
			break;
		case '\\': 
			fprintf(fp, "\\\\");
			break;
		default:
			if (c != '\0')
				fprintf(fp, "%c", c);
			else if (c == '\0')
				lyxerr[Debug::INFO] << "RoffAsciiTable:"
					" 0 char in structure.\n";
			break;
		}
	}
	par->table->RoffEndOfCell(fp, cell);
	fclose(fp);
	string cmd = lyxrc->ascii_roff_command + " >" + fname2;
	cmd.subst("$$FName",fname1);
	Systemcalls one(Systemcalls::System, cmd);
	if (!(lyxerr.debugging())) {
		remove(fname1.c_str());
	}
	if (!(fp=fopen(fname2.c_str(),"r"))) {
		WriteFSAlert(_("Error! Can't open temporary file:"), fname2);
		return;
	}
	// now output the produced file
	fprintf(file, "\n\n");
	c = fgetc(fp);
	if (feof(fp))
		WriteAlert(_("Error!"),
			   _("Error executing *roff command on table"));
	// overread leading blank lines
	while(!feof(fp) && (c == '\n'))
		c = fgetc(fp);
	while(!feof(fp)) {
		for(j=0; j<par->depth; ++j)
			fprintf(file, "  ");
		while(!feof(fp) && (c != '\n')) {
			fputc(c,file);
			c = fgetc(fp);
		}
		fputc('\n',file);
		// overread trailing blank lines
		while(!feof(fp) && (c == '\n'))
			c = fgetc(fp);
	}
	fclose(fp);
	remove(fname2.c_str());
#endif
}


bool Buffer::isLinuxDoc()
{
	return textclasslist.nameOfClass(params.textclass) == "linuxdoc";
}





void Buffer::validate(LaTeXFeatures &features)
{
        LyXTextClass const *tclass = textclasslist.textClass(params.textclass);
    
        // AMS Style is at document level
    
        features.amsstyle = (params.use_amsmath ||
			     prefixIs(tclass->latexname, "ams"));

	// This uses recursive decent.
	paragraph->validate(features);
	
	// the bullet shapes are buffer level not paragraph level
	// so they are tested here
	for (int i = 0; i < 4; ++i) {
		if (params.user_defined_bullets[i] != ITEMIZE_DEFAULTS[i]) {
			int font = params.user_defined_bullets[i].getFont();
			if (font == 0) {
				int c = params.user_defined_bullets[i].getCharacter();
				if (c == 16
				   || c == 17
				   || c == 25
				   || c == 26
				   || c == 31) {
					features.latexsym = true;
				}
			}
			if (font == 1) {
				features.amssymb = true;
			}
			else if ((font >= 2 && font <=5)) {
				features.pifont = true;
			}
		}
	}
	
	if (lyxerr.debugging(Debug::LATEX)) {
		features.showStruct(params);
	}
}


void Buffer::setPaperStuff()
{
	char c1, c2;
 
	params.papersize = PAPER_DEFAULT;
	c1 = params.paperpackage;
	if (c1 == PACKAGE_NONE) {
		c2 = params.papersize2;
		if (c2 == VM_PAPER_USLETTER)
			params.papersize = PAPER_USLETTER;
		else if (c2 == VM_PAPER_USLEGAL)
			params.papersize = PAPER_LEGALPAPER;
		else if (c2 == VM_PAPER_USEXECUTIVE)
			params.papersize = PAPER_EXECUTIVEPAPER;
		else if (c2 == VM_PAPER_A3)
			params.papersize = PAPER_A3PAPER;
		else if (c2 == VM_PAPER_A4)
			params.papersize = PAPER_A4PAPER;
		else if (c2 == VM_PAPER_A5)
			params.papersize = PAPER_A5PAPER;
		else if ((c2 == VM_PAPER_B3) || (c2 == VM_PAPER_B4) ||
			 (c2 == VM_PAPER_B5))
			params.papersize = PAPER_B5PAPER;
	} else if ((c1 == PACKAGE_A4) || (c1 == PACKAGE_A4WIDE) ||
		   (c1 == PACKAGE_WIDEMARGINSA4))
		params.papersize = PAPER_A4PAPER;
}


/* This function should be in Buffer because it's a buffer's property (ale) */
string Buffer::getIncludeonlyList(char delim)
{
	string list;
	paragraph->getIncludeonlyList(list, delim);
	lyxerr << "Includeonly(" << list << ")\n";
	return list;
}


/* This is also a buffer property (ale) */ 
string Buffer::getReferenceList(char delim)
{
	string list;
	paragraph->getReferenceList(list, delim);
	lyxerr[Debug::INFO] << "References(" << list << ")\n";
	return list;
}


LyXCursor Buffer::getLabelCursor(const string &label)
{
	return paragraph->getLabelCursor(label);
}


bool Buffer::isDepClean(string const & name) const
{
	DEPCLEAN* item = dep_clean;
	while (item && item->master != name)
		item = item->next;
	if (!item) return true;
	return item->clean;
}


void Buffer::markDepClean(string const & name)
{
	if (!dep_clean) {
		dep_clean = new DEPCLEAN;
		dep_clean->clean = true;
		dep_clean->master = name;
		dep_clean->next = 0;
	} else {
		DEPCLEAN* item = dep_clean;
		while (item && item->master != name)
			item = item->next;
		if (item) {
			item->clean = true;
		} else {
			item = new DEPCLEAN;
			item->clean = true;
			item->master = name;
			item->next = 0;;
		}
	}
}

void Buffer::markDirty()
{
    if (lyx_clean) {
	lyx_clean = false;
	updateTitles();
    }
    dvi_clean = false;
    bak_clean = false;
    DEPCLEAN* tmp = dep_clean;
    while (tmp) {
	tmp->clean = false;
	tmp = tmp->next;
    }
}
