/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Word Processor
 *
 *           Copyright (C) 1995 Matthias Ettrich
 *           Copyright (C) 1995-1999 The LyX Team. 
 *
 *           This file is Copyright (C) 1996-1999
 *           Lars Gullik Bjnnes
 *
 *======================================================
 */

#ifdef __GNUG__
#pragma implementation
#endif

#include <config.h>
#include <sys/types.h>
#include <utime.h>
#include <algorithm>
#include "bufferlist.h"
#include "lyx_main.h"
#include "minibuffer.h"
#include "FileInfo.h"
#include "filetools.h"
#include "lyx_gui_misc.h"
#include "lastfiles.h"
#include "debug.h"
#include "lyxrc.h"
#include "lyxscreen.h"
#include "lyxtext.h"
#include "lyx_cb.h"
#include "gettext.h"
#include "LAssert.h"
#include "lstrings.h"

// 	$Id: bufferlist.C,v 1.14 1998/10/20 13:10:19 larsbj Exp $	

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: bufferlist.C,v 1.14 1998/10/20 13:10:19 larsbj Exp $";
#endif /* lint */

extern MiniBuffer *minibuffer;
extern int RunLinuxDoc(int, string const &);


//
// Class BufferList
//
BufferList::BufferList()
{
	_state = BufferList::OK;
}


BufferList::~BufferList()
{
	// I am sure something should be done here too.
}


bool BufferList::empty() const
{
	return bstore.empty();
}


bool BufferList::QwriteAll()
{
        bool askMoreConfirmation = false;
        string unsaved;
	for (BStore::iterator itr = bstore.begin();
	     itr != bstore.end(); ++itr) {
		if(itr->isLyxClean()) continue;
		
		switch(AskConfirmation(_("Changes in document:"),
				       MakeDisplayPath(itr->filename,
						       50),
				       _("Save document?"))) {
		case 1: // Yes
			menuWrite(itr);
			break;
		case 2: // No
			askMoreConfirmation = true;
			unsaved += MakeDisplayPath(itr->filename,50);
			unsaved += "\n";
			break;
		case 3: // Cancel
			return false;
		}
	}
	// This should not be placed here. Lgb.
        if (askMoreConfirmation &&
            lyxrc->exit_confirmation &&
            !AskQuestion(_("Some documents were not saved:"),
                         unsaved, _("Exit anyway?"))) {
                return false;
        }
        return true;
}


// Should probably be moved to somewhere else: BufferView? LyXView?
bool BufferList::write(Buffer *buf, bool makeBackup)
{
	minibuffer->Set(_("Saving document"),
			MakeDisplayPath(buf->filename),"...");

	// We don't need autosaves in the immediate future. (Asger)
	buf->resetAutosaveTimers();

	// make a backup
	if (makeBackup) {
		string s = buf->filename + '~';
		// Rename is the wrong way of making a backup,
		// this is the correct way.
		/* truss cp fil fil2:
		   lstat("LyXVC3.lyx", 0xEFFFF898)                 Err#2 ENOENT
		   stat("LyXVC.lyx", 0xEFFFF688)                   = 0
		   open("LyXVC.lyx", O_RDONLY)                     = 3
		   open("LyXVC3.lyx", O_WRONLY|O_CREAT|O_TRUNC, 0600) = 4
		   fstat(4, 0xEFFFF508)                            = 0
		   fstat(3, 0xEFFFF508)                            = 0
		   read(3, " # T h i s   f i l e   w".., 8192)     = 5579
		   write(4, " # T h i s   f i l e   w".., 5579)    = 5579
		   read(3, 0xEFFFD4A0, 8192)                       = 0
		   close(4)                                        = 0
		   close(3)                                        = 0
		   chmod("LyXVC3.lyx", 0100644)                    = 0
		   lseek(0, 0, SEEK_CUR)                           = 46440
		   _exit(0)
		 */

		// Should proabaly have some more error checking here.
		// Should be cleaned up in 0.13, at least a bit.
		// Doing it this way, also makes the inodes stay the same.
		// This is still not a very good solution, in particular we
		// might loose the owner of the backup.
		FileInfo finfo(buf->filename);
		if (finfo.exist()) {
			mode_t fmode = finfo.getMode();

			struct utimbuf *times =
				(struct utimbuf*)new char[sizeof(struct utimbuf)];
			times->actime = finfo.getAccessTime();
			times->modtime = finfo.getModificationTime();
			long blksize = finfo.getBlockSize();
			lyxerr[Debug::INFO] << "BlockSize: "
					    << blksize << '\n';
			FilePtr fin(buf->filename,FilePtr::read);
			FilePtr fout(s,FilePtr::truncate);
			if (fin() && fout()) {
				char * cbuf = new char[blksize+1];
				size_t c_read = 0;
				size_t c_write = 0;
				do {
					c_read = fread(cbuf, 1, blksize, fin);
					if (c_read != 0)
						c_write =
							fwrite(cbuf, 1, c_read, fout);
				} while (c_read);
				fin.close();
				fout.close();
				chmod(s.c_str(), fmode);
				
				if (utime(s.c_str(), times)) {
					lyxerr << "utime error.\n";
				}
				delete [] cbuf;
			} else {
				lyxerr << "LyX was not able to make "
					"backupcopy. Beware.\n";
			}
			delete[] times;
		}
	}
	
	if (buf->writeFile(buf->filename,false)) {
		buf->markLyxClean();

		minibuffer->Set(_("Document saved as"),
				MakeDisplayPath(buf->filename));

		// now delete the autosavefile
		string a = OnlyPath(buf->filename);
		a += '#';
		a += OnlyFilename(buf->filename);
		a += '#';
		FileInfo fileinfo(a);
		if (fileinfo.exist()) {
			if (remove(a.c_str()) != 0) {
				WriteFSAlert(_("Could not delete "
					       "auto-save file!"), a);
			}
		}
	} else {
		// Saving failed, so backup is not backup
		if (makeBackup) {
			string s = buf->filename + '~';
			rename(s.c_str(), buf->filename.c_str());
		}
		minibuffer->Set(_("Save failed!"));
		return false;
	}

	return true;
}


void BufferList::closeAll()
{
	_state = BufferList::CLOSING;

	// Is this safe? Lgb.
	while (!bstore.empty()) {
		close(bstore.begin());
	}
	_state = BufferList::OK;
}


// Should perhaps be placed somewhere else. Lgb.
class resize_buffer {
public:
	void operator()(Buffer & b) {
		b.resize();
	}
};

void BufferList::resize()
{
	for_each(bstore.begin(), bstore.end(), resize_buffer());
}


bool BufferList::close(Buffer * ci)
{
	//	ci->InsetUnlock();
	
	if (ci->paragraph && !ci->isLyxClean() && !quitting) {
		ProhibitInput();
                switch(AskConfirmation(_("Changes in document:"),
				       MakeDisplayPath(ci->filename,50),
				       _("Save document?"))){
		case 1: // Yes
			if (write(ci)) {
				lastfiles->newFile(ci->filename);
			} else {
				AllowInput();
				return false;
			}
                        break;
		case 3: // Cancel
                        AllowInput();
                        return false;
                }
		AllowInput();
	}

	bstore.erase(ci);
	return true;
}


class add_to_pup {
	int pup;
public:
	add_to_pup(int p)
		: pup(p) {}
	void operator()(Buffer const & b) const {
		string relbuf = MakeDisplayPath(b.getFileName(), 30);
		fl_addtopup(pup, relbuf.c_str());
	}
};

void BufferList::makePup(int pup)
	/* This should be changed to return a char const *const
	   in the same way as for lastfiles.[hC]
	   */
{
	if (bstore.size() == 0)
		fl_addtopup(pup,_("No Documents Open!%t"));

	for_each(bstore.begin(), bstore.end(), add_to_pup(pup));
}


Buffer * BufferList::first()
{
	if (bstore.empty())
		return 0;
	
	return bstore.begin();
}


Buffer * BufferList::getBuffer(unsigned int choice)
{
	Assert(choice > 0 && choice < bstore.size() + 1);
	return &bstore[choice-1];
}


class find_updateinset {
	Inset * inset;
public:
	find_updateinset(Inset * i)
		: inset(i) {}
	bool operator()(Buffer & b) {
		if (b.getUser() && b.getUser()->text->UpdateInset(inset))
			return true;
		return false;
	}
};

void BufferList::updateInset(Inset *inset, bool mark_dirty)
{
	BStore::iterator it = find_if(bstore.end(), bstore.end(),
					find_updateinset(inset));
	if (it != bstore.end() && mark_dirty)
		it->markDirty();
}


#if 0
class find_locking_inset {
	UpdatableInset * inset;
public:
	find_locking_inset(UpdatableInset * i)
		: inset(i) {}
	bool operator()(Buffer const & b) const {
		if (b.the_locking_inset == inset) return true;
		return false;
	}
};
#endif

int BufferList::unlockInset(UpdatableInset *inset)
{
	/*
	if (!inset) return 1;

	BStore::const_iterator it = find_if(bstore.begin(), bstore.end(),
	find_locking_inset(inset));
	if (it != bstore.end()) {
	it->InsetUnlock();
	return 0;
	}
	*/
	return 1;
}


class update_included_tex_files {
	string tmpdir;
public:
	update_included_tex_files(string const & td)
		: tmpdir(td) {}
	
	void operator()(Buffer & b) {
		if (!b.isDepClean(tmpdir)) {
			string writefile(tmpdir);
			writefile += '/';
			writefile += ChangeExtension(b.getFileName(),
						     ".tex", true);
			b.makeLaTeXFile(writefile, tmpdir, true, true);
			b.markDepClean(tmpdir);
		}
	}
};

void BufferList::updateIncludedTeXfiles(string const & mastertmpdir)
{
	for_each(bstore.begin(), bstore.end(),
		 update_included_tex_files(mastertmpdir));
}

void BufferList::emergencyWriteAll()
{
	for (BStore::iterator it = bstore.begin(); it != bstore.end(); ++it) {
		if (it->isLyxClean()) continue;
		
		bool madeit=false;
		
		lyxerr << _("LyX: Attempting to save document ")
		       << it->filename
		       << _(" as...") << '\n';
		
		for (int i=0; i<3 && !madeit; ++i) {
			string s;
			
			// We try to save three places:
			// 1) Same place as document.
			// 2) In HOME directory.
			// 3) In "/tmp" directory
			switch(i) {
			case 0:
				s = it->filename;
				break;
			case 1:
				s = AddName(GetEnvPath("HOME"),
					    it->filename);
				break;
			case 2:
				// MakeAbsPath to prepend the
				// current drive letter on OS/2
				s = AddName(MakeAbsPath("/tmp/"),
					    it->filename);
				break;
			default:
				// if this happens it is a bug
				Assert(i>=0 && i <=2);
				break;
			}
			s += ".emergency";
			
			lyxerr << "  " << i+1 << ") " << s << '\n';
			
			if (it->writeFile(s,true)) {
				it->markLyxClean();
				lyxerr << _("  Save seems succesful. Phew.")
				       << '\n';
				madeit = true;
			} else if (i != 2) {
				lyxerr << _("  Save failed! Trying...")
				       << '\n';
			} else {
				lyxerr << _("  Save failed! Bummer. "
					    "Document is lost.") << '\n';
			}
		}
	}
}


Buffer * BufferList::readFile(string const & s, bool ronly)
{
	// We are not using push_back, because then we would not get
	// an iterator returned. Lgb.
	BStore::iterator b = bstore.insert(bstore.end(),
					   Buffer(s, lyxrc, ronly));
	b->params.useClassDefaults();
	
	string ts = s;
	string e = OnlyPath(s);
	string a = e;
	// File information about normal file
	FileInfo fileInfo2(s);

	// Check if emergency save file exists and is newer.
	e += OnlyFilename(s) + ".emergency";
	FileInfo fileInfoE(e);

	bool use_emergency = false;

	if (fileInfoE.exist() && fileInfo2.exist()) {
		if (fileInfoE.getModificationTime()
		    > fileInfo2.getModificationTime()) {
			if (AskQuestion(_("An emergency save of this document exists!"),
					MakeDisplayPath(s,50),
					_("Try to load that instead?"))) {
				ts = e;
				// the file is not saved if we load the
				// emergency file.
				b->markDirty();
				use_emergency = true;
			}
		}
	}

	if (!use_emergency) {
		// Now check if autosave file is newer.
		a += '#';
		a += OnlyFilename(s);
		a += '#';
		FileInfo fileInfoA(a);
		if (fileInfoA.exist() && fileInfo2.exist()) {
			if (fileInfoA.getModificationTime()
			    > fileInfo2.getModificationTime()) {
				if (AskQuestion(_("Autosave file is newer."),
						MakeDisplayPath(s,50),
						_("Load that one instead?"))) {
					ts = a;
					// the file is not saved if we load the
					// autosave file.
					b->markDirty();
				}
			}
		}
	}
	// not sure if this is the correct place to begin LyXLex
	LyXLex lex(0, 0);
	lex.setFile(ts);
	if (b->readFile(lex))
		return b;
	else {
		bstore.erase(b);
		return 0;
	}
}


class find_filename {
	string fn;
public:
	find_filename(string const & f)
		: fn(f) {}
	bool operator()(Buffer const & b) const {
		if (b.getFileName() == fn) return true;
		return false;
	}
};

bool BufferList::exists(string const & s) const
{
	BStore::const_iterator ci = find_if(bstore.begin(), bstore.end(),
					    find_filename(s));
	if (ci != bstore.end()) return true;
	return false;
}


Buffer * BufferList::getBuffer(string const &s)
{
	BStore::iterator it = find_if(bstore.begin(), bstore.end(),
				      find_filename(s));
	if (it != bstore.end()) return it;
	return 0;
}


Buffer * BufferList::newFile(string const & name, string tname)
{
	BStore::iterator b = bstore.insert(bstore.end(), Buffer(name, lyxrc));
	b->params.useClassDefaults();

	// use defaults.lyx as a default template if it exists.
	if (tname.empty()) {
		tname = LibFileSearch("templates", "defaults.lyx");
	}
	if (!tname.empty() && IsLyXFilename(tname)){
		bool templateok = false;
		LyXLex lex(0,0);
		lex.setFile(tname);
		if (lex.IsOK()) {
			if (b->readFile(lex)) {
				templateok = true;
			}
		}
		if (!templateok) {
			WriteAlert(_("Error!"),_("Unable to open template"), 
				   MakeDisplayPath(tname));
			// no template, start with empty buffer
			b->paragraph = new LyXParagraph();
		}
	} else {  // start with empty buffer
		b->paragraph = new LyXParagraph();
	}

	b->markDirty();
	b->setReadonly(false);
	
	return b;
}


Buffer* BufferList::loadLyXFile(string const & filename, bool tolastfiles)
{
	// make sure our path is absolute
	string s = MakeAbsPath(filename);
	
	// Is it LinuxDoc?
#warning Disabled...
#if 0
	if (IsSGMLFilename(s)) {
		if (!RunLinuxDoc(-1, s)) {
			s = ChangeExtension (s, ".lyx", false);
		} else { // sgml2lyx failed
			WriteAlert(_("Error!"),
				   _("Could not convert file"),s);
			return 0;
		}
	}
#endif
	// file already open?
	if (exists(s)) {
		if (AskQuestion(_("Document is already open:"), 
				MakeDisplayPath(s,50),
				_("Do you want to reload that document?"))) {
			// Reload is accomplished by closing and then loading
			if (!close(getBuffer(s))) {
				return 0;
			}
			// Fall through to new load. (Asger)
		} else {
			// Here, we pretend that we just loaded the
			// open document
			return getBuffer(s);
		}
	}
	Buffer *b=0;
	bool ro = false;
	switch (IsFileWriteable(s)) {
	case 0:
		minibuffer->Set(_("File `")+MakeDisplayPath(s,50)+
				_("' is read-only."));
		ro = true;
		// Fall through
	case 1:
		b=readFile(s, ro);
		if (b) {
			b->lyxvc.file_found_hook(s);
		}
		break; //fine- it's r/w
	case -1:
		// Here we probably should run
		if (LyXVC::file_not_found_hook(s)) {
			// Ask if the file should be checked out for
			// viewing/editing, if so: load it.
			lyxerr << "Do you want to checkout?\n";
		}
		if (AskQuestion(_("Cannot open specified file:"), 
				MakeDisplayPath(s,50),
				_("Create new document with this name?")))
	    	{
			// Find a free buffer
			b=newFile(s,string());
	    	}
		break;
	}

	if (b && tolastfiles)
		lastfiles->newFile(b->getFileName());

	return b;
}
