// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1999 Asger Alstrup Nielsen
 *
 *======================================================*/

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "Encoding.h"
#include "EncodingConverter.h"

// Structure to keep information about the encodings LyX supports.
struct EncodingData {
	Encoding encoding;
	char const * name;
	char const * gui_name;
	bool fixed_width;
	bool to_unicode;
};

/////////////////////////////////////////////////////////////////////////
//
// If you want to implement a new encoding, please implement the interface
// in EncodingConverter.h, add the name of the encoding to the Encoding
// enum in include/Encoding.h.  Add the new files to Makefile.am, and
// an appropriate #include below.
// Now, fill out this table and the accessor function further below 
// with the corresponding information. Then LyX will fully support the 
// new encoding. (Asger)
//

#include "EncLatin1.h"
#include "EncLatin2.h"
#include "EncLatin3.h"
#include "EncLatin4.h"
#include "EncLatin5.h"
#ifdef BIDIRECTIONAL
#include "EncLatin6.h"
#endif
#include "EncLatin7.h"
#ifdef BIDIRECTIONAL
#include "EncLatin8.h"
#endif
#include "EncLatin9.h"
#include "EncLatin14.h"
#include "EncLatin15.h"

static EncodingData database[] =
{ 
//  Encoding	Name		GUI Name		Fixed? To Unicode?
  { ISO8859_1,	"iso8859-1",	"Western (ISO-8859-1, Latin1)", true,  true },
  { ISO8859_2,	"iso8859-2",	"Central and East European (ISO-8859-2, Latin2)", true,  true },
  { ISO8859_3,	"iso8859-3",	"Maltese, Esperanto (ISO-8859-3, Latin3)", true,  true },
  { ISO8859_4,	"iso8859-4",	"Baltic, Greenlandic, Lappish (ISO-8859-4, Latin4)", true,  true },
  { ISO8859_5,	"iso8859-5",	"Cyrrilic (ISO-8859-5)", true,  true },
#ifdef BIDIRECTIONAL
  { ISO8859_6,	"iso8859-6",	"Arabic (ISO-8859-6)",	true,  true },
#endif
  { ISO8859_7,	"iso8859-7",	"Greek (ISO-8859-7)",	true,  true },
#ifdef BIDIRECTIONAL
  { ISO8859_8,	"iso8859-8",	"Hebrew, Yiddish (ISO-8859-8)",	true,  true },
#endif
  { ISO8859_9,	"iso8859-9",	"Turkish (ISO-8859-9, Latin5)",	true,  true },
  { ISO8859_14,	"iso8859-14",	"Celtic (ISO-8859-14, Latin8)",true,  true },
  { ISO8859_15,	"iso8859-15",	"Western with the Euro sign (ISO-8859-15, Latin9)",true,  true },
  { Encoding(0xFFFF), 0, 0, false, false }
};

// An encoding converter accessor.
static EncodingConverter & getEncodingConverter(Encoding e) {
	static EncLatin1 latin1; 
	static EncLatin2 latin2; 
	static EncLatin3 latin3; 
	static EncLatin4 latin4;
	static EncLatin5 latin5;
#ifdef BIDIRECTIONAL
	static EncLatin6 latin6; 
#endif	
	static EncLatin7 latin7; 
#ifdef BIDIRECTIONAL
	static EncLatin8 latin8; 
#endif	
	static EncLatin9 latin9; 
	static EncLatin14 latin14; 
	static EncLatin15 latin15; 
	switch(e) {
	case ISO8859_1: return latin1;
	case ISO8859_2: return latin2;
	case ISO8859_3: return latin3;
	case ISO8859_4: return latin4;
	case ISO8859_5: return latin5;
#ifdef BIDIRECTIONAL
	case ISO8859_6: return latin6;
#endif
	case ISO8859_7: return latin7;
#ifdef BIDIRECTIONAL
	case ISO8859_8: return latin8;
#endif
	case ISO8859_9: return latin9;
	case ISO8859_14: return latin14;
	case ISO8859_15: return latin15;
	}
	// To shut up gcc
	return latin1;
}

//
// No need to change anything below this point.
//
/////////////////////////////////////////////////////////////////////////


// Implementation of general encoding conversion
std::pair<LString, LString::const_iterator>
convertEncoding(LString const & s, Encoding from, Encoding to) {
	if (from == to) {
		return std::make_pair(s, s.end());
	}

	// Here, more efficient custom converters can be plugged in.

	// Otherwise, we use the default, slow implementation through Unicode
	std::pair<wstring, LString::const_iterator> unicode_result;
	unicode_result = convertToUnicode(s, from);
	if (unicode_result.second == s.begin()) {
		// Nothing was converted
		return std::make_pair(LString(), s.begin());
	}

	std::pair<LString, wstring::const_iterator> to_result;
	to_result = convertFromUnicode(unicode_result.first, to);
	if (to_result.second == unicode_result.first.end()) {
		// All of the Unicode was converted correctly
		return std::make_pair(to_result.first, unicode_result.second);
	}
	if (to_result.second == unicode_result.first.begin()) {
		// Nothing was converted correctly
		return std::make_pair(LString(), s.begin());
	}

	// Ok, we need to find out where the error appears

	if (isEncodingFixedWidth(from)) {
		// If the original string is fixed width encoded, we
		// can calculate it:
		return std::make_pair(to_result.first, s.begin() 
		     + (to_result.second - unicode_result.first.begin()) );
	}

	// Otherwise, we have to try each char until we find the one that fails
	LString::const_iterator i = s.begin();

	for (; i != unicode_result.second; ++i) {
		LString glyph(1, *i);
		if (convertEncoding(glyph, from, to).second == glyph.begin())
			break;
	}
	return std::make_pair(to_result.first, i);
}


/// Convert a string to Unicode
std::pair<wstring, LString::const_iterator>
convertToUnicode(LString const & s, Encoding from) {
	return getEncodingConverter(from).toUnicode(s);
}

/// Convert a string from Unicode
std::pair<LString, wstring::const_iterator>
convertFromUnicode(wstring const & s, Encoding to) {
	return getEncodingConverter(to).fromUnicode(s);
}

/// Return the name of an encoding as a single word in ASCII
LString getEncodingName(Encoding e) {
	return database[e].name;
}

/// Get the encoding from an ASCII name
Encoding getEncodingFromName(LString const & name) {
	for (int i=0; database[i].encoding != 0xFFFF; ++i) {
		if (database[i].name == name)
			return database[i].encoding;
	}
	return ISO8859_1; // To shut up gcc
}

/// Return the GUI name of an encoding
LString getEncodingGuiName(Encoding e) {
	return database[e].gui_name;
}

/// Get the encoding from a GUI name
Encoding getEncodingFromGuiName(LString const & name) {
	for (int i=0; database[i].encoding != 0xFFFF; ++i) {
		if (database[i].gui_name == name)
			return database[i].encoding;
	}
	return ISO8859_1; // To shut up gcc
};

/// Can all chars of an encoding be contained in a single LChar?
bool isEncodingFixedWidth(Encoding e) {
	return database[e].fixed_width;
}

/// Can this encoding be converted to Unicode? 
bool canConvertToUnicode(Encoding e) {
	return database[e].to_unicode;
}
