// -*- C++ -*-
/*
 *  Combox: A combination of two objects (a button and a browser) is
 *          encapsulated to get a combobox-like object. All XForms 
 *          functions are hidden.         
 * 
 *  GNU Copyleft (C) 1996 Alejandro Aguilar Sierra <asierra@servidor.unam.mx>
 *                        and the LyX Team.
 * 
 *  Dependencies:  Only XForms, but created to be used with LyX.
 * 
 */ 

/* Change log:
 *  
 *  2/06/1996,   Alejandro Aguilar Sierra 
 *    Created and tested.
 *  
 *  4/06/1996,   Alejandro Aguilar Sierra 
 *    Added droplist mode (a button with a black down arrow at right)
 *    and support for middle and right buttons, as XForms choice object.
 */ 

#ifndef _COMBOX_H_
#define _COMBOX_H_

#ifdef __GNUG__
#pragma interface
#endif

#include FORMS_H_LOCATION
#include <stdlib.h>

#include "LString.h"

/// callback prototype
typedef void (*FL_COMBO_CB) (int, void*);
/// input callback prototype
typedef void (*FL_COMBO_INPUT_CB) (void*);
/// pre post prototype
typedef void (*FL_COMBO_PRE_POST) ();


///
class Combox {
public:
	//
	enum combox_type {
		///
		FL_COMBOX_NORMAL,
		///
		FL_COMBOX_DROPLIST,
		///
		FL_COMBOX_INPUT
	};
	///
	Combox(combox_type t=FL_COMBOX_NORMAL);
	///
	~Combox();

	/** To add this object to a form. Note that there are two heights
	 for normal (button) and expanded (browser) mode each. */
	void add(int x, int y, int w, int hmin, int hmax);
	
	/// Add lines. Same as for fl_browser object
	void addline(string const &);
	/// Add lines. Same as for fl_browser object
	void addto(string const &);
	
	/// Returns the selected item
	int get();
   
	/// Returns a pointer to the selected line of text
	string getline();
   
	///  Select an arbitrary item
	void select(int);
	///
        bool select_text(string const &);
   
	///  Clear all the list
	void clear();

	/// Is the combox cleared (empty)
	bool empty() { return is_empty; }
	
	/// Remove the objects from the form they are in. 
	void remove();

	/**  Assign a callback to this object. The callback should be a void
	 function with a int and a void pointer as parameters. */
	void setcallback(FL_COMBO_CB, void *);

	/** Assign a special callback that is called when the input is changed
	    -- only useful if you are using an FL_COMBO_INPUT in conjunction
	    with an input_return.  A bit hacky but very useful for controlling
	    gui interface. ARRae 990103 */
	void setinputcallback(FL_COMBO_INPUT_CB, void *);
	
        ///  Pre handler
	void setpre(FL_COMBO_PRE_POST);
	/// Post handler
	void setpost(FL_COMBO_PRE_POST);
	
	///  XForms attributes
	void resize(unsigned);
	///
	void gravity(unsigned, unsigned);
	///
	void activate();
	///
	void deactivate();
	///
        void shortcut(string const &, int);
	///
	void input_return(int);
	///
	void Redraw();
	///
	void Show();
protected:
        /// At least Hide should not be public
	void Hide(int who = 0);
	///
	FL_OBJECT *browser;
private:
	///
	combox_type type;
	///
        int bw, bh;
	///
	int sel;
	///
	bool is_empty;
	///
	static void combo_cb(FL_OBJECT *, long);
	///
	static void input_cb(FL_OBJECT *, long);
	///
        static int  peek_event(FL_FORM *, void *);
	///
	FL_COMBO_CB callback;
	///
	void *cb_arg;
	///
	FL_COMBO_INPUT_CB input_callback;
	///
	void *icb_arg;
	///
	int input_return_type;
	///
	FL_COMBO_PRE_POST _pre;
	///
	FL_COMBO_PRE_POST _post;
	///
	FL_OBJECT *button;
	///
	FL_OBJECT *label;
	///
        FL_FORM* form;
};



//-----------------  Inline methods  --------------------------- 

inline
void Combox::addto(string const &text)
{
	if (browser) {
		fl_addto_browser(browser, text.c_str());
		is_empty = false;
	}
}

inline
void Combox::resize(unsigned r)
{
   fl_set_object_resize(button, r);
   if (label!=button) fl_set_object_resize(label, r); 
}

inline
void Combox::gravity(unsigned g1, unsigned g2)
{
   fl_set_object_gravity(button, g1, g2);
   if (label!=button) fl_set_object_gravity(label, g1, g2); 
}

inline
void Combox::shortcut(string const &s, int i)
{
   if (button)
      fl_set_object_shortcut(button,s.c_str(),i);
}

inline
void Combox::setcallback(FL_COMBO_CB cb, void *a = 0)
{
   callback = cb;
   cb_arg = a;
}

inline
void Combox::setinputcallback(FL_COMBO_INPUT_CB cb, void *a = 0)
{
   input_callback = cb;
   icb_arg = a;
}

inline
void Combox::setpre(FL_COMBO_PRE_POST cb)
{
	_pre = cb;
}

inline
void Combox::setpost(FL_COMBO_PRE_POST cb)
{
	_post = cb;
}

inline
int Combox::get()
{
   return sel;
}

inline
string Combox::getline()
{
    if (type==FL_COMBOX_INPUT) 
      return fl_get_input(label);
    else
      return ((browser) ? string(fl_get_browser_line(browser, sel)): string());
}

inline
void Combox::input_return(int style)
{
	if(label
	   && type == FL_COMBOX_INPUT) {
		fl_set_input_return(label, style);
		input_return_type = style;
	}
}

#endif


