// -*- C++ -*-
/* Communicator.h
 */

#ifndef COMMUNICATOR_H
#define COMMUNICATOR_H

#include "LString.h"

class LyXParagraphParameters;
class UpdatableInset;
class LyXParagraph;
class BufferParams;
class LyXViewBase;
class BufferView;  // temporary only
class Combox;
class VSpace;
class Buffer;
class Popups;
class Inset;


/** Communications class for GUI-to-LyX communication.
 * Holds the various methods required by popups/dialogs to communicate
 * with the rest of LyX.  Provides a simpler interface for GUI porters
 * although these methods may be useful for other areas of LyX also.
 * The header files required for each method are listed also.
 */
class Communicator {
public:
	///
	Communicator(LyXViewBase *);
	///
	~Communicator();
	///
	/**@name LyXView related methods */
	//@{

	/**
	 * LyX dispatcher, executes lyx actions.
	 * \#include "commandtags.h"
	 */
	string Dispatch(int action, char const* arg = 0) const;

	/**  
	 * Allows access to Popup signals.
	 * \#include "Popups.h"
	 */
	Popups & getPopups() const;

	/**
	 * @return The directory LyX was installed in.
	 * \#include "string.h"
	 */
	string getSystemLyXDir() const;

	/**
	 * @return The document directory a user has set in their lyxrc file.
	 * \#include "string.h"
	 */
	string getDefaultDocumentPath() const;
	
	/// Are we supposed to use a temporary directory?
	bool usingTempDir() const;
	/// Display string in the minibuffer
	void setMinibuffer(char const *) const;
	/// Ugly and very temporary hack to support form_bibkey hack
	BufferView * getCurrentView() const;
	//@}
	
	/**@name Buffer related methods */
	//@{
	/// Is the current buffer view available.
	bool isAvailable() const;
	///
	void redoBuffer() const;
	/// Tell LyX the buffer contents have changed
	void markDirty() const;

	/**
	 * @return Current buffer
	 * \#include "buffer.h"
	 */
	Buffer * getBuffer() const;

	/**
	 * @return current buffers parameters
	 * \#include "bufferparams.h"
	 */
	BufferParams & getBufferParams() const;
	
	///
	int switchLayoutsBetweenClasses(char, char) const;
	///
	void setPaperStuff() const;
	///
	bool isReadOnly() const;
	///
	bool isLinuxDoc() const;
	///
	string getFileName() const;
	///
	string getTmpDir() const;
	///
	void setPreamble(char const *) const;
	///
	char const * getPreamble() const;
	///
	bool hasLockingInset() const;
	///
	UpdatableInset * getLockingInset() const;
	///
	bool openNewInset(UpdatableInset *) const;
	///
	void updateInset(Inset *inset, bool mark_dirty) const;
	///
	void bibitemUpdate(Combox *combox) const;
	//@}
	
	/**@name Paragraphs at the current cursor position */
	//@{
	///
	void setParagraph(bool line_top,      bool line_bottom,
			  bool pagebreak_top, bool pagebreak_bottom,
			  VSpace space_top,   VSpace space_bottom,
			  char align,         string labelwidthstring,
			  bool noindent) const;
	///
	void setParagraphExtraOpt(int type,
                                  char const *width,
                                  char const *widthp,
                                  int alignment, bool hfill,
                                  bool start_minipage) const;
	/// This should only be temporary until pextra's are moved to pparams
	LyXParagraph * getParAtCursor() const;
	///
	int getParLayout() const;
	///
	int getParAlign() const;
	///
	string getParLabelWidthString() const;

	/**
	 * @return The current paragraphs parameter structure.
	 * \#include "paraparams.h"
	 */
	LyXParagraphParameters & getParaParams() const;
	//@}
private:
	///
	Communicator() {}
	///
	Communicator(Communicator const &) {}
	///
	LyXViewBase *l_view;
};

#endif
