// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1998 The LyX Team
 *
 *======================================================*/

#ifndef FILEFORMAT_H
#define FILEFORMAT_H

#include <config.h>

#ifdef __GNUG__
#pragma interface
#endif

#include "LString.h"
#include "Encoding.h"
class FilePtr;

/** This is a struct that is used to encapsulate the names
    of some useful logical formatting characters. The name 
    is short to allow easy access to the characters.
    We could probably use a namespace instead, but I'm not
    sure older compilers will accept that. 
 */
struct fmt {
	/// A bunch of useful logical formatting symbols
	enum Symbol {
		/// Forced line-break / end of paragraph
		br,
		/// Non-breakable space
		nbsp,
		/// Start of a paragraph
		par,
		/// An hfill
		hfill
	};
};


/** This is an abstract base class for all file formats.

    So all supported file formats should derive from this.
    Have a look at TextFormat.h to see an example.

    The only functionality this base class provides is
    hiding the raw file writing stuff.
    (Asger)
*/
class FileFormat {
public:
	/// All supported fileformats
	enum Format {
		/// Implemented as LatexFormat
		LaTeX,
		/// Implemented as AsciiFormat
		ASCII,
		/// Implemented as DocBookFormat
		DocBook
	};
	
	/// The encoding we get, and the encoding we should produce
	FileFormat(Encoding & inenc_, Encoding & outenc_);

	///
	virtual ~FileFormat();

	/** Define the file to write to.
	    The filename should include a full path. */
	void setFile(string const & filename);

	/** Write a string to the output file. If we fail, return false.
	    Sometimes, the string is already in the correct output encoding.
	    Then, let "convert" be false.
	    We throw a writeError if we have trouble writing to the file.
	*/
	virtual void write(string const &, bool convert=true) = 0;

	/** Write a symbol to the output file. If we fail, return false. 
	    We throw a writeError if we have trouble writing to the file. */
	virtual void write(fmt::Symbol) = 0;

	/** End the writing -- flushes any buffers and closes the file.
	    Throws a writeError if we have trouble. */
	virtual void close();

	/// What kind of file-format are we?
	virtual Format kind() const = 0;

	/// When we have a write error, we throw one of these
	class writeError {
	public:
		///
		writeError(string const & what_)
			: what(what_)
		{ }
		///
		string what;
	};
protected:
	/// Spit this verbatim to the file.  This can throw a writeError. 
	void print(char const *);

	/// What encoding do we get?
	Encoding * inenc;

	/// What encoding should we write?
	Encoding * outenc;
private:
	/// What file do we write to?
	// IOSTREAM!!!
	FilePtr * file;
};

#endif FILEFORMAT_H
