// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1998 The LyX Team
 *
 *======================================================*/

#ifndef TEXTFORMAT_H
#define TEXTFORMAT_H

#include <config.h>

#ifdef __GNUG__
#pragma interface
#endif

#include <vector>
#include <stack>

#include "FileFormat.h"

/** This is an abstract base class for textual file formats derived from
    FileFormat.

    All textual file formats that are based on some kind of ASCII
    format should derive from this, and expand it with methods that 
    help the production of the particular file format.

    The basic feature of this class is to enable the nice creation 
    of a file with margin, word wrap and indentation control, 
    while still being generic enough to support a wide range of textual
    formats.  

    The class uses logical formatting characters to control the output.
    There are two mayor modes of operation, that can be mixed without
    restriction: the line model and the paragraph model.
    
    In a plain line model, the input is fed as lines ended by fmt::br 
    logical formatting symbols.  Each line is formatted to obey word-wrapping 
    and the left and right margins.
    Every space in the input is treated as a breakable space.  You can
    use the fmt::nbsp logical formatting symbol to get unbreakable
    spaces.  These are output as ordinary spaces in the output, but will
    not be used as break-points in the output.

    If the word-wrap feature is disabled, the right margin will not be 
    respected, and the output is allowed to extend past the specified line 
    width.  If the left margin is set to 0, this can be used to get a free
    format model, where the user of the class has full control of the formatting
    of the output (useful for, for instance, verbatim environments.)

    If word-wrapping is enabled, the setting "breakwords" specifies whether
    we should strictly enforce the line width.  It can happen that a very
    long word can not be contained on a single line within the margins.  If 
    the "breakwords" property is "true", we will break these words at the 
    right margin, even if there is no suitable break-point.  If the flag is 
    false, we will spill the word across the right margin, and possibly even 
    beyond the line width.

    The class also supports an optional paragraph model:  Besides use as in
    the line model, the input can be fed as paragraphs delimited by the
    fmt::par and fmt::br logical formatting symbols.  When these are
    used correctly, the output is formatted with the correct indentation of 
    the first line in a paragraph, according to the "parindent" setting (in 
    addition to the margin control as in the line model.)
    Besides ordinary paragraph indents, the parindent can be used to obtain 
    hanging indents and bullets by specifying a negative paragraph indentation.

    Other configurable items are the used "end-of-line" string.
    Should tabs be use to do identation?

    Also, derivate classes can add addition features by overriding some key
    methods.  In this way, the AsciiFormat class implements alignment control,
    and hfills.

    (Asger)
*/
class TextFormat : public FileFormat {
public:
	/** What encoding do we get, and what encoding should we write?  
	    What end-of-line string should we use?
	    How wide is a line at the most in the output file? 
	*/
	TextFormat(Encoding & inenc_, Encoding & outenc_, 
		   string eol_ = string("\n"), int width_ = 75);

	///
	virtual ~TextFormat();

	/** Write a string to the output file. If we fail, return false.
	    Sometimes, the string is already in the correct output encoding.
	    Then, let "convert" be false.
	    The string is NOT allowed to contain any line break characters,
	    or other formatting symbols.  All spaces are treated as breakable
	    if word-wrapping is enabled.
	    We throw a writeError if we have trouble writing to the file.
	*/
	virtual void write(string const &, bool convert=true);

	/** Process a logical formatting symbol.
	    We throw a writeError if we have trouble writing to the file. */
	virtual void write(fmt::Symbol);

	/** End the writing.
	    This is required to flush the last line and close the file 
	    properly. */
	virtual void close();

	/// What kind of file format are we?
	virtual Format kind() const = 0;

	/// Enable/disable word-wrapping
	virtual void setWordwrap(bool wrap = true) {
		wordwrap = wrap;
	}

	/// Enable/disable breaking of long words
	virtual void setBreakWords(bool break_ = true) {
		breakwords = break_;
	}

	/// What is the width of a line?
	virtual int paperwidth() {
		return width;
	}

	/** Set margins.
	    The left margin is always enforced.  The right margin is
	    only enforced when word wrapping is enabled.
	    The margins can only be set right after a paragraph end
	    or linebreak (not enforced). */
	virtual void pushMargins(int right_ = 0, int left_ = 0) {
		margins.push(right);
		margins.push(left);
		right = right_;
		left = left_;
	}

	/// Restore margins
	virtual void popMargins() {
		left = margins.top();
		margins.pop();
		right = margins.top();
		margins.pop();
	}

	/** Indent margins by amount.
	    The numbers relate to how many characters we should
	    eat of the available space.  Passing negative numbers
	    will give more space.
	    See note to pushMargins regarding when to use. */
	virtual void indent(int right_ = 2, int left_=0) {
		pushMargins(right + right_, left - left_);
	}

	/// Dedent margins.  Undoes last indent.
	virtual void dedent() {
		popMargins();
	}

	/** Define the paragraph indentation.
	    Will apply to the next paragraph start.  Therefor,
	    call this before you start a new paragraph. */
	virtual void setParIndent(int c) {
		parindent = c;
	}

protected:
	// Override these to modify the behaviour:

	/** A derived class can get the oppurtunity to modify the
	    line before we spit it out to the file.  This can
	    be used to implement for instance hfill and alignment 
	    support.
	*/
	virtual void processLine(string & line_) 
	{ }

	/// Flush line if applicable (no line break is printed)
	virtual void flushline();

	/** Use tabs to indent?
	    If this is not wanted, override this method. */
	virtual bool usetabs() const {
		return true;
	}

	/// Force a line break
	void linebreak();

	/// Begin a new paragraph
	void parstart();

	/// Write a non-breakable space
	void nbsp();

	/// Get this much space
	string doindent(int);

	/// How much is left of this line?
	int available();

	/// Break this line at the rightmost breakpoint and flush it
	void splitline();

	// Setup information

	/// What end-of-line string is used?
	string eol;

	/// How wide do we allow a line to be?
	int width;

	/// Should we word-wrap?
	bool wordwrap;

	/// Are we allowed to force a break in very long words?
	bool breakwords;

	/// How much padding to the left and right?
	int right, left;

	/// How much indentation on the first line of a new paragraph?
	int parindent;

	// Running information

	/// What line number are we at?
	int lineno;

	/// What column are we at?
	int column;

	/// The current line we are collecting
	string line;

	/// The leftmargin of the current line
	int leftmargin;

	/// A stack of margins
	stack<int> margins;

	/// Keep a list of breakable spaces in the current line
	vector<int> spaces;
};

#endif TEXTFORMAT_H
