// Created by Lars Gullik Bjnnes
// Copyright 1999 Lars Gullik Bjnnes (larsbj@lyx.org)
// Released under the Gnu General Public License

// Implemented and tested on g++ 2.7.2.3

// Primarily developed for use in the LyX Project http://www.lyx.org/
// but should be adaptable to any project.

#ifndef DEBUGSTREAM_H
#define DEBUGSTREAM_H

//#define FSTREAM_HEADER <fstream.h>
#include FSTREAM_HEADER
#include "LString.h"

/** This is a streambuffer that never prints out anything, at least
    that is the intention. You can call it a no-op streambuffer, and
    the ostream that uses it will be a no-op stream.
*/
class nullbuf : public streambuf {
public:
	///
	nullbuf() { setbuf(0,0); }
protected:
	///
	virtual int sync() { return 0; }
	/// 
	virtual streamsize xsputn(char const *, streamsize n) {
		// fakes a purge of the buffer by returning n
		return n;
	}
	///
	virtual int overflow(int c = EOF) {
		// fakes success by returning c
		return c;
	}
};

/** A streambuf that sends the output to cerr and optionally to 
    a file f.
*/
class debugbuf : public streambuf {
public:
	///
	debugbuf() { setbuf(0,0); }
	///
	debugbuf(char const * f)
		: out(f) { 
		setbuf(0, 0); 
		out.rdbuf()->setbuf(0, 0); // in new STL we must use pubsetbuf
	}
protected:
	///
	virtual int sync() { 
		if (out.is_open()) out.flush();
		return 0;
	}
	///
	virtual streamsize xsputn(char const * p, streamsize n) {
		cerr.write(p, n);
		if (out.is_open()) out.write(p, n);
		return n;
	}
	///
	virtual int overflow(int c = EOF) {
		if (c != EOF) {
			cerr.put(c);
			if (out.is_open()) out.put(c);
		}
		return c;
	}
private:
	///
	ofstream out;
};


/** Ideally this should have been a namespace, but since we try to be
    compilable on older C++ compilators too, we use a struct instead.
    This is all the different debug levels that we have.
*/
struct Debug {
	///
	enum type {
		///
		NONE = 0,
		///
		INFO       = (1 << 0),   // 1
		///
		INIT       = (1 << 1),   // 2
		///
		KEY        = (1 << 2),   // 4
		///
		TOOLBAR    = (1 << 3),   // 8
		///
		PARSER     = (1 << 4),   // 16
		///
		LYXRC      = (1 << 5),   // 32
		///
		KBMAP      = (1 << 6),   // 64
		///
		LATEX      = (1 << 7),   // 128
		///
		MATHED     = (1 << 8),   // 256 // Alejandro, please use this.
		///
		FONT       = (1 << 9),   // 512
		///
		TCLASS     = (1 << 10),  // 1024
		///
		LYXVC      = (1 << 11),  // 2048
		///
		LYXSERVER  = (1 << 12),  // 4096
	};
	///
	static const type ANY = type(INFO | INIT | KEY | TOOLBAR |
				     PARSER | LYXRC | KBMAP | LATEX |
				     MATHED | FONT | TCLASS | LYXVC |
				     LYXSERVER);

	/** A function to convert symbolic string names on debug levels
	    to their numerical value.
	*/
	static Debug::type value(string const & val) {
		if (val == "NONE") return Debug::NONE;
		if (val == "INFO") return Debug::INFO;
		if (val == "INIT") return Debug::INIT;
		if (val == "KEY") return Debug::KEY; 
		if (val == "TOOLBAR") return Debug::TOOLBAR;
		if (val == "PARSER") return Debug::PARSER; 
		if (val == "LYXRC") return Debug::LYXRC; 
		if (val == "KBMAP") return Debug::KBMAP;  
		if (val == "LATEX") return Debug::LATEX;  
		if (val == "MATHED") return Debug::MATHED; 
		if (val == "FONT") return Debug::FONT;   
		if (val == "TCLASS") return Debug::TCLASS; 
		if (val == "LYXVC") return Debug::LYXVC;  
		if (val == "LYXSERVER") return Debug::LYXSERVER;
		return Debug::NONE;
	}
};


/** DebugStream is a ostream intended for debug output. It has also support
    for a logfile. Debug output is output to cerr and if the logfile is set,
    to the logfile.

    Example of Usage:
    DebugStream debug;
    debug.level(Debug::INFO);
    debug.debug(Debug::WARN) << "WARN\n";
    debug[Debug::INFO] << "INFO\n";
    debug << "Always\n";

    Will output:
    INFO
    Always

    If you want to have debug output from time critical code you should 
    use this construct:
    if (debug.debugging(Debug::INFO)) {
         debug << "...debug output...\n";
    }
    
    To give debug info even if no debug (NONE) is requested:
    debug << "... always output ...\n";

    To give debug output regardless of what debug level is set (!NONE):
    debug.debug() << "...on debug output...\n";
    debug[Debug::ANY] << "...on debug output...\n";

    To give debug output when a specific debug level is set (INFO):
    debug.debug(Debug::INFO) << "...info...\n";
    debug[Debug::INFO] << "...info...\n";

    To give debug output when either on of debug levels is set (INFO or CRIT):
    debug.debug(Debug::type(Debug::INFO | Debug::CRIT)) << "...info/crit...\n";
    debug[Debug::type(Debug::INFO | Debug::CRIT)] << "...info/crit...\n";

*/
class DebugStream : public ostream {
public:
	/// Constructor, sets the debug level to t.
	DebugStream(Debug::type t = Debug::NONE) 
		: ostream(new debugbuf), dt(t), 
		  nullstream(new nullbuf) {}

	/// Constructor, sets the log file to f, and the debug level to t.
	DebugStream(char const * f, Debug::type t = Debug::NONE)
		: ostream(new debugbuf(f)), dt(t), 
		  nullstream(new nullbuf) {}

	/// Sets the debug level to t.
	void level(Debug::type t) {
		dt = t;
	}

	/// Returns the current debug level.
	Debug::type level() const {
		return dt;
	}

	/// Adds t to the current debug level.
	void addLevel(Debug::type t) {
		dt = Debug::type(dt | t);
	}

	/// Deletes t from the current debug level.
	void delLevel(Debug::type t) {
		dt = Debug::type(dt & ~t);
	}

	/// Sets the debugstreams' logfile to f.
	void logFile(char const * f) {
		if (f) {
			streambuf * old = rdbuf(new debugbuf(f));
			delete old;
		}
	}

	/// Returns true if t is part of the current debug level.
	bool debugging(Debug::type t = Debug::ANY) const {
		if (dt & t) return true;
		return false;
	}

	/** Returns the no-op stream if t is not part of the
	    current debug level otherwise the real debug stream
	    is used.
	*/
	ostream & debug(Debug::type t = Debug::ANY) {
		if (dt & t) return *this;
		return nullstream;
	}

	/** This is an operator to give a more convenient use:
	    dbgstream[Debug::INFO] << "Info!\n";
	    Besides that it is equal to debug except that it has
	    not a default argument.
	*/
	ostream & operator[](Debug::type t) {
		return debug(t);
	}

private:
	/// The current debug level
	Debug::type dt;
	/// The no-op stream.
	ostream nullstream;
};

extern DebugStream lyxerr;

//#define TEST_DEBUGSTREAM
#ifdef TEST_DEBUGSTREAM

// Example debug stream
DebugStream debugstream;

int main(int, char**)
{
  // I have been running some tests on this to see how much overhead
  // this kind of permanent debug code has. My conclusion is: not 
  // much. In all, but the most time critical code, this will have 
  // close to no impact at all.

  // In the tests that I have run the use of
  // if (debugstream.debugging(DebugStream::INFO))
  //    debugstream << "some debug\n";
  // has close to no overhead when the debug level is not 
  // DebugStream::INFO.

  // The overhead for
  //  debugstream.debug(DebugStream::INFO) << "some debug\n";
  // is also very small when the debug level is not
  // DebugStream::INFO. However the overhead for this will increase
  // if complex debugging information is output.

  // The overhead when the debug level is DebugStream::INFO can be
  // significant, but since we then are running in debug mode it is 
  // of no concern.

  // Why should we use this instead of the class Error that we already
  // have? First of all it uses C++ iostream and constructs, secondly
  // it will be a lot easier to output the debug info that we need
  // without a lot of manual conversions, thirdly we can now use 
  // iomanipulators and the complete iostream formatting functions.
  // pluss it will work for all types that have a operator<< 
  // defined, and can be used in functors that take a ostream& as
  // parameter. And there should be less need for temporary objects.
  // And one nice bonus is that we get a log file almost for
  // free.
  
  // Some of the names are of course open to modifications. I will try
  // to use the names we already use in LyX.

  // Just a few simple debugs to show how it can work.
  debugstream << "Debug level set to Debug::NONE\n";
  if (debugstream.debugging()) {
	  debugstream << "Something must be debugged\n";
  }
  debugstream.debug(Debug::WARN) << "more debug(WARN)\n";
  debugstream.debug(Debug::INFO) << "even more debug(INFO)\n";
  debugstream.debug(Debug::CRIT) << "even more debug(CRIT)\n";
  debugstream.level(Debug::value("INFO"));
  debugstream << "Setting debug level to Debug::INFO\n";
  if (debugstream.debugging()) {
	  debugstream << "Something must be debugged\n";
  }
  debugstream.debug(Debug::WARN) << "more debug(WARN)\n";
  debugstream.debug(Debug::INFO) << "even more debug(INFO)\n";
  debugstream.debug(Debug::CRIT) << "even more debug(CRIT)\n";
  debugstream.addLevel(Debug::type(Debug::CRIT | 
				   Debug::WARN));
  debugstream << "Adding Debug::CRIT and Debug::WARN\n";
  debugstream[Debug::WARN] << "more debug(WARN)\n";
  debugstream[Debug::INFO] << "even more debug(INFO)\n";
  debugstream[Debug::CRIT] << "even more debug(CRIT)\n";
  debugstream.delLevel(Debug::INFO);
  debugstream << "Removing Debug::INFO\n";
  debugstream[Debug::WARN] << "more debug(WARN)\n";
  debugstream[Debug::INFO] << "even more debug(INFO)\n";
  debugstream[Debug::CRIT] << "even more debug(CRIT)\n";
  debugstream.debug(string("a string with debug"), Debug::INFO);
  debugstream.print(string("some more string debug"));
  debugstream.logFile("logfile");
  debugstream << "Setting logfile to \"logfile\"\n";
  debugstream << "Value: " << 123 << " " << "12\n";
  int i = 0;
  int *p = new int;
  // note: the (void*) is needed on g++ 2.7.x since it does not
  // support partial specialization. In egcs this should not be
  // needed.
  debugstream << "automatic " << (void*)&i 
	      << ", free store " << (void*)p << '\n';
}
#endif
#endif
