// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *
 *======================================================*/
#ifndef _LYXINSET_H
#define _LYXINSET_H

#ifdef __GNUG__
#pragma interface
#endif

#include "config.h"
#include <stdio.h>
#include <X11/Xlib.h>
#include "gettext.h"
#include "LString.h"

class Buffer;
class BufferView;
class LyXFont;
class LyXLex;
class Painter;
class LaTeXFeatures;

/// Insets
class Inset {
public:
	/** This is not quite the correct place for this enum, but it is
	  better than definitions.h. I think the correct would be to let
	  each subclass of Inset declare its own enum code. Actually the
	  notion of an Inset::Code should be avoided, but I am not sure how
	  this could be done in a cleaner way. */
	enum Code {
		///
		NO_CODE,
		///
		TOC_CODE,  // table of contents
		///
		LOF_CODE,  // list of figures
		///
		LOT_CODE, // list of tables
		///
		LOA_CODE, // list of algorithms
		///
		QUOTE_CODE, // quote inset
		///
		MARK_CODE, // ???
		///
		REF_CODE, // reference inset
		///
		URL_CODE, // url inset
		///
		HTMLURL_CODE, // html inset
		///
		SEPARATOR_CODE, // ???
		///
		ENDING_CODE, // ???
		///
		LABEL_CODE, // reference label
		///
		IGNORE_CODE, // ???
		///
		ACCENT_CODE, // latex accent
		///
		MATH_CODE, // math inset (mathed)
		///
		INDEX_CODE, // index inset
		///
		INCLUDE_CODE, // include inset
		///
		GRAPHICS_CODE, // figure inset ??
		///
		PARENT_CODE, // parent inset 
		///
		BIBTEX_CODE, // bibtex inset ?
		/// Identifies citation insets for use by Communicator::bibitemUpdate
		CITATION_CODE,
		///
		FLOAT_CODE, // float insets
		///
		FOOT_CODE, // footnote
                ///
                TABULAR_CODE, // tabular insets
                ///
                TEXT_CODE, // text insets
		///
		MARGIN_CODE, // Margin note
		///
		NOTE_CODE  // Old note inset
	};

	enum EDIT {
		NOT_EDITABLE,
		EDITABLE,
		HIGHLY_EDITABLE,
		EDITABLE_NO_UNDO
	};
	
	///
	virtual ~Inset(){};

	/// How far above the baseline do we go?
	virtual int ascent(Painter &, LyXFont const &) const=0;
	/// How far below the baseline do we go?
	virtual int descent(Painter &, LyXFont const &) const=0;
	/// How wide is this inset?
	virtual int width(Painter &, LyXFont const &) const=0;
	/// Painter encapsulates all graphics parameters
	virtual void draw(Painter &, LyXFont const &, int baseline, float & x)=0;

	///
	virtual LyXFont ConvertFont(LyXFont font);
	///
	/// what appears in the minibuffer when opening
	virtual string EditMessage() const {return _("Opened inset");}
	///
	virtual void Edit(int x, int y, unsigned int button, BufferView *);
	///
	virtual Inset::EDIT Editable() const;
	///
	virtual bool AutoDelete() const;
	///
	virtual void Write(FILE *file)=0;
	///
	virtual void Read(LyXLex &lex)=0;
	/** returns the number of rows (\n's) of generated tex code.
	 fragile != 0 means, that the inset should take care about
	 fragile commands by adding a \protect before.
	 */
	virtual int Latex(FILE *file, signed char fragile)=0;
	virtual int Latex(string &file, signed char fragile)=0;
	/// Updates needed features for this inset.
	virtual void Validate(LaTeXFeatures &features) const;
	///
	virtual bool Deletable() const;

	/// returns LyX code associated with the inset. Used for TOC, ...)
	virtual Inset::Code LyxCode() const = 0;
  
	/// Get the label that appears at screen
	virtual string getLabel(int) const {
		return string();
	}

	///
	virtual Inset* Clone()=0;

	/// Returns true if the inset should be centered alone
	virtual bool display() const { return false; }  
 
	/// Changes the display state of the inset
	virtual void SetDisplay(bool) {  }  
	///
	virtual int GetNumberOfLabels() const {
		return 0;
	}
};




//  Updatable Insets. These insets can be locked and receive
//  directly user interaction. Currently used only for mathed.
//  Note that all pure methods from Inset class are pure here too.
//  [Alejandro 080596] 

/** Extracted from Matthias notes:
  * 
  * An inset can simple call LockInset in it's edit call and *ONLY* 
  * in it's edit call.
  *
  * Unlocking is either done by LyX or the inset itself with a 
  * UnlockInset-call
  *
  * During the lock, all button and keyboard events will be modified
  * and send to the inset through the following inset-features. Note that
  * Inset::InsetUnlock will be called from inside UnlockInset. It is meant
  * to contain the code for restoring the menus and things like this.
  * 
  * If a inset wishes any redraw and/or update it just has to call
  * UpdateInset(this).
  * 
  * It's is completly irrelevant, where the inset is. UpdateInset will
  * find it in any paragraph in any buffer. 
  * Of course the_locking_inset and the insets in the current paragraph/buffer
  *  are checked first, so no performance problem should occur.
  */
class UpdatableInset: public Inset {
public:
	/** Dispatch result codes
	    Now that nested updatable insets are allowed, the local dispatch
	    becomes a bit complex, just two possible results (boolean)
	    are not enough. 
	 
	    DISPATCHED   = the inset catched the action
	    FINISHED     = the inset must be unlocked as a result
	                   of the action
	    UNDISPATCHED = the action was not catched, it should be
	                   dispatched by lower level insets
	 
	    [Alejandro 19981105] 
	 */ 
	enum RESULT {
	    UNDISPATCHED=0,
	    DISPATCHED,
	    FINISHED
	};
    
        /// To convert old binary dispatch results
        RESULT DISPATCH_RESULT(bool b) {
	    return (b) ? DISPATCHED: FINISHED;
	}
    
	///
	UpdatableInset() { scx = mx_scx = 0; }
	///
	virtual ~UpdatableInset() { };
	///
	virtual Inset::EDIT Editable() const;
   
	/// may call ToggleLockedInsetCursor
	virtual void ToggleInsetCursor();
	///
	virtual void GetCursorPos(int&, int&) { }
	///
	virtual int InsetInInsetY() { return 0; }
	///
	virtual void InsetButtonPress(int x, int y, int button);
	///
	virtual void InsetButtonRelease(int x, int y, int button);
	
	///
	virtual void InsetKeyPress(XKeyEvent *ev);
	///
	virtual void InsetMotionNotify(int x, int y, int state);
	///
	virtual void InsetUnlock();
   
	///  An updatable inset could handle lyx editing commands
	virtual RESULT LocalDispatch(int, char const*);
	//
	bool isCursorVisible() const { return cursor_visible; }
	///
	virtual void Edit(int, int, unsigned int, BufferView *);
	///
	virtual void draw(Painter &, LyXFont const &, int baseline, float & x);
	///
	virtual void SetFont(LyXFont const &) { return; }
	///
	virtual bool InsertInset(Inset *) { return false; }
	///
	virtual UpdatableInset *GetLockingInset() { return this; }
	///
	virtual bool UpdateInsetInInset(BufferView*, Inset*) { return false; }
	///
	virtual bool UnlockInsetInInset(Inset*,bool /*lr*/=false) { return false; }
	///
        virtual int getMaxWidth(UpdatableInset*) const { return -1; }

protected:
	///
	bool cursor_visible;
	///
	virtual void scroll(float);
	///
	int
	        top_x,
		top_baseline;
	BufferView *bview;
	///
	virtual void UpdateLocal(bool flag=true);

private:
	///
	int mx_scx, scx;
};


#endif
