// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *          Copyright (C) 1995-1999 The LyX Team.
 *
 *======================================================*/

// Wrap string is a wrapper around a C++ STD basic_string, it is provided
// so that there can be two different types based on the same basic_string.
// A typedef would not provide that.
// Lgb

#ifndef WRAPSTRING_H
#define WRAPSTRING_H 


template <class String>
class wrap_string {
  private:
  String str_;
public:
	/**@name Typedefs */
	//@{
	
	///
	typedef typename String::size_type size_type;

	///
	typedef typename String::value_type value_type;

	///
	typedef typename String::pointer pointer;
	///
	typedef typename String::const_reference const_reference;
	///
	typedef typename String::reference reference;
	///
	typedef typename String::iterator iterator;
	///
	typedef typename String::const_iterator const_iterator;
#if 0
	///
	typedef typename String::reverse_iterator reverse_iterator;
	///
	typedef typename String::reverse_iterator const_reverse_iterator;

#endif
	//@}

	///
	iterator begin() { return str_.begin(); }
	///
	const_iterator begin() const { return str_.begin(); }
	///
	iterator end() { return str_.end(); }
	///
	const_iterator end() const { return str_.end(); }
#if 0
	///
	reverse_iterator rbegin() { return str_.rbegin(); }
	///
	const_reverse_iterator rbegin() const { return str_.rbegin(); }
	///
	reverse_iterator rend() { return str_.rend(); }
	///
	const_reverse_iterator rend() const { return str_.rend(); }
#endif
	/**@name Constructors and Deconstructors. wrap_string has the same
	  constructors as STL, except the one using iterators, one other
	  difference is that wrap_string, do not allow the optional allocator
	  parameter. */
	//@{
	
	/// "all characters" marker
	static const size_type npos = static_cast<size_type>(-1);

	/// #wrap_string x;# 
	  wrap_string() {}
	
	/// #wrap_string x(wrap_string ...)# 
	wrap_string(wrap_string const & ws,
		    size_type pos = 0,
		    size_type n = npos)
	  : str_(ws.str_, pos, n)
	  {}
	
	/// #wrap_string x("abc", 2) -> "ab"#
	    wrap_string(value_type const * cc, size_type n)
	      : str_(cc, n) {}
	
	/// #wrap_string x("abc")#
	wrap_string(value_type const * cc)
	  : str_(cc) {}
	
	/// wrap_string(5, 'n') -> "nnnnn"
	wrap_string(size_type n, value_type c)
	  : str_(n, c) {}

	///
	wrap_string(iterator first, iterator last)
	  : str_(first, last) {}

	///
	~wrap_string() {}

	//@}

	
	/**@name Size and Capacity */
	//@{
	
	/// number of characters
	size_type size() const { return str_.size(); }

	/// largest possible string
	size_type max_size() const { return str_.max_size(); }

	///
	size_type length() const { return str_.length(); }

	///
	bool empty() const { return str_.empty(); }

	///
	void resize(size_type n, value_type c) { str_.resize(n, c); }
	
	///
	void resize(size_type n) { resize(n, ' '); }
	
	/// size of the memory in number of elements allocated
	   size_type capacity() const { return str_.capacity(); }

	///
	   void reserve(size_type res_arg = 0) { str_.reserve(res_arg); }
	
	//@}


	/**@name Assignment */
	//@{

	///
	   wrap_string & operator=(wrap_string const & ws)
	     {
	       str_.operator=(ws.str_);
	       return *this;
	     }
	
	///
	wrap_string & operator=(value_type const * cc)
	  {
	    str_.operator=(cc);
	    return *this;
	  }
	
	///
	wrap_string & operator=(value_type c)
	  {
	    str_.operator=(c);
	    return *this;
	  }

	///
	wrap_string & assign(wrap_string const & ws)
	  {
	    str_.assign(ws.str_);
	    return *this;
	  }
	
	///
	wrap_string & assign(wrap_string const & ws,
			     size_type pos, size_type n)
	  {
	    str_.assign(ws.str_, pos, n);
	    return *this;
	  }
	
	///
	wrap_string & assign(value_type const * p, size_type n)
	  {
	    str_.assign(p, n);
	    return *this;
	  }
	
	///
	wrap_string & assign(value_type const * p)
	  {
	    str_.assign(p);
	    return *this;
	  }
	
	///
	wrap_string & assign(size_type n, value_type c)
	  {
	    str_.assign(n, c);
	    return *this;
	  }

	///
	wrap_string & assign(iterator first, iterator last)
	  {
	    str_.assign(first, last);
	    return *this;
	  }
	
	//@}

	/**@name Element Access. Since wrap_string does not use exceptions,
	  an abort is called on out-of-range access to at(). */
	
	/// unchecked access
	const_reference operator[](size_type s) const
	  {
	    return str_.operator[](s);
	  }
	
	/// unchecked access
	reference operator[](size_type s)
	  {
	    return str_.operator[](s);
	  }
	
	/// checked access
	const_reference at(size_type s) const
	  {
	    return str_.at(s);
	  }
	
	/// checked access
	reference at(size_type s)
	  {
	    return str_.at(s);
	  }

	//@}

	
	/**@name Insert */
	//@{
	
	// add characters after (*this)[length()-1]:
	
	///
	wrap_string & operator+=(wrap_string const & ws)
	  {
	    str_.operator+=(ws.str_);
	    return *this;
	  }
	
	///
	wrap_string & operator+=(value_type const * cc)
	  {
	    str_.operator+=(cc);
	    return *this;
	  }
	
	///
	wrap_string & operator+=(value_type c)
	  {
	    str_.operator+=(c);
	    return *this;
	  }

	///
	void push_back(value_type c)
	  {
	    str_.push_back(c);
	  }

	///
	wrap_string & append(wrap_string const & ws)
	  {
	    str_.append(ws.str_);
	    return *this;
	  }

	///
	wrap_string & append(wrap_string const & ws,
			     size_type pos, size_type n)
	  {
	    str_.append(ws.str_, pos, n);
	    return *this;
	  }

	///
	wrap_string & append(value_type const * cc, size_type n)
	  {
	    str_.append(cc, n);
	    return *this;
	  }

	///
	wrap_string & append(value_type const * cc)
	  {
	    str_.append(cc);
	    return *this;
	  }

	///
	wrap_string & append(size_type n, value_type c)
	  {
	    str_.append(n, c);
	    return *this;
	  }

	///
	wrap_string & append(iterator first, iterator last)
	  {
	    str_.append(first, last);
	    return *this;
	  }
	
	// insert characters before (*this)[pos]:

	///
	wrap_string & insert(size_type pos, wrap_string const & ws)
	  {
	    str_.insert(pos, ws.str_);
	    return *this;
	  }
	
	///
	wrap_string & insert(size_type pos, wrap_string const & ws,
			size_type pos2, size_type n)
	  {
	    str_.insert(pos, ws.str_, pos2, n);
	    return *this;
	  }
	
	///
	wrap_string & insert(size_type pos, value_type const * p, size_type n)
	  {
	    str_.insert(pos, p, n);
	    return *this;
	  }

	///
	wrap_string & insert(size_type pos, value_type const * p)
	  {
	    str_.insert(pos, p);
	    return *this;
	  }

	///
	wrap_string & insert(size_type pos, size_type n, value_type c)
	  {
	    str_.insert(pos, n, c);
	    return *this;
	  }

	// insert characters before p

	///
	iterator insert(iterator p, value_type c)
	  {
	    return str_.insert(p, c);
	  }

	///
	void insert(iterator p, size_type n , value_type c)
	  {
	    str_.insert(p, n, c);
	  }

	///
	void insert(iterator p, iterator first, iterator last)
	  {
	    str_.insert(p, first, last);
	  }
	
	//@}

	/**@name Find */
	//@{

	///
	size_type find(wrap_string const & ws, size_type i = 0) const
	  {
	    return str_.find(ws.str_, i);
	  }
	
	///
	size_type find(value_type const * p, size_type i, size_type n) const
	  {
	    return str_.find(p, i, n);
	  }
	
	///
	size_type find(value_type const * p, size_type i = 0) const
	  {
	    return str_.find(p, i);
	  }
	
	///
	size_type find(value_type c, size_type i = 0) const
	  {
	    return str_.find(c, i);
	  }

	///
	size_type rfind(wrap_string const & ws, size_type i = npos) const
	  {
	    return str_.rfind(ws.str_, i);
	  }
	
	///
	size_type rfind(value_type const * p, size_type i, size_type n) const
	  {
	    return str_.rfind(p, i, n);
	  }
	
	///
	size_type rfind(value_type const * p, size_type i = npos) const
	  {
	    return str_.rfind(p, i);
	  }
	
	///
	size_type rfind(value_type c, size_type i = npos) const
	  {
	    return str_.rfind(c, i);
	  }

	///
	size_type find_first_of(wrap_string const & ws, size_type i = 0) const
	  {
	    return str_.find_first_of(ws.str_, i);
	  }
	
	///
	size_type find_first_of(value_type const * p, size_type i, size_type n) const
	  {
	    return str_.find_first_of(p, i, n);
	  }
	
	///
	size_type find_first_of(value_type const * p, size_type i = 0) const
	  {
	    return str_.find_first_of(p, i);
	  }
	
	///
	size_type find_first_of(value_type c, size_type i = 0) const
	  {
	    return str_.find_first_of(c, i);
	  }

	///
	size_type find_last_of(wrap_string const & ws,
			       size_type i = npos) const
	  {
	    return str_.find_last_of(ws.str_, i);
	  }
	
	///
	size_type find_last_of(value_type const * p, size_type i, size_type n) const
	  {
	    return str_.find_last_of(p, i, n);
	  }
	
	///
	size_type find_last_of(value_type const * p, size_type i = npos) const
	  {
	    return str_.find_last_of(p, i);
	  }
	
	///
	size_type find_last_of(value_type c, size_type i = npos) const
	  {
	    return str_.find_last_of(c, i);
	  }

	///
	size_type find_first_not_of(wrap_string const & ws,
				    size_type i = 0) const
	  {
	    return str_.find_first_not_of(ws.str_, i);
	  }
	
	///
	size_type find_first_not_of(value_type const * p, size_type i,
				    size_type n) const
	  {
	    return str_.find_first_not_of(p, i, n);
	  }
	
	///
	size_type find_first_not_of(value_type const * p, size_type i = 0) const
	  {
	    return str_.find_first_not_of(p, i);
	  }
	
	///
	size_type find_first_not_of(value_type c, size_type i = 0) const
	  {
	    return str_.find_first_not_of(c, i);
	  }

	///
	size_type find_last_not_of(wrap_string const & ws,
				   size_type i = npos) const
	  {
	    return str_.find_last_not_of(ws.str_, i);
	  }
	
	///
	size_type find_last_not_of(value_type const * p, size_type i,
				   size_type n) const
	  {
	    return str_.find_last_not_of(p, i, n);
	  }
	
	///
	size_type find_last_not_of(value_type const * p, size_type i = npos) const
	  {
	    return str_.find_last_not_of(p, i);
	  }
	
	///
	size_type find_last_not_of(value_type c, size_type i = npos) const
	  {
	    return str_.find_last_not_of(c, i);
	  }

	//*}

	
	/**@name Replace */
	//@{

	// replace [(*this)[i], (*this)[i+n]] with other value_typeacters:

	///
	wrap_string & replace(size_type i,size_type n, wrap_string const & str)
	  {
	    str_.replace(i, n, str.str_);
	    return *this;
	  }

	///
	wrap_string & replace(size_type i,size_type n, wrap_string const & s,
			  size_type i2, size_type n2)
	  {
	    str_.replace(i, n, s.str_, i2, n2);
	    return *this;
	  }

	///
	wrap_string & replace(size_type i, size_type n, value_type const * p,
			  size_type n2)
	  {
	    str_.replace(i, n, p, n2);
	    return *this;
	  }

	///
	wrap_string & replace(size_type i, size_type n, value_type const * p)
	  {
	    str_.replace(i, n, p);
	    return *this;
	  }

	///
	wrap_string & replace(size_type i, size_type n, size_type n2, value_type c)
	  {
	    str_.replace(i, n, n2, c);
	    return *this;
	  }

	///
	wrap_string & replace(iterator i, iterator i2, const wrap_string & str)
	  {
	    str_.replace(i, i2, str.str_);
	    return *this;
	  }

	///
	wrap_string & replace(iterator i, iterator i2,
			  value_type const * p, size_type n)
	  {
	    str_.replace(i, i2, p, n);
	    return *this;
	  }

	///
	wrap_string & replace(iterator i, iterator i2, value_type const * p)
	  {
	    str_.replace(i, i2, p);
	    return *this;
	  }

	///
	wrap_string & replace(iterator i, iterator i2,
			      size_type n , value_type c)
	  {
	    str_.replace(i, i2, n, c);
	    return *this;
	  }
	
	///
	wrap_string & replace(iterator i, iterator i2,
			      iterator j, iterator j2)
	  {
	    str_.replace(i, i2, j, j2);
	    return *this;
	  }
	
	/// Erase n value_types from position i.
	wrap_string & erase(size_type i = 0, size_type n = npos)
	  {
	    str_.erase(i, n);
	    return *this;
	  }

	///
	iterator erase(iterator i)
	  {
	    return str_.erase(i);
	  }

	///
	iterator erase(iterator first, iterator last)
	  {
	    return str_.erase(first, last);
	  }

	//@}

	
	/**@name Conversion to C-style Strings */
	//@{
	
	/// 
	value_type const *c_str() const
	  {
	    return str_.c_str();
	  }

	/** Note that this is STL compilant, so you can not assume
	  that the returned array has a trailing '\0'. */
	value_type const *data() const
	  {
	    return str_.data();
	  }

	/** This one returns a verbatim copy. Not the trailing '\0'
	  The caller must provide a buffer with engough room.
	  */
	size_type copy(value_type * buf, size_type len, size_type pos = 0) const
	  {
	    return str_.copy(buf, len, pos);
	  }

	//@}

	
	/**@name Comparisons. Combined > and == */
	//@{
	
	///
	int compare(wrap_string const & str) const
	  {
	    return str_.compare(str.str_);
	  }

	///
	int compare(value_type const * p) const
	  {
	    return str_.compare(p);
	  }

	///
	int compare(size_type pos, size_type n, wrap_string const & str) const
	  {
	    return str_.compare(pos, n, str.str_);
	  }

	///
	int compare(size_type pos, size_type n, wrap_string const & str,
		    size_type pos2, size_type n2) const
	  {
	    return str_.compare(pos, n, str.str_, pos2, n2);
	  }

	///
	int compare(size_type pos, size_type n, value_type const * p,
		    size_type n2 = npos) const
	  {
	    return str_.compare(pos, n, p, n2);
	  }
	
	//@}

	
	
	/**@name Substrings */
	//@{

	///
	wrap_string substr(size_type i = 0, size_type n = npos) const
	  {
	    wrap_string<String> st;
	    st.str_ = str_.substr(i, n);
	    return st;
	  }
	
	//@}

private:
	///
	wrap_string & operator+(int);
	///
	wrap_string & operator=(int);
};

// The usual comparison operators ==, !=, >, <, >=, <= are
// provided for wrap_strings

template <class S>
bool operator==(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) == 0;
}

template <class S>
bool operator==(wrap_string<S>::value_type const * a , wrap_string<S> const & b)
{
	return b.compare(a) == 0;
}

template <class S>
bool operator==(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) == 0;
}


template <class S>
bool operator!=(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) != 0;
}

template <class S>
bool operator!=(wrap_string<S>::value_type const * a, wrap_string<S> const & b)
{
	return b.compare(a) != 0;
}

template <class S>
bool operator!=(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) != 0;
}


template <class S>
bool operator>(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) > 0;
}

template <class S>
bool operator>(wrap_string<S>::value_type const * a, wrap_string<S> const & b)
{
	return b.compare(a) < 0; // args reversed
}

template <class S>
bool operator>(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) > 0;
}


template <class S>
bool operator<(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) < 0;
}

template <class S>
bool operator<(wrap_string<S>::value_type const * a, wrap_string<S> const & b)
{
	return b.compare(a) > 0; // args reversed
}

template <class S>
bool operator<(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) < 0;
}


template <class S>
bool operator>=(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) >= 0;
}

template <class S>
bool operator>=(wrap_string<S>::value_type const * a, wrap_string<S> const & b)
{
	return b.compare(a) <= 0; // args reversed
}

template <class S>
bool operator>=(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) >= 0;
}


template <class S>
bool operator<=(wrap_string<S> const & a, wrap_string<S> const & b)
{
	return a.compare(b) <= 0;
}

template <class S>
bool operator<=(wrap_string<S>::value_type const * a, wrap_string<S> const & b)
{
	return b.compare(a) >= 0; // args reversed
}

template <class S>
bool operator<=(wrap_string<S> const & a, wrap_string<S>::value_type const * b)
{
	return a.compare(b) <= 0;
}


// Concatenation
template <class S>
wrap_string<S> operator+(wrap_string<S> const &a,
			 wrap_string<S> const &b)
{
	wrap_string<S> ws(a);
	ws.append(b);
	return ws;
}

template <class S>
wrap_string<S> operator+(wrap_string<S>::value_type const *a, wrap_string<S> const &b)
{
	wrap_string<S> ws(a);
	ws.append(b);
	return ws;
}

template <class S>
wrap_string<S> operator+(wrap_string<S>::value_type a, wrap_string<S> const &b)
{
	wrap_string<S> ws;
	ws = a;
	ws.append(b);
	return ws;
}

template <class S>
wrap_string<S> operator+(wrap_string<S> const &a, wrap_string<S>::value_type const *b)
{
	wrap_string<S> ws(a);
	ws.append(b);
	return ws;
}

template <class S>
wrap_string<S> operator+(wrap_string<S> const &a, wrap_string<S>::value_type b)
{
	wrap_string<S> ws(a);
	ws +=b;
	return ws;
}

#endif
