/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *          Copyright (C) 1995-1998 The LyX Team.
 *
 *======================================================*/

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "insetcommand.h"
#include "painter.h"
#include "lyxlex.h"
#include "debug.h"
#include "lyxfont.h"

// 	$Id: insetcommand.C,v 1.9 1998/10/22 16:56:12 larsbj Exp $	

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: insetcommand.C,v 1.9 1998/10/22 16:56:12 larsbj Exp $";
#endif /* lint */


InsetCommand::InsetCommand()
{
}


InsetCommand::InsetCommand(string const & cmd, string const & arg, 
			   string const & opt)
	: command(cmd), options(opt), contents(arg)
{
}


InsetCommand::~InsetCommand()
{
}


int InsetCommand::ascent(Painter & pain, LyXFont const &) const
{
	LyXFont font(LyXFont::ALL_SANE);
	font.decSize();
	
	int width, ascent, descent;
        string s = getScreenLabel();

        if (Editable()) {
		pain.buttonText(0, 0, s, font,
				false, width, ascent, descent);
	} else {
		pain.rectText(0, 0, s, font,
			      LColor::commandbg, LColor::commandframe,
			      false, width, ascent, descent);
	}
	return ascent;
}


int InsetCommand::descent(Painter & pain, LyXFont const &) const
{
	LyXFont font(LyXFont::ALL_SANE);
	font.decSize();
	
	int width, ascent, descent;
        string s = getScreenLabel();

        if (Editable()) {
		pain.buttonText(0, 0, s, font,
				false, width, ascent, descent);
	} else {
		pain.rectText(0, 0, s, font,
			      LColor::commandbg, LColor::commandframe,
			      false, width, ascent, descent);
	}
	return descent;
}


int InsetCommand::width(Painter & pain, LyXFont const &) const
{
	LyXFont font(LyXFont::ALL_SANE);
	font.decSize();
	
	int width, ascent, descent;
        string s = getScreenLabel();

        if (Editable()) {
		pain.buttonText(0, 0, s, font,
				false, width, ascent, descent);
	} else {
		pain.rectText(0, 0, s, font,
			      LColor::commandbg, LColor::commandframe,
			      false, width, ascent, descent);
	}
	return width;
}


void InsetCommand::draw(Painter &pain, LyXFont const &,
			int baseline, float &x)
{
	// Draw it as a box with the LaTeX text
	LyXFont font(LyXFont::ALL_SANE);
	font.setColor(LColor::command);
	font.decSize();
	
	int width;
        string s = getScreenLabel();

        if (Editable()) {
		pain.buttonText(int(x), baseline, s, font,
				true, width);
	} else {
		pain.rectText(int(x), baseline, s, font,
			      LColor::commandbg, LColor::commandframe,
			      true, width);
	}

	x +=  width;
}


// In lyxf3 this will be just LaTeX
void InsetCommand::Write(FILE *file)
{
	fprintf(file, "\\begin_inset LatexCommand %s\n\\end_inset\n",
		getCommand().c_str());
}


void InsetCommand::scanCommand(string const &cmd)
{
	string tcommand, toptions, tcontents;

	if (cmd.empty()) return;

	enum { WS, Command, Option, Content } state = WS;
	
	// Used to handle things like \command[foo[bar]]{foo{bar}}
	int nestdepth = 0;

	for (unsigned int i=0; i<cmd.length(); ++i) {
		char c = cmd[i];
		if ((state==Command && c == ' ') ||
		    (state==Command && c == '[') ||
		    (state==Command && c == '{')) {
			state = WS;
		}
		if ((state==Option  && c == ']') ||
		    (state==Content && c == '}')) {
			if (nestdepth==0) {
				state = WS;
			} else {
				--nestdepth;
			}
		}
		if ((state==Option  && c == '[') ||
		    (state==Content && c == '{')) {
		    	++nestdepth;
		}
		switch (state) {
		case Command:	tcommand += c; break;
		case Option: 	toptions += c; break;
		case Content:	tcontents += c; break;
		case WS:
			if (c == '\\') {
				state = Command;
			} else if (c == '[') {
				state = Option;
				nestdepth = 0; // Just to be sure
			} else if (c == '{') {
				state = Content;
				nestdepth = 0; // Just to be sure
			}
			break;
		}
	}

	// Don't mess with this.
	if (!tcommand.empty()) command = tcommand;
	if (!toptions.empty()) options = toptions;
	if (!tcontents.empty()) setContents(tcontents); 
                	// setContents is overloaded in InsetInclude

	if (lyxerr.debugging(Debug::PARSER))
		lyxerr << "Command <" << cmd
		       << "> == <" << getCommand()
		       << "> == <" << getCmdName()
		       << '|' << getContents()
		       << '|' << getOptions() << ">\n";
}


// This function will not be necessary when lyx3
void InsetCommand::Read(LyXLex &lex)
{    
	if (lex.EatLine()) {
		string t = lex.GetString();
		scanCommand(t);
	} else {
		lex.printError("InsetCommand: Parse error: `$$Token'");
	}
	// step \end_inset
	lex.nextToken();
	if (lex.GetString() != "\\end_inset") {
		lex.printError("Missing \\end_inset at this point."
			       " Read: `$$Token'");
	}
}


int InsetCommand::Latex(FILE *file, signed char /*fragile*/)
{
	fprintf(file, "%s", getCommand().c_str());
	return 0;
}


int InsetCommand::Latex(string &file, signed char /*fragile*/)
{
	file += getCommand();
	return 0;
}


Inset* InsetCommand::Clone()
{
	InsetCommand *result = new InsetCommand(command, contents, options);
	return result;
}


string InsetCommand::getCommand() const
{	
	string s;
	if (!command.empty()) s += "\\"+command;
	if (!options.empty()) s += "["+options+']';
	if (!contents.empty()) s += "{"+contents+'}';
	return s;
}
