#include <config.h>

#include <stdlib.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "insetinclude.h"
#include "bufferlist.h"
#include "debug.h"
#include "LyXView.h"
#include "LaTeXFeatures.h"
#include "gettext.h"
#include "FileInfo.h"
#include "textclasslist.h"

extern BufferList bufferlist;


InsetInclude::InsetInclude(string const & fname, Buffer *bf)
	: InsetCommand("include") 
{
	master = bf;
	setContents(fname);
	flag = InsetInclude::INCLUDE;
	noload = false;
}


InsetInclude::~InsetInclude()
{
	gone();
}


string InsetInclude::getMasterFilename() const {
	return master->getFileName();
}


void InsetInclude::Edit(int, int, unsigned int, BufferView *bv)
{
	bv->getOwner()->getPopups().showFormInclude(this);
}


void InsetInclude::Write(FILE *file)
{
	fprintf(file, "\\begin_inset Include %s\n\\end_inset\n",
		getCommand().c_str());
}


void InsetInclude::Read(LyXLex &lex)
{
	InsetCommand::Read(lex);
    
	if (getCmdName()=="include")
		setInclude();
	else if (getCmdName() == "input")
		setInput();
	else if (contains(getCmdName(), "verbatim")) {
		setVerb();
		if (getCmdName() == "verbatiminput*")
			setVisibleSpace(true);
	}
}


string InsetInclude::getScreenLabel() const
{
	string temp;
	if (isInput())
		temp += _("Input");
	else if (isVerb()) {
		temp += _("Verbatim Input");
		if (isVerbVisibleSpace()) temp += '*';
	} else temp += _("Include");
	temp += ": ";
	
	if (contents.empty()) {
		temp+="???";
	} else {
		temp+=contents;
	}
	return temp;
}


bool InsetInclude::loadIfNeeded() const
{
	if (isNoLoad() || isVerb()) return false;
	if (!IsLyXFilename(getFileName())) return false;
	
	if (bufferlist.exists(getFileName())) return true;
	
	// the readonly flag can/will be wrong, not anymore I think.
	FileInfo finfo(getFileName());
	bool ro = !finfo.writable();
	return ( bufferlist.readFile(getFileName(), ro) != 0 );
}


int InsetInclude::Latex(FILE *file, signed char /*fragile*/)
{
	string include_file;
	signed char dummy = 0;
	Latex(include_file, dummy);
	fprintf(file, "%s", include_file.c_str());
	return 0;
}


int InsetInclude::Latex(string &file, signed char /*fragile*/)
{
	string writefile, incfile;

	// Do nothing if no file name has been specified
	if (contents.empty())
		return 0;
 
	// Use += to force a copy of contents (JMarc)
	incfile += contents;

	if (loadIfNeeded()) {
		Buffer *tmp = bufferlist.getBuffer(getFileName());
		
		if (tmp->params.textclass != master->params.textclass) {
			lyxerr << "ERROR: Cannot handle include file `"
			       << MakeDisplayPath(getFileName())
			       << "' which has textclass `"
			       << textclasslist.nameOfClass(tmp->params.textclass)
			       << "' instead of `"
			       << textclasslist.nameOfClass(master->params.textclass)
			       << "'.\n";
			return 0;
               }

		// write it to a file (so far the complete file)
		writefile = ChangeExtension(getFileName(), ".tex", false);
		if (!master->getTmpDir().empty()) {
//		    && !master->niceFile) {
			incfile = subst(incfile, '/','@');
			#ifdef __EMX__
			incfile.subst(':', '$');
			#endif
			writefile = AddName(master->getTmpDir(), incfile);
		} else
			writefile = getFileName();
		writefile = ChangeExtension(writefile,".tex",false);
		lyxerr[Debug::LATEX] << "incfile:" << incfile << '\n';
		lyxerr[Debug::LATEX] << "writefile:" << writefile << '\n';
		
		tmp->markDepClean(master->getTmpDir());
		
		tmp->makeLaTeXFile(writefile,
				   OnlyPath(getMasterFilename()), 
				   true /* master->niceFile */, true);
	} 

	if (isVerb()) {
		file += '\\';
		file += command + '{';
		file += incfile + '}';
	} else if (isInput()) {
		// \input wants file with extension (default is .tex)
		if (!IsLyXFilename(getFileName())) {
			file += '\\';
			file += command + '{';
			file += incfile + '}';
		} else {
			file += '\\';
			file += command + '{';
			file +=	ChangeExtension(incfile, ".tex", false)
				+ '}';
		}
	} else {
		// \include don't want extension and demands that the
		// file really have .tex
		file += '\\';
		file += command + '{';
		file +=	ChangeExtension(incfile, string(), false)
			+ '}';
	}

	return 0;
}


void InsetInclude::Validate(LaTeXFeatures& features) const
{
	if (isVerb())
		features.verbatim = true;

	// Here we must do the fun stuff...
	// Load the file in the include if it needs
	// to be loaded:
	if (loadIfNeeded()) {
		// a file got loaded
		Buffer *tmp = bufferlist.getBuffer(getFileName());
		tmp->validate(features);
	}
}
