// -*- C++ -*-
/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 *           Copyright (C) 1995 Matthias Ettrich
 *           Copyright (C) 1995-1998 The LyX Team.
 *
 *           This file is Copyright (C) 1996-1998
 *           Lars Gullik Bj<F8>nnes
 *
 * ======================================================
 */

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "insetnote.h"
#include "lyxparagraph.h"
#include "textclasslist.h"
#include "buffer.h"
#include "insetlatexaccent.h"
#include "insetspecialchar.h"
#include "debug.h"
#include "lyxlex.h"

InsetNote::InsetNote(Buffer *b)
	: buffer(b), par(0)
{
}


InsetNote::~InsetNote()
{
	if (par) delete par;
}


int InsetNote::ascent(Painter &, LyXFont const & font) const
{
	return font.maxAscent() + 3;
}


int InsetNote::descent(Painter &, LyXFont const & font) const
{
	return font.maxDescent() + 3;
}


int InsetNote::width(Painter &, LyXFont const &) const
{
	return 50;
}


LyXFont InsetNote::GetFont(LyXParagraph* par, int pos)
{
    char par_depth = par->GetDepth();

    LyXLayout const *layout = textclasslist.style(buffer->params.textclass,
				       par->GetLayout());

    // We specialize the 95% common case:
    if (par->footnoteflag == LyXParagraph::NO_FOOTNOTE && !par_depth) {
        if (pos >= 0) {
            // 95% goes here
            if (layout->labeltype == LABEL_MANUAL
                && pos < BeginningOfMainBody(par)) {
		// 1% goes here
                return par->GetFontSettings(pos).realize(layout->reslabelfont);
            } else
                return par->GetFontSettings(pos).realize(layout->resfont);
        } else {
            // 5% goes here.
            // process layoutfont for pos == -1 and labelfont for pos < -1
            if (pos == -1)
                return layout->resfont;
            else
                return layout->reslabelfont;
        }
    }
    // The uncommon case need not be optimized as much

    LyXFont layoutfont, tmpfont;

    if (pos >= 0){
        // 95% goes here
        if (pos < BeginningOfMainBody(par)) {
            // 1% goes here
            layoutfont = layout->labelfont;
        } else {
            // 99% goes here
            layoutfont = layout->font;
        }
        tmpfont = par->GetFontSettings(pos);
        tmpfont.realize(layoutfont);
    } else{
        // 5% goes here.
        // process layoutfont for pos == -1 and labelfont for pos < -1
        if (pos == -1)
            tmpfont = layout->font;
        else
            tmpfont = layout->labelfont;
    }
    
    // Resolve against environment font information
    //if (par->GetDepth()){ // already in while condition
    while (par && par_depth && !tmpfont.resolved()) {
        par = par->DepthHook(par_depth - 1);
        if (par) {
            tmpfont.realize(textclasslist.style(buffer->params.textclass,
                                           par->GetLayout())->font);
            par_depth = par->GetDepth();
        }
    }
    tmpfont.realize(textclasslist.textClass(buffer->params.textclass)->defaultfont);

    // Cosmetic improvement: If this is an open footnote, make the font 
    // smaller.
    if (par->footnoteflag == LyXParagraph::OPEN_FOOTNOTE
        && par->footnotekind == LyXParagraph::FOOTNOTE) {
        tmpfont.decSize();
    }
    return tmpfont;
}


void InsetNote::draw(Painter &pain, const LyXFont &font,
		     int baseline, float &x)
{
#warning A lot of work to do here...
	int w = 0;
	pain.buttonText(pain.paperMargin(), baseline + ascent(pain, font),
			"Foot", font, true, w);
	x = pain.paperMargin() + w;
#if 0
	// Here we should just call the paragraph.draw method. And get its
	// geometry. Yet not possible:
	int height = 0;
	par.draw(pain, font, baseline, float(x)/*temporary*/, height);
	pain.rectangle(x, baseline, pain.paperWidth() - x,
		       heigth /* + litt */, LColor::footnoteframe);
#else
	char ch = 0;
	for(int p = 0; (p < par->Last()) && !par->IsNewline(p); ++p) {
		ch = par->GetChar(p);
		LyXFont font = GetFont(par, p);
		if (IsFloatChar(ch)) {
			// ignore for now
		} else if (ch == LYX_META_INSET) {
			Inset *tmpinset = par->GetInset(p);
			if (tmpinset)
				tmpinset->draw(pain, font, baseline, x);
			x += tmpinset->width(pain, font);
		} else {
			pain.text(int(x), baseline, ch, font);
			x += pain.width(ch, font);
		}
	}
#endif
}


string InsetNote::EditMessage() const
{
	return _("Opened Note Inset");
}


Inset::EDIT InsetNote::Editable() const
{
	return Inset::HIGHLY_EDITABLE;
}


void InsetNote::Write(FILE *file)
{
	fprintf(file, "\\begin_inset Footnote\n");
	LyXFont font1 = LyXFont(LyXFont::ALL_INHERIT);
	LyXFont font2;
	int column = 0;
	char c = 0;
	
	for (int i = 0; i < par->last(); ++i) {
		// Write font changes
		font2 = par->GetFontSettings(i);
		if (font2 != font1) {
			font2.lyxWriteChanges(font1, file);
			column = 0;
			font1 = font2;
		}
		c = par->GetChar(i);
		// A newline before tags
		if (column > 0 &&
		    (c == LYX_META_INSET ||
		     c == LYX_META_NEWLINE ||
		     c == '\\')) {
			fprintf(file, "\n");
			column = 0;
		}
		
		switch (c) {
		case LYX_META_INSET: {
			Inset *inset = par->GetInset(i);
			if (inset) {
				inset->Write(file);
				column = 0;
			}
		}
		break;
		case LYX_META_NEWLINE: 
			fprintf(file, "\\newline\n");
			column = 0;
			break;
		case '\\': 
			fprintf(file, "\\backslash\n");
			column = 0;
			break;
		default:
			if (column > 65 && c==' ') {
				fprintf(file, "\n");
				column = 0;
			}
			// this check is to amend a bug. LyX sometimes
			// inserts '\0' this could cause problems.
			if (c != '\0')
				fprintf(file, "%c", c);
			else
				lyxerr << "ERROR (InsetNote::writeFile):"
					" '\\0' char in structure.\n";
			column++;
			break;
		}
	}
	// A newline if the last c was not a tag.
	if (column > 0 &&
	    (c != LYX_META_INSET &&
	     c != LYX_META_NEWLINE &&
	     c != '\\')) {
		fprintf(file, "\n");
		column = 0;
	}
	fprintf(file, "\\end_inset\n");
}


void InsetNote::Read(LyXLex &lex)
{
	string token;
	int pos = 0;
	LyXFont font = LyXFont(LyXFont::ALL_INHERIT);
	if (par)
		delete par;
	par = new LyXParagraph;
	
	while (lex.IsOK()) {
		lex.nextToken();
		token = lex.GetString();
		
		if (token.empty())
			continue;
		else if (token[0] != '\\') {
			int n = token.length();
			for (int i = 0; i < n; ++i) {
				par->InsertChar(pos, token[i]);
				par->SetFont(pos, font);
				++pos;
			}
		} else if (token == "\\newline") {
			par->InsertChar(pos, LYX_META_NEWLINE);
			par->SetFont(pos, font);
			++pos;
		} else if (token == "\\family") {
			lex.next();
			font.setLyXFamily(lex.GetString());
		} else if (token == "\\series") {
			lex.next();
			font.setLyXSeries(lex.GetString());
		} else if (token == "\\shape") {
			lex.next();
			font.setLyXShape(lex.GetString());
		} else if (token == "\\size") {
			lex.next();
			font.setLyXSize(lex.GetString());
		} else if (token == "\\latex") {
			lex.next();
			string tok = lex.GetString();
			// This is dirty, but gone with LyX3. (Asger)
			if (tok == "no_latex")
				font.setLatex(LyXFont::OFF);
			else if (tok == "latex")
				font.setLatex(LyXFont::ON);
			else if (tok == "default")
				font.setLatex(LyXFont::INHERIT);
			else
				lex.printError("Unknown LaTeX font flag "
					       "`$$Token'");
		}else if (token == "\\emph") {
			lex.next();
			font.setEmph(font.setLyXMisc(lex.GetString()));
		} else if (token == "\\bar") {
			lex.next();
			string tok = lex.GetString();
			// This is dirty, but gone with LyX3. (Asger)
			if (tok == "under")
				font.setUnderbar(LyXFont::ON);
			else if (tok == "no")
				font.setUnderbar(LyXFont::OFF);
			else if (tok == "default")
				font.setUnderbar(LyXFont::INHERIT);
			else
				lex.printError("Unknown bar font flag "
					       "`$$Token'");
		} else if (token == "\\noun") {
			lex.next();
			font.setNoun(font.setLyXMisc(lex.GetString()));
		} else if (token == "\\color") {
			lex.next();
			font.setLyXColor(lex.GetString());
		}
#if 0
		else if (token == "\\begin_inset") {
			Inset *inset = 0;
			lex.next();
			string tmptok = lex.GetString();
			// Test the different insets.
			if (tmptok == "Quotes") {
				inset = new InsetQuotes;
				inset->Read(lex);
			} else if (tmptok == "LaTeXAccent" || tmptok == "\\i") {
				inset = new InsetLatexAccent;
				inset->Read(lex);
			} else if (tmptok == "FormulaMacro") {
				inset = new InsetFormulaMacro;
				inset->Read(lex);
			} else if (tmptok == "Formula") {
				inset = new InsetFormula;
				inset->Read(lex);
			} else if (tmptok == "Figure") {
				inset = new InsetFig(100,100, buffer);
				inset->Read(lex);
			} else if (tmptok == "Tabular") {
				inset = new InsetTabular(buffer);
				inset->Read(lex);
			} else if (tmptok == "Info") {
				inset = new InsetInfo;
				inset->Read(lex);
			} else if (tmptok == "Include") {
				inset = new InsetInclude(string(), buffer);
				inset->Read(lex);
			} else if (tmptok == "LatexCommand") {
				InsetCommand inscmd;
				inscmd.Read(lex);
				if (inscmd.getCmdName()=="cite") {
					inset = new InsetCitation(inscmd.getContents(), inscmd.getOptions());
				} else if (inscmd.getCmdName()=="bibitem") {
					lex.printError("Wrong place for bibitem");
					inset = inscmd.Clone();
				} else if (inscmd.getCmdName()=="BibTeX") {
					inset = new InsetBibtex(inscmd.getContents(), inscmd.getOptions(), buffer);
				} else if (inscmd.getCmdName()=="index") {
					inset = new InsetIndex(inscmd.getContents());
				} else if (inscmd.getCmdName()=="include") {
					inset = new InsetInclude(inscmd.getContents(), buffer);
				} else if (inscmd.getCmdName()=="label") {
					inset = new InsetLabel(inscmd.getCommand());
				} else if (inscmd.getCmdName() == "ref"
					   || inscmd.getCmdName() == "pageref") {
					inset = new InsetRef(inscmd, buffer);
				}
			}
			if (inset) {
				par->InsertChar(pos, LYX_META_INSET);
				par->InsertInset(pos, inset);
				par->SetFont(pos, font);
				++pos;
			} else {
				lex.printError("Unknown inset `$$Token'. "
					       "Inserting as text.");
			}
		}
#endif 
#ifndef NO_COMPABILITY
		else if (token == "\\hfill") {
			// now obsolete, but we have a bak compability
			Inset *inset = new InsetSpecialChar(InsetSpecialChar::HFILL);
			par->InsertChar(pos, LYX_META_INSET);
			par->InsertInset(pos, inset);
			par->SetFont(pos, font);
			++pos;
		} else if (token == "\\protected_separator") {
			// now obsolete, but we have a back compability
			//par->InsertChar(pos, LYX_META_PROTECTED_SEPARATOR);
			Inset *inset = new InsetSpecialChar(InsetSpecialChar::PROTECTED_SPACE);
			par->InsertChar(pos, LYX_META_INSET);
			par->InsertInset(pos, inset);
			par->SetFont(pos, font);
			++pos;
#endif
		} else if (token == "\\bibitem") {  // ale970302
			if (!par->bibkey)
				par->bibkey = new InsetBibKey;
			par->bibkey->Read(lex);                 
		}else if (token == "\\backslash") {
			par->InsertChar(pos, '\\');
			par->SetFont(pos, font);
			++pos;
		} else if (token == "\\end_inset") {
			break;
		} else {
			lex.printError("Unknown note token `$$Token'. Not handled!");
			break;
		}
	}
	if (token != "\\end_inset") {
		lex.printError("Missing \\end_float at this point. "
			       "Read: `$$Token'");
	}
}


int InsetNote::Latex(FILE *, signed char /*fragile*/)
{
	return 0;
}


int InsetNote::Latex(string &, signed char /*fragile*/)
{
	return 0;
}


void InsetNote::Validate(LaTeXFeatures &) const
{
}


Inset::Code InsetNote::LyxCode() const
{
	return Inset::NOTE_CODE;
}


InsetNote* InsetNote::Clone()
{
	return new InsetNote(buffer);
}


bool InsetNote::display() const
{
	// will be dynamic later.
	return true;
}


int InsetNote::BeginningOfMainBody(LyXParagraph *par)
{
    if (textclasslist.style(buffer->params.textclass,
                       par->GetLayout())->labeltype != LABEL_MANUAL)
        return 0;
    else
        return par->BeginningOfMainBody();
}
