/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 *
 *           Copyright (C) 1995-1998 The LyX Team.
 *
 *======================================================
 */

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "insettabular.h"
#include "lyxlex.h"
#include "debug.h"
#include "lyxfont.h"
#include "lyxlex.h"
#include "commandtags.h"
#include "buffer.h"
#include "LyXView.h"
#include "BufferView.h"
#include "textutils.h"
#include "textclasslist.h"
#include "insetlatexaccent.h"
#include "insetquotes.h"
#include "formulamacro.h"
#include "insetfig.h"
#include "insetinfo.h"
#include "insetinclude.h"
#include "insetbibtex.h"
#include "insetcitation.h"
#include "insetcommand.h"
#include "insetindex.h"
#include "insetlabel.h"
#include "insetref.h"
#include "insetspecialchar.h"
#include "LaTeXFeatures.h"
#include "painter.h"
#include "lyx_gui_misc.h"

const int ADD_TO_HEIGHT = 2;
const int ADD_TO_TABLE_WIDTH = 2;

extern unsigned char getCurrentTextClass(Buffer*);

InsetTabular::InsetTabular(Buffer *buf, int rows=1, int columns=1)
{
    if (rows <= 0)
        rows = 1;
    if (columns <= 0)
        columns = 1;
    tabular = new LyXTable(rows,columns);
    par = new LyXParagraph();
    // inizialize paragraph
    for (int i=(rows * columns)-1; i >= 0; --i)
        par->InsertChar(0,LYX_META_NEWLINE);
    the_locking_inset = 0;
    buffer = buf;
    cursor_visible = false;
    old_x = -1;
    calculate_width_of_cells();
    sel_pos_start = sel_pos_end = 0;
    no_selection = false;
}

InsetTabular::InsetTabular(InsetTabular const & tab, Buffer *buf)
{
    tabular = new LyXTable(*(tab.tabular));
    par = new LyXParagraph(tab.par);
    the_locking_inset = 0;
    buffer = buf;
    cursor_visible = false;
    old_x = -1;
    calculate_width_of_cells();
    sel_pos_start = sel_pos_end = 0;
    no_selection = false;
}

InsetTabular::~InsetTabular()
{
    if (par)
        delete par;
}


InsetTabular* InsetTabular::Clone()
{
    InsetTabular* t = new InsetTabular(*this, buffer);
    return t;
}


void InsetTabular::Write(FILE *file)
{
    fprintf(file, "\\begin_inset Tabular\n");
    tabular->Write(file);

    LyXFont font1 = LyXFont(LyXFont::ALL_INHERIT);
    LyXFont font2;
    int column = 0;
    char c = 0;

    for (int i = 0; i < par->last(); ++i) {
        // Write font changes
        font2 = par->GetFontSettings(i);
        if (font2 != font1) {
            font2.lyxWriteChanges(font1, file);
            column = 0;
            font1 = font2;
        }
        c = par->GetChar(i);
	// A newline before tags
	if (column > 0 &&
	    (c == LYX_META_INSET ||
	     c == LYX_META_NEWLINE ||
	     c == '\\')) {
		fprintf(file, "\n");
		column = 0;
	}
	
        switch (c) {
	case LYX_META_INSET: {
              Inset *inset = par->GetInset(i);
              if (inset) {
                  inset->Write(file);
                  column = 0;
              }
	}
              break;
          case LYX_META_NEWLINE: 
              fprintf(file, "\\newline\n");
              column = 0;
              break;
          case '\\': 
              fprintf(file, "\\backslash\n");
              column = 0;
              break;
          default:
              if (column > 65 && c==' ') {
                  fprintf(file, "\n");
                  column = 0;
              }
              // this check is to amend a bug. LyX sometimes
              // inserts '\0' this could cause problems.
              if (c != '\0')
                  fprintf(file, "%c", c);
              else
		      lyxerr << "ERROR (InsetTabular::writeFile):"
			      " '\\0' char in structure.\n";
              column++;
              break;
        }
    }
	// A newline if the last c was not a tag.
	if (column > 0 &&
	    (c != LYX_META_INSET &&
	     c != LYX_META_NEWLINE &&
	     c != '\\')) {
		fprintf(file, "\n");
		column = 0;
	}
	fprintf(file, "\\end_inset\n");
}


void InsetTabular::Read(LyXLex &lex)
{
#ifndef NO_COMPABILITY
    bool old_format = (lex.GetString() == "\\LyXTable");
#endif
    string token, tmptok;
    LyXFont font = LyXFont(LyXFont::ALL_INHERIT);
    int pos = 0;

    if (tabular)
        delete tabular;
    if (par)
        delete par;
    
    tabular = new LyXTable(lex);
    par = new LyXParagraph;
    
    while (lex.IsOK()) {
        lex.nextToken();
        token = lex.GetString();
        if (token.empty())
            continue;
        else if (token[0] != '\\') {
            int n = token.length();
            for (int i=0; i < n; ++i) {
                par->InsertChar(pos, token[i]);
                par->SetFont(pos, font);
                ++pos;
            }
        } else if (token == "\\newline") {
            par->InsertChar(pos, LYX_META_NEWLINE);
            par->SetFont(pos, font);
            ++pos;
        } else if (token == "\\family") { 
            lex.next();
            font.setLyXFamily(lex.GetString());
        } else if (token == "\\series") {
            lex.next();
            font.setLyXSeries(lex.GetString());
        } else if (token == "\\shape") {
            lex.next();
            font.setLyXShape(lex.GetString());
        } else if (token == "\\size") {
            lex.next();
            font.setLyXSize(lex.GetString());
        } else if (token == "\\latex") {
            lex.next();
            string tok = lex.GetString();
            // This is dirty, but gone with LyX3. (Asger)
            if (tok == "no_latex")
                font.setLatex(LyXFont::OFF);
            else if (tok == "latex")
                font.setLatex(LyXFont::ON);
            else if (tok == "default")
                font.setLatex(LyXFont::INHERIT);
            else
                lex.printError("Unknown LaTeX font flag "
                               "`$$Token'");
        } else if (token == "\\emph") {
            lex.next();
            font.setEmph(font.setLyXMisc(lex.GetString()));
        } else if (token == "\\bar") {
            lex.next();
            string tok = lex.GetString();
            // This is dirty, but gone with LyX3. (Asger)
            if (tok == "under")
                font.setUnderbar(LyXFont::ON);
            else if (tok == "no")
                font.setUnderbar(LyXFont::OFF);
            else if (tok == "default")
                font.setUnderbar(LyXFont::INHERIT);
            else
                lex.printError("Unknown bar font flag "
                               "`$$Token'");
        } else if (token == "\\noun") {
            lex.next();
            font.setNoun(font.setLyXMisc(lex.GetString()));
        } else if (token == "\\color") {
            lex.next();
            font.setLyXColor(lex.GetString());
       } else if (token == "\\begin_inset") {
           Inset *inset = 0;
           lex.next();
           tmptok = lex.GetString();
           // Test the different insets.
           if (tmptok == "Quotes") {
               inset = new InsetQuotes(buffer);
               inset->Read(lex);
           } else if (tmptok == "LaTeXAccent" || tmptok == "\\i") {
               inset = new InsetLatexAccent;
               inset->Read(lex);
           } else if (tmptok == "FormulaMacro") {
               inset = new InsetFormulaMacro;
               inset->Read(lex);
           } else if (tmptok == "Formula") {
               inset = new InsetFormula;
               inset->Read(lex);
           } else if (tmptok == "Figure") {
               inset = new InsetFig(100,100, buffer);
               inset->Read(lex);
           } else if (tmptok == "Tabular") {
               inset = new InsetTabular(buffer);
               inset->Read(lex);
           } else if (tmptok == "Info") {
               inset = new InsetInfo;
               inset->Read(lex);
           } else if (tmptok == "Include") {
               inset = new InsetInclude(string(), buffer);
               inset->Read(lex);
           } else if (tmptok == "LatexCommand") {
               InsetCommand inscmd;
               inscmd.Read(lex);
               if (inscmd.getCmdName()=="cite") {
                   inset = new InsetCitation(inscmd.getContents(),
                                             inscmd.getOptions());
               } else if (inscmd.getCmdName()=="bibitem") {
                   lex.printError("Wrong place for bibitem");
                   inset = inscmd.Clone();
               } else if (inscmd.getCmdName()=="BibTeX") {
                   inset = new InsetBibtex(inscmd.getContents(),
                                           inscmd.getOptions(), buffer);
               } else if (inscmd.getCmdName()=="index") {
                   inset = new InsetIndex(inscmd.getContents());
               } else if (inscmd.getCmdName()=="include") {
                   inset = new InsetInclude(inscmd.getContents(), buffer);
               } else if (inscmd.getCmdName()=="label") {
                   inset = new InsetLabel(inscmd.getCommand());
               } else if (inscmd.getCmdName() == "ref" ||
                          inscmd.getCmdName() == "pageref") {
                   inset = new InsetRef(inscmd, buffer);
               } else
                   // The following three are only for compatibility
                   if (inscmd.getCmdName()=="-") {
                       inset = new InsetSpecialChar(InsetSpecialChar::HYPHENATION);
                   } else if (inscmd.getCmdName()=="@.") {
                       inset = new InsetSpecialChar(InsetSpecialChar::END_OF_SENTENCE);
                   } else if (inscmd.getCmdName()=="ldots") {
                       inset = new InsetSpecialChar(InsetSpecialChar::LDOTS);
                   } else
                       inset = inscmd.Clone();
           }
           if (inset) {
               par->InsertChar(pos, LYX_META_INSET);
               par->InsertInset(pos, inset);
               par->SetFont(pos, font);
               ++pos;
           } else {
               lex.printError("Unknown inset `$$Token'. "
                              "Inserting as text.");
           }
#ifndef NO_COMPABILITY
        } else if (token == "\\hfill") {
            // now obsolete, but we have a bak compability
            Inset *inset = new InsetSpecialChar(InsetSpecialChar::HFILL);
            par->InsertChar(pos, LYX_META_INSET);
            par->InsertInset(pos, inset);
            par->SetFont(pos, font);
            ++pos;
        } else if (token == "\\protected_separator") {
            // now obsolete, but we have a back compability
            //par->InsertChar(pos, LYX_META_PROTECTED_SEPARATOR);
            Inset *inset = new InsetSpecialChar(InsetSpecialChar::PROTECTED_SPACE);
            par->InsertChar(pos, LYX_META_INSET);
            par->InsertInset(pos, inset);
            par->SetFont(pos, font);
            ++pos;
#endif
        } else if (token == "\\bibitem") {  // ale970302
            if (!par->bibkey)
                par->bibkey = new InsetBibKey;
            par->bibkey->Read(lex);                 
        }else if (token == "\\backslash") {
            par->InsertChar(pos, '\\');
            par->SetFont(pos, font);
            ++pos;
        } else if (token == "\\end_inset") {
            break;
        } else {
#ifndef NO_COMPABILITY
            if (old_format) {
                // When old_format we must assume that upon the first
                // encounter of an unknown token we have reached the
                // end of the tabular, so break out and return.
                lex.pushToken(token);
                break;
            }
#endif
            lex.printError("Unknown tabular token `$$Token'. Not handled!");
	    break;
        }
    }
    calculate_width_of_cells();
#ifndef NO_COMPABILITY
    if (old_format) return;
#endif
    if (token != "\\end_inset") {
        lex.printError("Missing \\end_inset at this point. "
                       "Read: `$$Token'");
    }
}


int InsetTabular::ascent(Painter &, LyXFont const &/*font*/) const
{
    return tabular->AscentOfRow(0);
    // return font.maxAscent();
}


int InsetTabular::descent(Painter &, LyXFont const &) const
{
    return tabular->HeightOfTable() - tabular->AscentOfRow(0);
}


int InsetTabular::width(Painter &, LyXFont const &) const
{
    return tabular->WidthOfTable() + (2 * ADD_TO_TABLE_WIDTH);
}


void InsetTabular::draw(Painter &pain, const LyXFont &, int baseline, float &x)
{
    string
        text;
    int
	s_start, s_end,
        i,j,p=0,
	oldp,
	oldcell,
        cell=-1,
        width,
	asc,desc,
        cell_width;
    float
	ssel_x, esel_x,
        oldx,
        newx;
    char
        ch;
    LyXFont
        font;
    bool
	do_reset_pos;

    UpdatableInset::draw(pain,font,baseline,x);
    do_reset_pos = (x != top_x) || (baseline != top_baseline);
    x += ADD_TO_TABLE_WIDTH;
    top_x = int(x);
    top_baseline = baseline;
    oldx = x;
    for(i=0;i<tabular->rows;++i) {
	ssel_x = esel_x = x = oldx;
	if (hasCharSelection()) {
	    if (sel_pos_start > sel_pos_end) {
		s_start = sel_pos_end;
		s_end = sel_pos_start;
	    } else {
		s_start = sel_pos_start;
		s_end = sel_pos_end;
	    }
	    oldcell = cell;
	    oldp = p;
	    for(j=0;j<tabular->columns;++j) {
		if (tabular->IsPartOfMultiColumn(i,j))
		    continue;
		++cell;
		width = tabular->WidthOfColumn(cell);
		cell_width = tabular->GetWidthOfCell(cell);
		newx = x + tabular->GetBeginningOfTextInCell(cell);
		x += width;
		for(;(p < par->last()) && !par->IsNewline(p);++p) {
		    if (p == s_start)
			ssel_x = newx;
		    if ((p >= s_start) && (p <= s_end))
			esel_x = newx;
		    ch = par->GetChar(p);
		    font = GetFont(par,p);
		    if (IsFloatChar(ch)) {
			// skip for now
		    } else if (ch == LYX_META_INSET) {
			Inset *tmpinset = par->GetInset(p);
			newx += tmpinset->width(pain, font);
		    } else {
			newx += pain.width(ch,font);
		    }
		}
		if (p == s_start)
		    ssel_x = newx;
		if ((p >= s_start) && (p <= s_end))
		    esel_x = newx;
		++p; // p is now on META_NEWLINE
	    }
	    if (ssel_x < esel_x) {
		asc = tabular->AscentOfRow(i)-1;   // how get the lineheight?
		desc = tabular->DescentOfRow(i);
		pain.fillRectangle(int(ssel_x), baseline-asc,
				   int(esel_x - ssel_x), asc+desc,
				   LColor::selection);
	    }
	    x = oldx;
	    p = oldp;
	    cell = oldcell;
	}
        for(j=0;j<tabular->columns;++j) {
            if (tabular->IsPartOfMultiColumn(i,j))
                continue;
            ++cell;
            width = tabular->WidthOfColumn(cell);
            cell_width = tabular->GetWidthOfCell(cell);
            newx = x + tabular->GetBeginningOfTextInCell(cell);
            x += width;
            for(;(p < par->last()) && !par->IsNewline(p);++p) {
                ch = par->GetChar(p);
                font = GetFont(par,p);
                if (IsFloatChar(ch)) {
                    // skip for now
                } else if (ch == LYX_META_INSET) {
                    Inset *tmpinset = par->GetInset(p);
                    if (tmpinset) 
                        tmpinset->draw(pain, font, baseline, newx);
                } else {
                    pain.text(int(newx), baseline, ch, font);
                    newx += pain.width(ch,font);
                }
            }
            DrawCellLines(pain, int(x), baseline, i, cell);
            ++p; // p is now on META_NEWLINE
        }
        // this should be no problem as the tabular functions already
        // check for a row-overflow
        baseline += tabular->DescentOfRow(i) + tabular->AscentOfRow(i+1)
            + tabular->AdditionalHeight(cell+1);
    }
    x += ADD_TO_TABLE_WIDTH;
    if (!the_locking_inset && do_reset_pos) {
	HideInsetCursor();
	resetPos();
	ShowInsetCursor();
    }
}


string InsetTabular::EditMessage() const
{
    return _("Opened Tabular Inset");
}


void InsetTabular::Edit(int x, int y, unsigned int button, BufferView *bv)
{
    UpdatableInset::Edit(x, y, button, bv);

    bv->lockInset(this);
    the_locking_inset = 0;
    inset_pos = inset_x = inset_y = 0;
    no_selection = true;
    setPos(x,y);
    sel_pos_start = sel_pos_end = actpos;
    sel_cell_start = sel_cell_end = actcell;
    current_font = real_current_font = GetFont(par, actpos);
    bv->getOwner()->getPopups().updateFormTable();
}


void InsetTabular::InsetUnlock()
{
    if (the_locking_inset)
	the_locking_inset->InsetUnlock();
    HideInsetCursor();
    the_locking_inset = 0;
    if (hasCharSelection()) {
	sel_pos_start = sel_pos_end = actpos;
	UpdateLocal(false);
    } else
	sel_pos_start = sel_pos_end = actpos;
    no_selection = false;
}

bool InsetTabular::UnlockInsetInInset(Inset *inset, bool lr=false)
{
    if (!the_locking_inset)
	return false;
    if (the_locking_inset == inset) {
	the_locking_inset->InsetUnlock();
	the_locking_inset = 0;
	if (lr)
	    moveRight(false);
	return true;
    }
    return the_locking_inset->UnlockInsetInInset(inset);
}


void InsetTabular::UpdateLocal(bool flag = true)
{
    resetPos();
    bview->updateInset(this, flag);
}

bool InsetTabular::UpdateInsetInInset(BufferView *bv, Inset *inset)
{
    if (!the_locking_inset)
	return false;
    if (the_locking_inset != inset)
	return the_locking_inset->UpdateInsetInInset(bv, inset);
    UpdateLocal();
    return true;
#if 0
    float
	x = inset_x;
    UpdateLocal();
    inset->draw(bv->getPainter(), real_current_font, inset_y, x);
    return true;
#endif
}

void InsetTabular::InsetButtonRelease(int x, int y, int button)
{
    if (the_locking_inset) {
        the_locking_inset->InsetButtonRelease(x-inset_x,y-inset_y,button);
        return;
    }
    no_selection = false;
}

void InsetTabular::InsetButtonPress(int x, int y, int button)
{
    if (hasCharSelection()) {
	sel_pos_start = sel_pos_end = 0;
	UpdateLocal(false);
    }
    no_selection = false;
    setPos(x,y,false);
    if (the_locking_inset) {
	UpdatableInset
	    *inset=0;
	if (par->GetChar(actpos)==LYX_META_INSET)
	    inset=(UpdatableInset*)par->GetInset(actpos);
	if (the_locking_inset == inset) {
	    the_locking_inset->InsetButtonPress(x-inset_x,y-inset_y,button);
	    return;
	} else if (inset) {
	    // otherwise unlock the_locking_inset and lock the new inset
	    inset_x = cx-top_x;
	    inset_y = cy;
	    inset_pos = actpos;
	    the_locking_inset->InsetUnlock();
	    the_locking_inset = inset;
	    the_locking_inset->Edit(x - inset_x, y - inset_y, button, bview);
	    return;
	}
	// otherwise only unlock the_locking_inset
	the_locking_inset->InsetUnlock();
    }
    int
	oldcell = actcell;
    setPos(x,y);
    the_locking_inset = 0;
    sel_pos_start = sel_pos_end = actpos;
    sel_cell_start = sel_cell_end = actcell;
    if (button == 3)
        bview->getOwner()->getPopups().showFormTable();
    else if (oldcell != actcell)
        bview->getOwner()->getPopups().updateFormTable();
}


void InsetTabular::InsetMotionNotify(int x, int y, int button)
{
    if (the_locking_inset) {
        the_locking_inset->InsetMotionNotify(x-inset_x,y-inset_y,button);
        return;
    }
    if (!no_selection) {
	int
	    oldcell = actcell,
	    old = sel_pos_end;

	setPos(x,y);
	sel_pos_end = actpos;
	sel_cell_end = actcell;
	if (old != sel_pos_end)
	    UpdateLocal(false);
	if (oldcell != actcell)
	    bview->getOwner()->getPopups().updateFormTable();
    }
    no_selection = false;
}


void InsetTabular::InsetKeyPress(XKeyEvent *xke)
{
    if (the_locking_inset) {
        the_locking_inset->InsetKeyPress(xke);
        return;
    }
}


UpdatableInset::RESULT InsetTabular::LocalDispatch(int action, char const *arg)
{
    no_selection = false;
    if (UpdatableInset::LocalDispatch(action, arg)) {
	resetPos();
	return DISPATCHED;
    }

    UpdatableInset::RESULT 
        result=DISPATCHED;
    if ((action < 0) && !arg)
        return FINISHED;

    if ((action != LFUN_DOWN) && (action != LFUN_UP) &&
        (action != LFUN_DOWNSEL) && (action != LFUN_UPSEL))
        old_x = -1;
    if (the_locking_inset) {
        result=the_locking_inset->LocalDispatch(action, arg);
	if (result == DISPATCHED) {
            setWidthOfCell(actpos,actcell,actrow);
	    the_locking_inset->ToggleInsetCursor();
	    UpdateLocal(false);
	    the_locking_inset->ToggleInsetCursor();
            return result;
        } else if (result == FINISHED) {
	    if ((action == LFUN_RIGHT) || (action == -1)) {
		actpos = inset_pos + 1;
		resetPos();
	    }
	    the_locking_inset=0;
	    result = DISPATCHED;
	    return result;
	}
    }
    int
        oldcell = actcell;

    HideInsetCursor();
    switch (action) {
        // Normal chars
      case -1:
          par->InsertChar(actpos,arg[0]);
	  par->SetFont(actpos,real_current_font);
          setWidthOfCell(actpos,actcell,actrow);
          ++actpos;
	  sel_pos_start = sel_pos_end = actpos;
	  UpdateLocal();
          break;
        // --- Cursor Movements ---------------------------------------------
      case LFUN_RIGHTSEL:
          moveRight(false);
	  sel_pos_end = actpos;
	  UpdateLocal(false);
	  break;
      case LFUN_RIGHT:
          result= DISPATCH_RESULT(moveRight());
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_LEFTSEL:
          moveLeft(false);
	  sel_pos_end = actpos;
	  UpdateLocal(false);
	  break;
      case LFUN_LEFT:
          result= DISPATCH_RESULT(moveLeft());
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_DOWNSEL:
          moveDown(false);
	  sel_pos_end = actpos;
	  UpdateLocal(false);
	  break;
      case LFUN_DOWN:
          result= DISPATCH_RESULT(moveDown());
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_UPSEL:
          moveUp(false);
	  sel_pos_end = actpos;
	  UpdateLocal(false);
	  break;
      case LFUN_UP:
          result= DISPATCH_RESULT(moveUp());
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_BACKSPACE:
          if (!actpos || par->IsNewline(actpos-1)) {
	      if (hasCharSelection()) {
		  sel_pos_start = sel_pos_end = 0;
		  UpdateLocal(false);
	      }
	      break;
	  }
          moveLeft();
      case LFUN_DELETE:
          if (Delete()) { // we need update
              setWidthOfCell(actpos,actcell,actrow);
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal();
          } else if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_HOME:
          for(;(actpos > 0) && !par->IsNewline(actpos-1);--actpos)
              cx -= SingleWidth(par, actpos-1);
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  } else
	      sel_pos_start = sel_pos_end = actpos;
          break;
      case LFUN_END:
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  }
          for(;(actpos < par->last()) && !par->IsNewline(actpos);++actpos)
              cx += SingleWidth(par, actpos);
	  sel_pos_start = sel_pos_end = actpos;
	  resetPos();
          break;
      case LFUN_TAB:
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  }
	  sel_pos_start = sel_pos_end = actpos;
          moveNextCell();
          break;
      case LFUN_MATH_MODE:   // Open or create a math inset
          InsertInset(new InsetFormula);
	  if (hasCharSelection()) {
	      sel_pos_start = sel_pos_end = actpos;
	      UpdateLocal(false);
	  }
          return DISPATCHED;
          break;
      default:
	  /// A little inset doesn't have to do lyxfunc job [ale]
          result = UNDISPATCHED;
          break;
    }
    if (result!=FINISHED) {
	if (!the_locking_inset) {
	    if (oldcell != actcell)
		bview->getOwner()->getPopups().updateFormTable();
	    ShowInsetCursor();
	}
    } else
        bview->unlockInset(this);
    return result;
}


int InsetTabular::Latex(FILE *file, signed char fragile)
{
    string
        fstr;
    int
        i;

    i=Latex(fstr, fragile);
    fprintf(file, fstr.c_str());
    return i;
}


int InsetTabular::Latex(string &file, signed char /* fragile */)
{
    TexRow
        texrow;

    return SimpleTeXOneTablePar(file, texrow);
}


void InsetTabular::Validate(LaTeXFeatures &features) const
{
    if (tabular->IsLongTable())
        features.longtable = true;
}


bool InsetTabular::calculate_width_of_cells()
{
    int
        cell=-1,
        i,j,p=0,
        width,
        asc,desc,
        maxAsc, maxDesc;
    
    asc = desc = 0;
    for(i=0; i<tabular->rows; ++i) {
        maxAsc = maxDesc = 0;
        for(j=0; j<tabular->columns; ++j) {
            if (tabular->IsPartOfMultiColumn(i,j)) {
                continue;
            }
            ++cell;
            width = 0;
            for(; (p < par->last()); ++p) {
                SingleHeight(par, p, asc, desc);
                if (maxAsc < asc)
                    maxAsc = asc;
                if (maxDesc < desc)
                    maxDesc = desc;
		if (par->IsNewline(p))
		    break;
                width += SingleWidth(par, p);
            }
	    ++p; // p is now on META_NEWLINE
            tabular->SetWidthOfCell(cell,width);
        }
        tabular->SetAscentOfRow(i,maxAsc+ADD_TO_HEIGHT);
        tabular->SetDescentOfRow(i,maxDesc+ADD_TO_HEIGHT);
    }
    the_locking_inset=0;
    return true;
}


// Returns the width of a character at a certain spot
int InsetTabular::SingleWidth(LyXParagraph *par, int pos)
{
    LyXFont
        font = GetFont(par, pos);
    char
        c = par->GetChar(pos);

    // The most common case is handled first (Asger)
    if (IsPrintable(c)) {
        return font.width(c);
    } else if (c == LYX_META_INSET) {
        Inset *tmpinset=par->GetInset(pos);
        if (tmpinset)
            return par->GetInset(pos)->width(buffer->getUser()->getPainter(),
                                             font);
        else
            return 0;
    } else if (IsSeparatorChar(c))
        c = ' ';
    else if (IsNewlineChar(c))
        c = 'n';
    return font.width(c);
}


// Returns the width of a character at a certain spot
void InsetTabular::SingleHeight(LyXParagraph *par,int pos, int &asc, int &desc)
{
    LyXFont
        font = GetFont(par, pos);
    char
        c = par->GetChar(pos);

    asc = desc = 0;
    // The most common case is handled first (Asger)
    if (c == LYX_META_INSET) {
        Inset *tmpinset=par->GetInset(pos);
        if (tmpinset) {
            asc = tmpinset->ascent(buffer->getUser()->getPainter(),font);
            desc = tmpinset->descent(buffer->getUser()->getPainter(),font);
        }
    } else {
        asc = font.maxAscent();
        desc = font.maxDescent();
    }
    return;
}


// Gets the fully instantiated font at a given position in a paragraph
// Basically the same routine as LyXParagraph::getFont() in paragraph.C.
// The difference is that this one is used for displaying, and thus we
// are allowed to make cosmetic improvements. For instance make footnotes
// smaller. (Asger)
// If position is -1, we get the layout font of the paragraph.
// If position is -2, we get the font of the manual label of the paragraph.
LyXFont InsetTabular::GetFont(LyXParagraph* par, int pos)
{
    char par_depth = par->GetDepth();

    LyXLayout const *layout =
	    textclasslist.style(buffer->params.textclass, par->GetLayout());

    // We specialize the 95% common case:
    if (par->footnoteflag == LyXParagraph::NO_FOOTNOTE && !par_depth) {
        if (pos >= 0) {
            // 95% goes here
            if (layout->labeltype == LABEL_MANUAL
                && pos < BeginningOfMainBody(par)) {
		// 1% goes here
                return par->GetFontSettings(pos).realize(layout->reslabelfont);
            } else
                return par->GetFontSettings(pos).realize(layout->resfont);
        } else {
            // 5% goes here.
            // process layoutfont for pos == -1 and labelfont for pos < -1
            if (pos == -1)
                return layout->resfont;
            else
                return layout->reslabelfont;
        }
    }
    // The uncommon case need not be optimized as much

    LyXFont layoutfont, tmpfont;

    if (pos >= 0){
        // 95% goes here
        if (pos < BeginningOfMainBody(par)) {
            // 1% goes here
            layoutfont = layout->labelfont;
        } else {
            // 99% goes here
            layoutfont = layout->font;
        }
        tmpfont = par->GetFontSettings(pos);
        tmpfont.realize(layoutfont);
    } else{
        // 5% goes here.
        // process layoutfont for pos == -1 and labelfont for pos < -1
        if (pos == -1)
            tmpfont = layout->font;
        else
            tmpfont = layout->labelfont;
    }
    
    // Resolve against environment font information
    //if (par->GetDepth()){ // already in while condition
    while (par && par_depth && !tmpfont.resolved()) {
        par = par->DepthHook(par_depth - 1);
        if (par) {
            tmpfont.realize(textclasslist.style(buffer->params.textclass,
                                           par->GetLayout())->font);
            par_depth = par->GetDepth();
        }
    }
    tmpfont.realize(textclasslist.textClass(buffer->params.textclass)->
		    defaultfont);
    return tmpfont;
}


int InsetTabular::BeginningOfMainBody(LyXParagraph *par)
{
    if (textclasslist.style(buffer->params.textclass,
                       par->GetLayout())->labeltype != LABEL_MANUAL)
        return 0;
    else
        return par->BeginningOfMainBody();
}


void InsetTabular::DrawCellLines(Painter &pain, int x, int baseline,
                                 int row, int cell)
{
	// Juregen, have you thought about drawing the on-off lines in a
	// different color (gray of some kind), especially since those
	// lines will not be there on the hardcopy anyway. (Lgb)
    int
        x_old = x;
    bool
        on_off;

    x_old -= tabular->WidthOfColumn(cell);
    on_off = !tabular->TopLine(cell);
    if ((!on_off || !tabular->TopAlreadyDrawed(cell)) &&
        !tabular->IsContRow(cell))
        pain.line(x_old, baseline - tabular->AscentOfRow(row),
                  x, baseline -  tabular->AscentOfRow(row),
                  LColor::tableline,
                  on_off ? Painter::line_onoffdash:Painter::line_solid);
    on_off = !tabular->BottomLine(cell);
    if ((!on_off && !tabular->RowHasContRow(cell)) || 
        tabular->VeryLastRow(cell))
        pain.line(x_old ,baseline +  tabular->DescentOfRow(row),
                  x, baseline +  tabular->DescentOfRow(row),
                  LColor::tableline,
                  on_off ? Painter::line_onoffdash:Painter::line_solid);
    on_off = !tabular->LeftLine(cell);
    pain.line(x_old, baseline -  tabular->AscentOfRow(row),
              x_old, baseline +  tabular->DescentOfRow(row),
              LColor::tableline,
              on_off ? Painter::line_onoffdash:Painter::line_solid);
    on_off = !tabular->RightLine(cell);
    pain.line(x - tabular->AdditionalWidth(cell),
              baseline -  tabular->AscentOfRow(row),
              x - tabular->AdditionalWidth(cell),
              baseline +  tabular->DescentOfRow(row),
              LColor::tableline,
              on_off ? Painter::line_onoffdash:Painter::line_solid);
}


void InsetTabular::GetCursorPos(int& x, int& y)
{
    x = cx-top_x;
    y = cy;
}


void InsetTabular::ToggleInsetCursor()
{
    if (the_locking_inset) {
        the_locking_inset->ToggleInsetCursor();
	return;
    }
    int
	x=0,y=0,
        asc,desc;
    LyXFont
        font = GetFont(par, actpos);

    asc = font.maxAscent();
    desc = font.maxDescent();
  
    if (cursor_visible)
        bview->hideLockedInsetCursor(cx+x, cy+y, asc, desc);
    else
        bview->showLockedInsetCursor(cx+x, cy+y, asc, desc);
    cursor_visible = !cursor_visible;
}


void InsetTabular::ShowInsetCursor()
{
    if (!cursor_visible) {
	int
	    asc,desc;
	LyXFont
	    font = GetFont(par, actpos);
    
	asc = font.maxAscent();
	desc = font.maxDescent();
	bview->fitLockedInsetCursor(cx, cy, asc, desc);
	bview->showLockedInsetCursor(cx, cy, asc, desc);
	cursor_visible = true;
    }
}


void InsetTabular::HideInsetCursor()
{
    if (cursor_visible)
        ToggleInsetCursor();
}


void InsetTabular::setPos(int x, int y, bool /*activate_inset*/)
{
    int
        lx, ly,
        cell,
	swh,
        sw=0;

    // search right X-pos x==0 -> top_x
    actpos = actcell = actrow = actcol = 0;
    cy = 0;
    ly = tabular->DescentOfRow(actrow);
    while((ly < y) && (actrow < tabular->rows)) {
        cy += tabular->DescentOfRow(actrow) + tabular->AscentOfRow(actrow+1) +
            tabular->AdditionalHeight(tabular->GetCellNumber(actcol,actrow+1));
        actrow++;
        actcell = tabular->GetCellNumber(actcol, actrow);
        ly = cy + tabular->DescentOfRow(actrow);
    }
    // now search the right cell & position
    for(cell=0;cell<actcell;++cell) {
        for(;(actpos < par->last()) && !par->IsNewline(actpos);++actpos)
            ;
        if (par->IsNewline(actpos))
            ++actpos;
    }
    cx = top_x;
    x += top_x;
    while((actpos < par->last()) && (cx < x)) {
        lx = cx + tabular->WidthOfColumn(actcell);
        // step empty cells
        while((actpos < par->last()) && par->IsNewline(actpos) && (lx < x)) {
            cx += tabular->WidthOfColumn(actcell++);
            lx = cx + tabular->WidthOfColumn(actcell);
            ++actcol;
            while(tabular->IsPartOfMultiColumn(actrow,actcol))
                ++actcol;
            ++actpos;
        }
        if (actpos >= par->last()) {
            cx += tabular->GetBeginningOfTextInCell(actcell);
            break;
        }
        sw = swh = SingleWidth(par,actpos);
	if (par->GetChar(actpos)!=LYX_META_INSET)
	    swh /= 2;
        lx = cx + tabular->GetBeginningOfTextInCell(actcell);
        while ((actpos < par->last()) && !par->IsNewline(actpos) &&
               ((lx+swh) < x)) {
            lx += sw;
            ++actpos;
            sw = swh = SingleWidth(par,actpos);
	    if (par->GetChar(actpos)!=LYX_META_INSET)
		swh /= 2;
        }
        if (((lx+swh) >= x) || ((cx+tabular->WidthOfColumn(actcell)) >= x)) {
            cx = lx;
            break;
        } else if (actcol >= (tabular->columns-1)) {
            cx = lx;
            break;
        } else if ((actpos < par->last()) && par->IsNewline(actpos)) {
            cx += tabular->WidthOfColumn(actcell++);
            ++actpos;
            ++actcol;
        } else {
            cx = lx;
            --actcol;
            break;
        }
    }
    while(tabular->IsPartOfMultiColumn(actrow,actcol))
        ++actcol;
    if (cx == top_x)
        cx += tabular->GetBeginningOfTextInCell(actcell);
    if (par->GetChar(actpos)==LYX_META_INSET) {
        the_locking_inset=(UpdatableInset*)par->GetInset(actpos);
	inset_x = cx-top_x;
	inset_y = cy;
	inset_pos = actpos;
        the_locking_inset->Edit(cx - inset_x, cy - inset_y, 0, bview);
    }
}

void InsetTabular::setWidthOfCell(int pos, int cell, int row)
{
    int
        width = 0,
        asc, desc,
        maxAsc = tabular->AscentOfRow(row)-ADD_TO_HEIGHT,
        maxDesc = tabular->DescentOfRow(row)-ADD_TO_HEIGHT;

    // search start of cell
    for(;(pos > 0) && !par->IsNewline(pos-1);--pos)
        ;
    for(;(pos < par->last()) && !par->IsNewline(pos);++pos) {
        width += SingleWidth(par, pos);
        SingleHeight(par, pos, asc, desc);
        if (maxAsc < asc)
            maxAsc = asc;
        if (maxDesc < desc)
            maxDesc = desc;
    }
    tabular->SetWidthOfCell(cell,width);
    tabular->SetAscentOfRow(row,maxAsc+ADD_TO_HEIGHT);
    tabular->SetDescentOfRow(row,maxDesc+ADD_TO_HEIGHT);
}

bool InsetTabular::moveRight(bool /*activate_inset*/)
{
    if (actpos < par->last()) {
        if (par->GetChar(actpos)==LYX_META_INSET) {
	    inset_x = cx-top_x;
	    inset_y = cy;
	    inset_pos = actpos;
            the_locking_inset=(UpdatableInset*)par->GetInset(actpos);
            the_locking_inset->Edit(0, 0, 0, bview);
            return true;
        }
        if (par->IsNewline(actpos)) {
            // search begin of cell
            for(--actpos;(actpos >= 0) && !par->IsNewline(actpos);--actpos)
                cx -= SingleWidth(par, actpos);
            cx -= tabular->GetBeginningOfTextInCell(actcell);
            cx += tabular->WidthOfColumn(actcell);
            for(++actcol;(actcol < tabular->columns) &&
                    tabular->IsPartOfMultiColumn(actrow, actcol);++actcol)
                ;
            if (actcol < tabular->columns) {
                cx += tabular->GetBeginningOfTextInCell(actcell+1);
                setPos(cx-top_x,cy);
            } else {
                cx = top_x;
                old_x = -1;
                return moveDown();
            }
        } else {
            cx += SingleWidth(par, actpos);
            ++actpos;
        }
    } else {
        return false;
    }
    return true;
}

bool InsetTabular::moveLeft(bool activate_inset=true)
{
    if (actpos > 0) {
        if (par->IsNewline(actpos-1)) {
            if (actcol > 0) {
                cx -= (tabular->GetBeginningOfTextInCell(actcell) +
                       tabular->AdditionalWidth(actcell-1));
            } else {
                cx = tabular->WidthOfTable() + top_x;
                old_x = -1;
                moveUp();
            }
            setPos(cx-top_x,cy);
        } else {
            cx -= SingleWidth(par, actpos-1);
            --actpos;
	    if (activate_inset && par->GetChar(actpos)==LYX_META_INSET) {
		the_locking_inset=(UpdatableInset*)par->GetInset(actpos);
		resetPos();
		inset_x = cx-top_x;
		inset_y = cy;
		inset_pos = actpos;
		the_locking_inset->Edit(the_locking_inset->
					width(bview->getPainter(),
					      GetFont(par,actpos)),
					0, 0, bview);
	    }
        }
    } else {
        return false;
    }
    return true;
}

bool InsetTabular::moveUp(bool activate_inset=true)
{
    if (old_x >= 0)
        cx = old_x;
    else
        old_x = cx;
    if (actrow > 0) {
        cy -= (tabular->AscentOfRow(actrow) + tabular->DescentOfRow(actrow-1) +
            tabular->AdditionalHeight(actcell));
        setPos(cx-top_x,cy,activate_inset);
    } else
        return false;
    return true;
}

bool InsetTabular::moveDown(bool activate_inset=true)
{
    if (old_x >= 0)
        cx = old_x;
    else
        old_x = cx;
    if ((actrow + 1) < tabular->rows) {
        cy += tabular->DescentOfRow(actrow) + tabular->AscentOfRow(actrow+1) +
            tabular->AdditionalHeight(tabular->GetCellNumber(actcol,actrow+1));
        setPos(cx-top_x,cy,activate_inset);
    } else
        return false;
    return true;
}

bool InsetTabular::moveNextCell()
{
    // search begin of cell
    if (actcell >= tabular->GetNumberOfCells())
        return false;
    for(++actcol;(actcol < tabular->columns) &&
            tabular->IsPartOfMultiColumn(actrow, actcol);++actcol)
        ;
    if (actcol < tabular->columns) {
        for(--actpos;(actpos >= 0) && !par->IsNewline(actpos);--actpos)
            cx -= SingleWidth(par, actpos);
        cx -= tabular->GetBeginningOfTextInCell(actcell);
        cx += tabular->WidthOfColumn(actcell);
        cx += tabular->GetBeginningOfTextInCell(actcell+1);
        setPos(cx-top_x,cy);
    } else {
        cx = top_x;
        old_x = -1;
        moveDown();
    }
    return true;
}

bool InsetTabular::movePrevCell()
{
    return true;
}

void InsetTabular::resetPos()
{
    int
	row,
        lx,sw,
        old_pos = actpos;

    cy = 0;
    for(row=0; row < actrow; ++row) {
        cy += tabular->DescentOfRow(row) + tabular->AscentOfRow(row+1) +
            tabular->AdditionalHeight(tabular->GetCellNumber(0,row+1));
    }
    setPos(0,cy);
    cx = top_x;
    while(actpos < old_pos) {
        lx = cx + tabular->WidthOfColumn(actcell);
        // step empty cells
        while((actpos < old_pos) && par->IsNewline(actpos)) {
            cx += tabular->WidthOfColumn(actcell++);
            lx = cx + tabular->WidthOfColumn(actcell);
            ++actcol;
            while(tabular->IsPartOfMultiColumn(actrow,actcol))
                ++actcol;
            ++actpos;
        }
        if (actpos >= old_pos) {
            cx += tabular->GetBeginningOfTextInCell(actcell);
            break;
        }
        sw = SingleWidth(par,actpos);
        lx = cx + tabular->GetBeginningOfTextInCell(actcell);
        while ((actpos < old_pos) && !par->IsNewline(actpos)) {
            lx += sw;
            ++actpos;
            sw = SingleWidth(par,actpos);
        }
        if (actpos >= old_pos)
            cx = lx;
    }
    while(tabular->IsPartOfMultiColumn(actrow,actcol))
        ++actcol;
    if (cx == top_x)
        cx += tabular->GetBeginningOfTextInCell(actcell);
}

bool InsetTabular::Delete()
{
    if (par->IsNewline(actpos))
        return false;

    /* some insets are undeletable here */
    if (par->GetChar(actpos)==LYX_META_INSET) {
        /* force complete redo when erasing display insets */ 
        /* this is a cruel mathod but save..... Matthias */ 
        if (par->GetInset(actpos)->Deletable() &&
            par->GetInset(actpos)->display()) {
            par->Erase(actpos);
            return true;
        }
        return false;
    }
    par->Erase(actpos);
    return true;
}

void  InsetTabular::SetFont(LyXFont const & font)
{
     fprintf(stderr,"SetFont called\n");
    // if there is no selection just set the current_font
    if (!hasCharSelection()) {
	// Determine basis font
	LyXFont layoutfont;
	if (actpos < BeginningOfMainBody(par))
	    layoutfont = GetFont(par, -2);
	else
	    layoutfont = GetFont(par, -1);
	
	// Update current font
	real_current_font.update(font);
	
	// Reduce to implicit settings
	current_font = real_current_font;
	current_font.reduce(layoutfont);
	// And resolve it completely
	real_current_font.realize(layoutfont);
	return;
    }
    int
	s_start, s_end;

    if (sel_pos_start > sel_pos_end) {
	s_start = sel_pos_end;
	s_end = sel_pos_start;
    } else {
	s_start = sel_pos_start;
	s_end = sel_pos_end;
    }
    LyXFont
	newfont;
    while(s_start < s_end) {
	newfont = GetFont(par,s_start);
	newfont.update(font);
	SetCharFont(s_start, newfont);
	++s_start;
    }
    calculate_width_of_cells();
    UpdateLocal();
}

void InsetTabular::SetCharFont(int pos, LyXFont font)
{
    /* let the insets convert their font */ 
    if (par->GetChar(pos) == LYX_META_INSET) {
	if (par->GetInset(pos))
	    font = par->GetInset(pos)->ConvertFont(font);
    }
    LyXLayout const
	*layout = textclasslist.style(buffer->params.textclass,
				      par->GetLayout());

    // Get concrete layout font to reduce against
    LyXFont layoutfont;

    if (pos < BeginningOfMainBody(par))
	layoutfont = layout->labelfont;
    else
	layoutfont = layout->font;


    layoutfont.realize(textclasslist.textClass(buffer->params.textclass)->
		       defaultfont);

    // Now, reduce font against full layout font
    font.reduce(layoutfont);

    par->SetFont(pos, font);
}

void InsetTabular::TableFeatures(int feature, string val="")
{
    int
	i,
	selection_start,
	selection_end,
        setLines = 0,
        setAlign = LYX_ALIGN_LEFT,
        lineSet;
    bool
        what;

    switch (feature) {
      case LyXTable::ALIGN_LEFT:
          setAlign=LYX_ALIGN_LEFT;
          break;
      case LyXTable::ALIGN_RIGHT:
          setAlign=LYX_ALIGN_RIGHT;
          break;
      case LyXTable::ALIGN_CENTER:
          setAlign=LYX_ALIGN_CENTER;
          break;
      default:
          break;
    }
    if (hasCellSelection()) {
	if (sel_cell_start > sel_cell_end) {
	    selection_start = sel_cell_end;
	    selection_end = sel_cell_start;
	} else {
	    selection_start = sel_cell_start;
	    selection_end = sel_cell_end;
	}
    } else
	selection_start = selection_end = actcell;
    switch (feature) {
      case LyXTable::SET_PWIDTH:
          tabular->SetPWidth(actcell,val);
          break;
      case LyXTable::SET_SPECIAL_COLUMN:
      case LyXTable::SET_SPECIAL_MULTI:
          tabular->SetAlignSpecial(actcell,val,feature);
          break;
      case LyXTable::APPEND_ROW:
      {
	  int
              pos = actpos,
              cell_org = actcell,
              cell = actcell;

          // if there is a ContRow following this row I have to add
          // the row after the ContRow's
          if ((pos < par->last()) && tabular->RowHasContRow(cell_org)) {
              while((pos < par->last()) && !tabular->IsContRow(cell)) {
                  while (pos < par->last() && !par->IsNewline(pos))
                      ++pos;
                  if (pos < par->last())
                      ++pos;
                  ++cell;
              }
              while((pos < par->last()) && tabular->IsContRow(cell)) {
                  while (pos < par->last() && !par->IsNewline(pos))
                      ++pos;
                  if (pos < par->last())
                      ++pos;
                  ++cell;
              }
              cell_org = --cell;
              if (pos < par->last())
                  --pos;
          }
          while ((pos < par->last()) && 
                 ((cell == cell_org) || !tabular->IsFirstCell(cell))) {
              while ((pos < par->last()) && !par->IsNewline(pos))
                  ++pos;
              if (pos < par->last())
                  ++pos;
              ++cell;
          }
          // insert the new cells
          int number = tabular->NumberOfCellsInRow(cell_org);
          for (i=0; i<number; ++i)
              par->InsertChar(pos, LYX_META_NEWLINE);
		
          // append the row into the table
          tabular->AppendRow(cell_org);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      }
      case LyXTable::APPEND_CONT_ROW:
      {
	  int
              pos = actpos,
              cell_org = actcell,
              cell = actcell;

          // if there is already a controw but not for this cell
          // the AppendContRow sets only the right values but does
          // not actually add a row
          if (tabular->RowHasContRow(cell_org) &&
              (tabular->CellHasContRow(cell_org) < 0)) {
              tabular->AppendContRow(cell_org);
              calculate_width_of_cells();
              UpdateLocal();
              return;
          }
          while ((pos < par->last()) &&
                 ((cell == cell_org) || !tabular->IsFirstCell(cell))) {
              while (pos < par->last() && !par->IsNewline(pos))
                  ++pos;
              if (pos < par->last())
                  ++pos;
              ++cell;
          }
          // insert the new cells
          int number = tabular->NumberOfCellsInRow(cell_org);
          for (i=0; i<number; ++i)
              par->InsertChar(pos, LYX_META_NEWLINE);
          // append the row into the table
          tabular->AppendContRow(cell_org);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      }
      case LyXTable::APPEND_COLUMN:
      { 
	  int
              pos = 0,
              cell_org = actcell,
              cell = 0;
          do {
              if (pos && (par->IsNewline(pos-1))) {
                  if (tabular->AppendCellAfterCell(cell_org, cell)) {
                      par->InsertChar(pos, LYX_META_NEWLINE);
                      if (pos <= actpos)
                          ++actpos;
                      ++pos;
                  }
                  ++cell;
              }
              ++pos;
          } while (pos <= par->last());
          // remember that the very last cell doesn't end with a newline.
          // This saves one byte memory per table ;-)
          if (tabular->AppendCellAfterCell(cell_org, cell))
              par->InsertChar(par->last(), LYX_META_NEWLINE);
          // append the column into the table
          tabular->AppendColumn(cell_org);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      }
      case LyXTable::DELETE_ROW:
          RemoveTableRow();
          calculate_width_of_cells();
          UpdateLocal();
          return;
      case LyXTable::DELETE_COLUMN:
      {
          int pos = 0;
          int cell = 0;
          do {
              if (!pos || (par->IsNewline(pos-1))){
                  if (tabular->DeleteCellIfColumnIsDeleted(cell, actcell)) {
                      // delete one cell
                      while (pos < par->last() && !par->IsNewline(pos))
                          par->Erase(pos);
                      if (pos < par->last())
                          par->Erase(pos);
                      else 
                          par->Erase(pos - 1); // the missing newline
                                               // at the end of a table
                      --pos; // because of ++pos below
                  }
                  ++cell;
              }
              ++pos;
          } while (pos <= par->last());
          /* delete the column from the table */ 
          tabular->DeleteColumn(actcell);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      }
      case LyXTable::TOGGLE_LINE_TOP:
          lineSet = !tabular->TopLine(actcell);
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetTopLine(i,lineSet);
          calculate_width_of_cells();
          UpdateLocal();
          return;
    
      case LyXTable::TOGGLE_LINE_BOTTOM:
          lineSet = !tabular->BottomLine(actcell); 
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetBottomLine(i,lineSet);
          calculate_width_of_cells();
          UpdateLocal();
          return;
		
      case LyXTable::TOGGLE_LINE_LEFT:
          lineSet = !tabular->LeftLine(actcell);
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetLeftLine(i,lineSet);
          calculate_width_of_cells();
          UpdateLocal();
          return;

      case LyXTable::TOGGLE_LINE_RIGHT:
          lineSet = !tabular->RightLine(actcell);
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetRightLine(i,lineSet);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      case LyXTable::ALIGN_LEFT:
      case LyXTable::ALIGN_RIGHT:
      case LyXTable::ALIGN_CENTER:
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetAlignment(i,setAlign);
          UpdateLocal();
          return;
      case LyXTable::MULTICOLUMN:
      {
	  if (tabular->row_of_cell(selection_start) !=
	      tabular->row_of_cell(selection_end)) {
	      WriteAlert(_("Impossible Operation!"), 
			 _("Multicolumns can only be horizontally."), 
			 _("Sorry."));
	      return;
	  }
	  // just multicol for one Single Cell
	  if (!hasCellSelection()) {
	      // check wether we are completly in a multicol
	      if (tabular->IsMultiColumn(actcell)) {
		  int
		      newlines,
		      pos = actpos;
		  if ((newlines=tabular->UnsetMultiColumn(actcell))) {
		      while ((pos < par->last()) && !par->IsNewline(pos))
			  ++pos;
		      for (;newlines;--newlines)
			  par->InsertChar(pos, LYX_META_NEWLINE);
		  }
		  calculate_width_of_cells();
	      } else {
		  tabular->SetMultiColumn(actcell, 1);
	      }
	      UpdateLocal();
	      return;
	  }
	  // we have a selection so this means we just add all this
	  // cells to form a multicolumn cell
	  int
	      number = 1,
	      s_start, s_end;

	  if (sel_pos_start > sel_pos_end) {
	      s_start = sel_pos_end;
	      s_end = sel_pos_start;
	  } else {
	      s_start = sel_pos_start;
	      s_end = sel_pos_end;
	  }
	  for(i=s_start; i < s_end; ++i) {
	      if (par->IsNewline(i)) {
		  par->Erase(i);
		  // check for double-blanks
		  if ((i && !par->IsLineSeparator(i-1)) &&
		      (i < par->last()) && !par->IsLineSeparator(i))
		      par->InsertChar(i, ' ');
		  else
		      --i;
		  ++number;
	      }
	  }
	  tabular->SetMultiColumn(selection_start,number);
	  actpos = s_start;
	  sel_cell_end = sel_cell_start;
	  calculate_width_of_cells();
	  UpdateLocal();
          return;
      }
      case LyXTable::SET_ALL_LINES:
          setLines = 1;
      case LyXTable::UNSET_ALL_LINES:
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetAllLines(i, setLines);
          calculate_width_of_cells();
          UpdateLocal();
          return;
      case LyXTable::SET_LONGTABLE:
          tabular->SetLongTable(true);
	  UpdateLocal(); // because this toggles displayed
          return;
      case LyXTable::UNSET_LONGTABLE:
          tabular->SetLongTable(false);
	  UpdateLocal(); // because this toggles displayed
          return;
      case LyXTable::SET_ROTATE_TABLE:
          tabular->SetRotateTable(true);
          return;
      case LyXTable::UNSET_ROTATE_TABLE:
          tabular->SetRotateTable(false);
          return;
      case LyXTable::SET_ROTATE_CELL:
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetRotateCell(i,true);
          return;
      case LyXTable::UNSET_ROTATE_CELL:
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetRotateCell(i,false);
          return;
      case LyXTable::SET_LINEBREAKS:
          what = !tabular->Linebreaks(tabular->FirstVirtualCell(actcell));
	  for(i=selection_start; i<=selection_end; ++i)
	      tabular->SetLinebreaks(i,what);
          return;
      case LyXTable::SET_LTFIRSTHEAD:
          tabular->SetLTHead(actcell,true);
          return;
      case LyXTable::SET_LTHEAD:
          tabular->SetLTHead(actcell,false);
          return;
      case LyXTable::SET_LTFOOT:
          tabular->SetLTFoot(actcell,false);
          return;
      case LyXTable::SET_LTLASTFOOT:
          tabular->SetLTFoot(actcell,true);
          return;
      case LyXTable::SET_LTNEWPAGE:
          what = !tabular->LTNewPage(actcell);
          tabular->SetLTNewPage(actcell,what);
          return;
    }
}

void InsetTabular::RemoveTableRow()
{
    int
        cell_org = actcell;

    // move to first cell in this row.
    while(!tabular->IsFirstCell(actcell)) {
        for(;(actpos > 0) && !par->IsNewline(actpos-1);--actpos)
            ;
        --actpos;
        --actcell;
    }
    for(;(actpos > 0) && !par->IsNewline(actpos-1);--actpos)
        ;
    int cell = actcell;
    // now we have to pay attention if the actual table is the
    // main row of TableContRows and if yes to delete all of them
    do {
        /* delete up to the next row */
        while ((actpos < par->last()) &&
               ((actcell == cell) || !tabular->IsFirstCell(actcell))) {
            while ((actpos < par->last()) && !par->IsNewline(actpos))
                par->Erase(actpos);
            ++actcell;
            if (actpos < par->last())
                par->Erase(actpos);
        }
        if (actpos && (actpos >= par->last())) {
            --actpos;
            par->Erase(actpos); // no newline at the very end!
            break;
        }
    } while (!tabular->IsContRow(cell_org) && tabular->IsContRow(actcell));
    tabular->DeleteRow(cell_org);
    return;
}

bool InsetTabular::InsertInset(Inset *inset)
{
    par->InsertChar(actpos, LYX_META_INSET);
    par->InsertInset(actpos, inset);
    UpdateLocal();
    the_locking_inset = (UpdatableInset*)inset;
    inset_x = cx-top_x;
    inset_y = cy;
    inset_pos = actpos;
    inset->Edit(0, 0, 0, bview);
    return true;
}

// This one spits out the text of a table paragraph
bool InsetTabular::SimpleTeXOneTablePar(string &file, TexRow &texrow)
{
	lyxerr[Debug::LATEX] << "SimpleTeXOneTablePar... "
			     << (void*)this << '\n';
    char c;
    int column, tmp;
    
    bool return_value = false;
    int current_cell_number = -1;

    LyXLayout const * style =
        textclasslist.style(getCurrentTextClass(buffer), par->GetLayout());
    LyXFont basefont;

    basefont = par->getFont(-1); // Get layout font
    // Which font is currently active?
    LyXFont running_font = basefont;
    // Do we have an open font change?
    bool open_font = false;
 
    column = 0;
    texrow.start(par, 0);
    tmp = tabular->TexEndOfCell(file,current_cell_number);
    for (;tmp>0;--tmp)
        texrow.newline();
    for (int i = 0; i < par->last(); ++i) {
        c = par->GetChar(i);
        if (tabular->IsContRow(current_cell_number+1)) {
            if (c == LYX_META_NEWLINE)
                ++current_cell_number;
            continue;
        }
        ++column;
        // Fully instantiated font
        LyXFont font = par->getFont(i);

        // Spaces at end of font change are simulated to be outside
        // font change. i.e. we write "\textXX{text} " rather
        // than "\textXX{text }". (Asger)
        if (open_font && c == ' ' && i <= par->last()-2 &&
            par->getFont(i+1) != running_font && par->getFont(i+1) != font) {
            font = par->getFont(i+1);
        }
        // We end font definition before blanks
        if (font != running_font && open_font) {
            column += running_font.latexWriteEndChanges(file, basefont);
            running_font = basefont;
            open_font = false;
        }
        // Blanks are printed before start of fontswitch
        if (c == ' '){
            par->SimpleTeXBlanks(file, texrow, i, column, font, style);
        }
        // Do we need to change font?
        if (font != running_font) {
            column += font.latexWriteStartChanges(file, basefont);
            running_font = font;
            open_font = true;
        }
        // Do we need to turn on LaTeX mode?
        if (font.latex() != running_font.latex()) {
            if (font.latex() == LyXFont::ON && style->needprotect) {
                file += "\\protect ";
                column += 9;
            }
        }
        if (c == LYX_META_NEWLINE) {
            // special case for inside a table
            // different from default case in
            // SimpleTeXSpecialChars()
            if (open_font) {
                column += running_font.latexWriteEndChanges(file, basefont);
                open_font = false;
            }
            basefont = par->getFont(-1);
            running_font = basefont;
            ++current_cell_number;
            if (tabular->CellHasContRow(current_cell_number) >= 0) {
                TeXContTableRows(file, i+1,current_cell_number,column, texrow);
            }
            // if this cell follow only ContRows till end don't
            // put the EndOfCell because it is put after the
            // for(...)
            if (tabular->ShouldBeVeryLastCell(current_cell_number)) {
                break;
            }
            int tmp = tabular->TexEndOfCell(file,current_cell_number);
            if (tmp>0) {
                column = 0;
            } else if (tmp < 0) {
                tmp = -tmp;
            }
            for (;tmp--;) {
                texrow.newline();
            }
            texrow.start(par, i+1);
        } else {
            par->SimpleTeXSpecialChars(file, texrow,
                                       font, running_font, basefont,
                                       open_font, style, i, column, c);
        }
    }
    // If we have an open font definition, we have to close it
    if (open_font) {
        running_font.latexWriteEndChanges(file, basefont);
    }
    ++current_cell_number;
    tmp = tabular->TexEndOfCell(file, current_cell_number);
    for (;tmp>0;--tmp)
        texrow.newline();
    lyxerr[Debug::LATEX] << "SimpleTeXOneTablePar...done "
			 << (void*)this << '\n';
    return return_value;
}

// This one spits out the text off ContRows in tables
bool InsetTabular::TeXContTableRows(string &file, int i,
				    int current_cell_number,
				    int &column, TexRow &texrow)
{
	lyxerr[Debug::LATEX] << "TeXContTableRows...   "
			     << (void*)this << '\n';
    char c;
    int cell;
    int lastpos;
    bool return_value = false;
    LyXLayout const * style =
        textclasslist.style(getCurrentTextClass(buffer), par->GetLayout());
    LyXFont basefont;
    basefont = par->getFont(-1); // Get layout font
    // Which font is currently active?
    LyXFont running_font = basefont;
    // Do we have an open font change?
    bool open_font = false;

    lastpos = i;
    cell = tabular->CellHasContRow(current_cell_number);
    ++current_cell_number;
    while(cell >= 0) {
        // first find the right position
        i = lastpos;
        for (; (i<par->last()) && (current_cell_number<cell); ++i) {
            c = par->GetChar(i);
            if (c == LYX_META_NEWLINE)
                ++current_cell_number;
        }
        lastpos = i;
        c=par->GetChar(i);
        if (tabular->Linebreaks(tabular->FirstVirtualCell(cell))) {
            file += " \\\\\n";
            texrow.newline();
            column = 0;
        } else if ((c != ' ') && (c != LYX_META_NEWLINE)) {
            file += ' ';
        }
        for (;i < par->last() && (c=par->GetChar(i)) != LYX_META_NEWLINE;++i) {
            ++column;
            // Fully instantiated font
            LyXFont font = par->getFont(i);
            // Spaces at end of font change are simulated to
            // be outside font change. i.e. we write
            // "\textXX{text} " rather than "\textXX{text }".
            // (Asger)
            if (open_font && c == ' ' && i <= par->last()-2 &&
                par->getFont(i+1) != running_font &&
                par->getFont(i+1) != font) {
                font = par->getFont(i+1);
            }
            // We end font definition before blanks
            if (font != running_font && open_font) {
                column += running_font.latexWriteEndChanges(file, basefont);
                running_font = basefont;
                open_font = false;
            }
            // Blanks are printed before start of fontswitch
            if (c==' ') {
                par->SimpleTeXBlanks(file, texrow, i, column, font, style);
            }
            // Do we need to change font?
            if (font != running_font) {
                column += font.latexWriteStartChanges(file, basefont);
                running_font = font;
                open_font = true;
            }
            // Do we need to turn on LaTeX mode?
            if (font.latex() != running_font.latex()) {
                if (font.latex() == LyXFont::ON && style->needprotect) {
                    file += "\\protect ";
                    column += 9;
                }
            }
            par->SimpleTeXSpecialChars(file, texrow, font,
                                       running_font, basefont,
                                       open_font, style, i, column, c);
        }
        // If we have an open font definition, we have to close it
        if (open_font) {
            running_font.latexWriteEndChanges(file, basefont);
            open_font = false;
        }
        basefont = par->getFont(-1);
        running_font = basefont;
        cell = tabular->CellHasContRow(current_cell_number);
    }
    lyxerr[Debug::LATEX] << "TeXContTableRows...done "
			 << (void*)this << '\n';
    return return_value;
}

#ifdef THIS_HAS_TO_BE_IMPLEMENTED

bool InsetTabular::RoffContTableRows(FILE *fp, int i, int actcell)
{
    LyXFont
        font1 = LyXFont(LyXFont::ALL_INHERIT),
        font2;
    Inset
        *inset;
    int
        lastpos, cell;
    char
        c;
    string
        fname2;
    FILE
        *fp2;

    fname2 = TmpFileName(string(),"RAT2");
    lastpos = i;
    cell = table->CellHasContRow(actcell);
    ++actcell;
    while(cell >= 0) {
        // first find the right position
        i = lastpos;
        for (; (i<last()) && (actcell<cell); ++i) {
            c = GetChar(i);
            if (c == LYX_META_NEWLINE)
                ++actcell;
        }
        lastpos = i;
        c=GetChar(i);
        if ((c != ' ') && (c != LYX_META_NEWLINE))
            fprintf(fp," ");
        for (; i < last() && (c=GetChar(i)) != LYX_META_NEWLINE; ++i) {
            font2 = GetFontSettings(i);
            if (font1.latex() != font2.latex()) {
                if (font2.latex() != LyXFont::OFF)
                    continue;
            }
            c = GetChar(i);
            switch (c) {
              case LYX_META_INSET:
                  if ((inset = GetInset(i))) {
                      if (!(fp2=fopen(fname2.c_str(),"w+"))) {
			      lyxerr << _("LYX_ERROR: Cannot open temporary file: ") << fname2 << '\n';
                          return false;
                      }
                      inset->Latex(fp2,-1);
                      rewind(fp2);
                      c = fgetc(fp2);
                      while(!feof(fp2)) {
                          if (c == '\\')
                              fprintf(fp,"\\\\");
                          else
                              fputc(c,fp);
                          c = fgetc(fp2);
                      }
                      fclose(fp2);
                  }
                  break;
              case LYX_META_NEWLINE:
                  break;
              case '\\': 
                  fprintf(fp, "\\\\");
                  break;
              default:
                  if (c != '\0')
                      fprintf(fp, "%c", c);
                  else
			  lyxerr[Debug::INFO] << "RoffAsciiTable: "
				  "0 char in structure.\n";
                  break;
            }
        }
        cell = table->CellHasContRow(actcell);
    }
    return true;
}
#endif
