#include <config.h>

#include "buffer.h"
#include "bufferlist.h"
#include "lastfiles.h"
#include "lyx_gui_misc.h"
#include "LString.h"
#include "lstrings.h"
#include "lyx_main.h"
#include "filedlg.h"
#include "minibuffer.h"
#include "FileInfo.h"
#include "path.h"
#include "lyxrc.h"
#include "syscall.h"
#include "lyx.h"
#include "print_form.h"
#include "insetlabel.h"
#include "LyXView.h"
#include "BufferView.h"
#include "lyxscreen.h"
#include "lyxtext.h"

extern BufferList bufferlist;
extern MiniBuffer *minibuffer;

extern FD_form_print *fd_form_print;
extern FD_form_toc *fd_form_toc;
extern FD_form_ref *fd_form_ref;

char ascii_type; /* for selection notify callbacks */
// this should be static, but I need it in buffer.C
bool quitting;	// flag, that we are quitting the program
extern bool finished; // all cleanup done just let it run through now.

extern void AllowInput();
extern void ProhibitInput();
extern bool MenuRunDvips(Buffer *buffer, bool wait=false);
extern int send_fax(string const &fname, string const &sendcmd);
extern void TocUpdateCB(FL_OBJECT *, long);
extern bool RunScript(Buffer *buffer, bool wait,
	       string const & command, string const & orgname = string(),
	       bool need_shell=true);


void BufferList::menuWrite(Buffer* buffer)
{
	if (!write(buffer)) {
		string fname = buffer->getFileName();
		string s = MakeAbsPath(fname);
		if (AskQuestion(_("Save failed. Rename and try again?"),
				 MakeDisplayPath(s,50),
				 _("(If not, document is not saved.)"))) {
			menuWriteAs(buffer);
		}
	} else {
		lastfiles->newFile(buffer->getFileName());
	}
}


void BufferList::menuWriteAs(Buffer *buffer)
{
	// bad test
	if (!buffer) return;

	string fname = buffer->getFileName();
	string oldname = fname;
	LyXFileDlg fileDlg;

	ProhibitInput();
	fileDlg.SetButton(0, _("Documents"), lyxrc->document_path);
	fileDlg.SetButton(1, _("Templates"), lyxrc->template_path);

	if (!IsLyXFilename(fname))
		fname += ".lyx";

	fname = fileDlg.Select(_("Enter Filename to Save Document as"), 
			       OnlyPath(fname),
			       "*.lyx", 
			       OnlyFilename(fname));
	AllowInput();

	if (fname.empty()) {
		minibuffer->Set(_("Canceled."));
		return;
	}

	// Make sure the absolute filename ends with appropriate suffix
	string s= MakeAbsPath(fname);
	if (!IsLyXFilename(s))
		s += ".lyx";

	// Same name as we have already?
	if (s == oldname) {
		if (!AskQuestion(_("Same name as document already has:"),
				 MakeDisplayPath(s,50),
				 _("Save anyway?")))
			return;
		// Falls through to name change and save
	} 
	// No, but do we have another file with this name open?
	else if (exists(s)) {
		if (AskQuestion(_("Another document with same name open!"),
				MakeDisplayPath(s,50),
				_("Replace with current document?"))) {
			close(getBuffer(s));
			
			// Ok, change the name of the buffer,
			// but don't save!
			buffer->setFileName(s);
			buffer->markDirty();
			
			minibuffer->Set(_("Document renamed to '"),
					MakeDisplayPath(s),
					_("', but not saved..."));
		}
		return;
	} // Check whether the file exists
	else {
		FileInfo myfile(s);
		if (myfile.isOK() &&
		    !AskQuestion(_("Document already exists:"), 
				 MakeDisplayPath(s,50),
				 _("Replace file?")))
			return;
	}

	// Ok, change the name of the buffer
	buffer->setFileName(s);
	buffer->markDirty();
	// And save
	// Small bug: If the save fails, we have irreversible changed the name
	// of the document.
	menuWrite(buffer);
}    


int BufferView::runLaTeX()
{
	int ret;

	if (buffer_->isLinuxDoc())
		ret = runLinuxDoc(1, buffer_->getFileName());
	else 
		ret = buffer_->runLaTeX();
   
	if (ret > 0) {
		string s;
		string t;
		if (ret == 1) {
			s = _("One error detected");
			t = _("You should try to fix it.");
		} else {
			s += ret;
			s += _(" errors detected.");
			t = _("You should try to fix them.");
		}
		WriteAlert(_("There were errors during the LaTeX run."),
			   s, t);
	}
	return ret;
}


int BufferView::runChktex()
{
	int ret;

	if (buffer_->isLinuxDoc()) {
		WriteAlert(_("Chktex is no fun for LinuxDoc."));
		return 0;
	} else 
		ret = buffer_->runChktex();
   
	if (ret >= 0) {
		string s;
		string t;
		if (ret == 0) {
			s = _("No warnings found.");
		} else if (ret == 1) {
			s = _("One warning found.");
			t = _("Use 'Edit->Go to Error' to find it.");
		} else {
			s += ret;
			s += _(" warnings found.");
			t = _("Use 'Edit->Go to Error' to find them.");
		}
		WriteAlert(_("Chktex run successfully"), s, t);
	} else {
		WriteAlert(_("Error!"),_("It seems chktex does not work."));
	}
	return ret;
}


// Returns false if we fail
bool MenuPreviewPS(Buffer *buffer)
{
	// bad test
	if (!buffer)
		return false;

	// Generate postscript file
	if (!MenuRunDvips(buffer, true)) {
		return false;
	}

	// Start postscript viewer
	ProhibitInput();
	string ps = ChangeExtension (buffer->getFileName(),
				      ".ps_tmp", true);
	// push directorypath, if necessary 
        string path = OnlyPath(buffer->getFileName());
        if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)){
		path = buffer->getTmpDir();
        }
        //PathPush(path);
	Path p(path);
	bool ret = RunScript(buffer, false, lyxrc->view_ps_command, ps);
        //PathPop();
	AllowInput();
	return ret;
}


void MenuFax(Buffer *buffer)
{
	// bad test
	if (!buffer)
		return;

	// Generate postscript file
	if (!MenuRunDvips(buffer, true)) {
		return;
	}

	// Send fax
	string ps = ChangeExtension (buffer->getFileName(), ".ps_tmp", true);
	string path = OnlyPath (buffer->getFileName());
	if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)) {
		path = buffer->getTmpDir();
	}
	Path p(path);
	if (!lyxrc->fax_program.empty()) {
                string help2 = subst(lyxrc->fax_program, "$$FName",ps);
                help2 += " &";
                Systemcalls one(Systemcalls::System, help2);
	} else
		send_fax(ps,lyxrc->fax_command);
}


// Returns false if we fail
bool MenuPreview(Buffer *buffer)
{
	// bad test
	if (!buffer)
		return false;
   
	string paper;
	
	char real_papersize = buffer->params.papersize;
	if (real_papersize == PAPER_DEFAULT)
		real_papersize = lyxrc->default_papersize;
   
	switch (real_papersize) {
	case PAPER_USLETTER:
		paper = "us";
		break;
	case PAPER_A3PAPER:
		paper = "a3";
		break;
	case PAPER_A4PAPER:
		paper = "a4";
		break;
	case PAPER_A5PAPER:
		paper = "a5";
		break;
	case PAPER_B5PAPER:
		paper = "b5";
		break;
	case PAPER_EXECUTIVEPAPER:
		paper = "foolscap";
		break;
	case PAPER_LEGALPAPER:
		paper = "legal";
		break;
	default: /* If nothing else fits, keep the empty value */
		break;
	}
   
	if (paper.empty()) {
		if (buffer->params.orientation == ORIENTATION_LANDSCAPE)
			// we HAVE to give a size when the page is in
			// landscape, so use USletter.		
			paper = " -paper usr";
	} else {
		paper = " -paper " + paper;
		if (buffer->params.orientation == ORIENTATION_LANDSCAPE)
			paper+='r';
	}

	// push directorypath, if necessary 
        string path = buffer->getFileDir();
        if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)){
		path = buffer->getTmpDir();
        }
	Path p(path);
	// Run dvi-viewer
	string command = lyxrc->view_dvi_command + paper ;
#ifndef __EMX__
	bool ret = RunScript(buffer, false, command);
#else
	bool ret = RunScript(buffer, false, command, string(), false);
#endif
	return ret;
}


void BufferView::MenuMakeLaTeX()
{
	// bad test
	if (buffer_) {
		// Get LaTeX-Filename
		string s = SpaceLess(ChangeExtension(
						buffer_->getFileName(),
						".tex", false));

		if (buffer_->isLinuxDoc())
			runLinuxDoc(0, buffer_->getFileName());
		else
			buffer_->makeLaTeXFile(s, string(), false);
		owner_->getMiniBuffer()->Set(_("Nice LaTeX file saved as"),
				MakeDisplayPath(s));
		buffer_->markDviDirty();
	}
}


void MenuMakeLinuxDoc(Buffer *buffer)
{
	// bad test
	if (buffer) {

		if (!buffer->isLinuxDoc()) {
			WriteAlert(_("Error!"), _("Document class must be linuxdoc."));
			return;
		}

		// Get LinuxDoc-Filename
		string s = ChangeExtension (buffer->getFileName(), 
					     ".sgml", false);
		minibuffer->Set(_("Building LinuxDoc SGML file `"),
				MakeDisplayPath(s), "'..."); 

		buffer->makeLinuxDocFile(s, 65);
                buffer->redraw();
		minibuffer->Set(_("LinuxDoc SGML file"),
				MakeDisplayPath(s), _("built")); 
	}
}


void MenuMakeAscii(Buffer *buffer)
{
	// bad test
	if (!buffer) return;
	
	/* get LaTeX-Filename */
	string s = ChangeExtension (buffer->getFileName(),
				     ".txt", false);
	buffer->writeFileAscii(s, lyxrc->ascii_linelen);
	
	minibuffer->Set(_("Ascii file saved as"), MakeDisplayPath(s));
}


void MenuPrint(Buffer *buffer)
{
	string input_file;

	// bad test
	if (!buffer) 
		return;
	
	input_file = ChangeExtension(buffer->getFileName(),
				     lyxrc->print_file_extension,
				     true);
	fl_set_input(fd_form_print->input_file, input_file.c_str());
	
	if (fd_form_print->form_print->visible) {
		fl_raise_form(fd_form_print->form_print);
	} else {
		fl_show_form(fd_form_print->form_print,
			     FL_PLACE_MOUSE | FL_FREE_SIZE, FL_FULLBORDER,
			     _("Print"));
	}
}


void QuitLyX()
{
	lyxerr[Debug::INFO] << "Running QuitLyX.\n";

	if (!bufferlist.QwriteAll()) return;
	
	lastfiles->writeFile(lyxrc->lastfiles);

	// Set a flag that we do quitting from the program,
	// so no refreshes are necessary.
	quitting = true;

	// close buffers first
	bufferlist.closeAll();

	// do any other cleanup procedures now
	lyxerr[Debug::INFO] << "Deleting tmp dir " 
			    << system_tempdir << '\n';

	DestroyLyXTmpDir(system_tempdir);

	finished = true;
}


//
// (c) CHT Software Service GmbH
// Uwe C. Schroeder
//
// create new file with template
// SERVERCMD !
//
Buffer * NewLyxFile(string const & filename)
{
	// Split argument by :
	string name;
	string tmpname = split(filename, name, ':');
#warning __EMX__ FIX ME!
#ifdef __EMX__ // Fix me! lyx_cb.C may not be low leve|l enough to allow this.
	if (name.length() == 1 && isalpha((unsigned char) name[0]) &&
	    (prefixIs(tmpname, "/") || prefixIs(tmpname, "\\"))) {
		name += ':';
		name += token(tmpname, ':');
		tmpname.split(':');
	}
#endif
	lyxerr << "Arg is " << filename
	       << "\nName is " << name
	       << "\nTemplate is " << tmpname << '\n';

	// find a free buffer 
	Buffer *tmpbuf = bufferlist.newFile(name,tmpname);
	if (tmpbuf)
		lastfiles->newFile(tmpbuf->getFileName());
	return tmpbuf;
}


// Insert ascii file
void BufferView::insertAsciiFile(string const & f, bool asParagraph)
{
	if (!screen_) return;

	string fname = f;
	LyXFileDlg fileDlg;
      
	if (fname.empty()) {
		ProhibitInput();
  
		fname =	fileDlg.Select(_("File to Insert"), 
				       buffer_->getFileDir(),
				       "*");
  
		AllowInput();
 
		if (fname.empty()) return;
	}

	FileInfo fi(fname);
	FilePtr myfile(fname, FilePtr::read);

	if (!fi.exist() || !fi.readable() || !myfile()) {
		WriteFSAlert(_("Error! Cannot open specified file:"),
			     MakeDisplayPath(fname,50));
		return;
	}
	
	// this is still a stupid way to read a file, but now
	// at least we don't use constructs not meant for it.
	long file_size = fi.getSize();
	char *filearray = new char[file_size + 1];
	if (!feof(myfile())) {
		int arraycounter = 0;
		char c = '\0';
		do {
			c = fgetc(myfile());
			filearray[arraycounter++] = c;
		} while (!feof(myfile()));
		filearray[arraycounter] = '\0';
	}
	
	// insert the string
	screen_->hideCursor();
      
	// clear the selection
	beforeChange();

	if (!asParagraph)
		text->InsertStringA(filearray);
	else
		text->InsertStringB(filearray);
	delete[] filearray;
	update(BufferView::ResetSelection
	       |BufferView::FitCursor
	       |BufferView::BufferChange);
}


void MenuShowTableOfContents()
{
	TocUpdateCB(0, 0);
	if (fd_form_toc->form_toc->visible) {
		fl_raise_form(fd_form_toc->form_toc);
	} else {
		fl_show_form(fd_form_toc->form_toc,
			     FL_PLACE_MOUSE | FL_FREE_SIZE, FL_FULLBORDER,
			     _("Table Of Contents"));
	}
}


void BufferView::insertLabel()
{
	ProhibitInput();
	string label = strip(fl_show_input(_("Enter new label to insert:"),""));
	label = frontStrip(label);
	if (!label.empty()) {
		InsetLabel *new_inset = new InsetLabel;
		new_inset->setContents(label);
		insertInset(new_inset);
	}
	AllowInput();
}


void MenuInsertRef()
{
	RefUpdateCB(0, 0);
	if (fd_form_ref->form_ref->visible) {
		fl_raise_form(fd_form_ref->form_ref);
	} else {
		fl_show_form(fd_form_ref->form_ref,
			     FL_PLACE_MOUSE | FL_FREE_SIZE, FL_FULLBORDER,
			     _("Insert Reference"));
	}
}


// We should look closer at this function
// and where it should be placed.
// and the strange argument "at"
void BufferView::pasteSelection(char at)
{
	if (!screen_) return;

	ascii_type = at;
  
	Atom data_prop = XInternAtom(fl_display, 
				     "LyX_Primary",
				     False);
	if (data_prop == None) 
		return;
	XConvertSelection(fl_display,
			  XA_PRIMARY, XA_STRING, data_prop, 
			  owner_->getForm()->window, 0);
}


/*
 * SGML support:
 * (flag == -1) import SGML file
 * (flag == 0) make TeX output
 * (flag == 1) make dvi output
 */
int BufferView::runLinuxDoc(int flag, string const & filename)
{
	string s2;
	string add_flags;

	int errorcode = 0;

	// Generate a path-less extension name.
	string name = ChangeExtension (filename, ".sgml", true);
	string path = OnlyPath (filename);
	if (lyxrc->use_tempdir || (IsDirWriteable(path) < 1)) {
		path = buffer_->getTmpDir();
	}
	Path p(path);
	
	if (flag != -1) {
		if (!available()) return 0;
		
		buffer_->makeLinuxDocFile(name,0);
		LYX_PAPER_SIZE ps = (LYX_PAPER_SIZE) buffer_->params.papersize;
		switch (ps) {
		case PAPER_A4PAPER:  add_flags = "-p a4";     break;
		case PAPER_USLETTER: add_flags = "-p letter"; break;
		default: /* nothing to be done yet ;-) */     break; 
		}
	}
	
	ProhibitInput();
	
	Systemcalls one;
	switch (flag) {
	case -1: /* Import file */
		minibuffer->Set(_("Importing LinuxDoc SGML file `"), 
				MakeDisplayPath(filename), "'...");
		s2 = "sgml2lyx " + lyxrc->sgml_extra_options + ' ' 
			+ name;
		if (one.Startscript(Systemcalls::System, s2)) 
			errorcode = 1;
		break;
	case 0: /* TeX output asked */
		minibuffer->Set(_("Converting LinuxDoc SGML to TeX file..."));
		s2 = "sgml2latex " + add_flags + " -o tex "
			+ lyxrc->sgml_extra_options + ' ' + name;
		if (one.Startscript(Systemcalls::System, s2)) 
			errorcode = 1;
		break;
	case 1: /* dvi output asked */
		minibuffer->Set(_("Converting LinuxDoc SGML to dvi file..."));
		s2 = "sgml2latex " + add_flags + " -o dvi "
			+ lyxrc->sgml_extra_options + ' ' + name;
		if (one.Startscript(Systemcalls::System, s2)) {
			errorcode = 1;
		} else
			buffer_->markDviClean();
		break;
	default: /* unknown output */
		break;
	}
	
	AllowInput();
	return errorcode;
}


void BufferView::layoutSave()
{
	if (!screen_ || !available()) return;

	if (AskQuestion(_("Do you want to save the current settings"),
			_("for Character, Document, Paper and Quotes"),
			_("as default for new documents?")))
		buffer_->saveParamsAsDefaults();
}
