// -*- C++ -*-
/* This file is part of
 * ======================================================
 *	 
 *           LyX, The Document Processor
 *        	
 *           Copyright (C) 1995 Matthias Ettrich
 *           Copyright (C) 1995-1998 The LyX Team.
 *
 * ====================================================== */

/**
   Docu   : To use the lyxserver define the name of the pipe in your
   lyxrc:
   \serverpipe "/home/myhome/.lyxpipe"
   Then use .lyxpipe.in and .lyxpipe.out to communicate to LyX.
   Each message consists of a single line in ASCII. Input lines
   (client -> LyX) have the following format:
   "LYXCMD:<clientname>:<functionname>:<argument>"
   Answers from LyX look like this:
   "INFO:<clientname>:<functionname>:<data>"
   [asierra970531] Or like this in case of error:
   "ERROR:<clientname>:<functionname>:<error message>"
           where <clientname> and <functionname> are just echoed.
           If LyX notifies about a user defined extension key-sequence,
           the line looks like this:
           "NOTIFY:<key-sequence>"
 [asierra970531] New server-only messages to implement a simple protocol
           "LYXSRV:<clientname>:<protocol message>"
           where <protocol message> can be "hello" or "bye". If hello is
           received LyX will inform the client that it's listening its
           messages, and 'bye' will inform that lyx is closing.

           See development/server_monitor.c for an example client.
  Purpose: implement a client/server lib for LyX
*/

#include <config.h>

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include FORMS_H_LOCATION

#ifdef __GNUG__
#pragma implementation
#endif

#include "lyxserver.h"
#include "lyxfunc.h"
#include "lyx_main.h"
#include "LyXAction.h"
#include "debug.h"
#include "lstrings.h"

#ifdef __EMX__
#include <stdlib.h>
#include <io.h>
#define OS2EMX_PLAIN_CHAR
#define INCL_DOSNMPIPES
#define INCL_DOSERRORS
#include <os2.h>
#include "os2_errortable.h"
#endif

// 	$Id: lyxserver.C,v 1.6 1998/10/22 16:56:18 larsbj Exp $	

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: lyxserver.C,v 1.6 1998/10/22 16:56:18 larsbj Exp $";
#endif /* lint */
	
/* === variables ========================================================= */

extern LyXAction lyxaction;

// LyXComm class
 
 // Open pipes
void LyXComm::openConnection()
{
	lyxerr[Debug::LYXSERVER] << "LyXComm: Opening connection\n";
	
	// If we are up, that's an error
	if (ready) {
		lyxerr << "LyXComm: Already connected\n";
		return;
	}
	// We assume that we don't make it
	ready = false;
	
	if (pipename.empty()) return;
	
	// --- prepare input pipe ---------------------------------------
	
	string tmp = pipename + ".in";
	
#ifdef __EMX__
	HPIPE fd;
	APIRET rc;
	int errnum;
	// Try create one instance of named pipe with the mode O_RDONLY|O_NONBLOCK.
	// The current emx implementation of access() won't work with pipes.
	rc = DosCreateNPipe(tmp.c_str(), &fd, NP_ACCESS_INBOUND,
			    NP_NOWAIT|0x01, 0600, 0600, 0);
	if (rc == ERROR_PIPE_BUSY) {
#else	
	if (access(tmp.c_str(), F_OK) == 0) {
#endif
		lyxerr << "LyXComm: Pipe " << tmp << " already exists.\n";
		lyxerr << "If no other LyX program is active, please delete"
			" the pipe by hand and try again.\n";
		pipename.erase();
		return;
	}
#ifndef __EMX__
	if (mkfifo(tmp.c_str(), 0600) < 0) {
		lyxerr << "LyXComm: Could not create pipe " << tmp << '\n';
		lyxerr << strerror(errno) << '\n';
		return;
	}
	infd = open(tmp.c_str(), O_RDONLY|O_NONBLOCK);
#else
	if (rc != NO_ERROR) {
		errnum = TranslateOS2Error(rc);
		lyxerr << "LyXComm: Could not create pipe " << tmp << '\n';
		lyxerr << strerror(errnum) << '\n';
		return;
	}
	// Listen to it.
	rc = DosConnectNPipe(fd);
	if (rc != NO_ERROR && rc != ERROR_PIPE_NOT_CONNECTED) {
		errnum = TranslateOS2Error(rc);
		lyxerr << "LyXComm: Could not create pipe " << tmp << '\n';
		lyxerr << strerror(errnum);
		return;
	}
	// Imported handles can be used both with
	// OS/2 APIs and emx library functions.
	infd = _imphandle(fd);
#endif
	if (infd < 0) {
		lyxerr << "LyXComm: Could not open pipe " << tmp << '\n';
		lyxerr << strerror(errno) << '\n';
		return;
	}
	fl_add_io_callback(infd, FL_READ, callback, (void*)this);
		
		// --- prepare output pipe ---------------------------------------
 
	tmp = pipename + ".out";
       
#ifdef __EMX__       
	rc = DosCreateNPipe(tmp.c_str(), &fd, NP_ACCESS_DUPLEX,
			    NP_NOWAIT|0x01, 0600, 0600, 0);
			
	if (rc == ERROR_PIPE_BUSY)
#else	
	if (access(tmp.c_str(), F_OK) == 0)
#endif
	{
		lyxerr << "LyXComm: Pipe " << tmp << " already exists.\n";
		lyxerr << "If no other LyX program is active, please delete"
			" the pipe by hand and try again.\n";
		pipename.erase();
		ready = true;
		lyxerr[Debug::LYXSERVER]
			<< "LyXComm: Connection established\n";
	}
}

/// Close pipes
void LyXComm::closeConnection()
{
#ifdef __EMX__
	APIRET rc;
	int errnum;
#endif
	lyxerr[Debug::LYXSERVER] << "LyXComm: Closing connection\n";

	if (pipename.empty()) {
		return;
	}

	if (!ready) {
		lyxerr << "LyXComm: Already disconnected\n";
		return;
	}
	
	if(infd > -1) {
		fl_remove_io_callback(infd, FL_READ, callback);
		
		string tmp = pipename + ".in";
#ifdef __EMX__		// Notify the operating system.
		rc = DosDisConnectNPipe(infd);
		if (rc != NO_ERROR) {
			errnum = TranslateOS2Error(rc);
			lyxerr << "LyXComm: Could not disconnect pipe "
			       << tmp << '\n';
			lyxerr << strerror(errnum) << '\n';
			return;
		}
#endif
		if (close(infd) < 0) {
			lyxerr << "LyXComm: Could not close pipe " 
			       << tmp << '\n';
			lyxerr << strerror(errno) << '\n';
						}
#ifndef __EMX__		// OS/2 named pipes will be automatically removed.
		if (unlink(tmp.c_str()) < 0){
			lyxerr << "LyXComm: Could not remove pipe "
			       << tmp << '\n';
			lyxerr << strerror(errno) << '\n';
		};
#endif
	}
	if(outfd > -1) {
		string tmp = pipename + ".out";
#ifdef __EMX__
		rc = DosDisConnectNPipe(outfd);
		if (rc != NO_ERROR) {
			errnum = TranslateOS2Error(rc);
			lyxerr << "LyXComm: Could not disconnect pipe "
			       << tmp << '\n';
			lyxerr << strerror(errnum) << '\n';
			return;
		}
#endif
		if (close(outfd) < 0) {
			lyxerr << "LyXComm: Could not close pipe "
			       << tmp << '\n';
			lyxerr << strerror(errno) << '\n';
		}
#ifndef __EMX__
		if (unlink(tmp.c_str()) < 0){
			lyxerr << "LyXComm: Could not remove pipe "
			       << tmp << '\n';
			lyxerr << strerror(errno) << '\n';
		}
#endif
	}
	ready = false;
}
 
// Receives messages and sends then to client
void LyXComm::callback(int fd, void *v)
{
	LyXComm * c = (LyXComm *) v;
	
	// --- read everything -------------------------------------------
 
	if (lyxerr.debugging(Debug::LYXSERVER)) {
		lyxerr << "LyXComm: Receiving from fd " << fd << '\n';
	}
	
	const int CMDBUFLEN = 100;
	char charbuf[CMDBUFLEN];
  	string cmd;
// nb! make lsbuf a class-member for multiple sessions
	static string lsbuf;

	errno = 0;
	int status;
	// the single = is intended here.
	while((status = read(fd,charbuf,CMDBUFLEN-1)))
	{// break and return in loop
		if(status > 0) // got something
		{
			charbuf[status]='\0'; // turn it into a c string
			lsbuf += charbuf;
			lsbuf=strip(lsbuf,'\r');
			// commit any commands read
			while(contains(lsbuf, "\n"))
			{ // while still commands left 
				// split() grabs the entire string if
				// the delim /wasn't/ found. ?:-P 
				lsbuf = split(lsbuf,cmd,'\n');
				lyxerr[Debug::LYXSERVER]
					<< "LyXComm: status:" << status
					<< ", lsbuf:" << lsbuf 
					<< ", cmd:" << cmd << '\n';
				if(!cmd.empty())
					c->clientcb(c->client, cmd); 
				        //\n or not \n?
			}
		}
		if(errno == EAGAIN)
		{  // EAGAIN is not really an error , it means we're
		   // only reading too fast for the writing process on
		   // the other end of the pipe. 
			errno = 0;
			return; // up to libforms select-loop (*crunch*)
		}
		if(errno != 0 )
		{
			lyxerr << "LyXComm: " << strerror(errno) << '\n';
			if (!lsbuf.empty()) {
				lyxerr << "LyxComm: truncated command: " 
				       << lsbuf << '\n';
				lsbuf.erase();
			}
			break; // reset connection
		}
	}
	c->closeConnection();
	c->openConnection();
	errno=0;
}

void LyXComm::send(string const & msg) {
	if (msg.empty()) {
		lyxerr << "LyXComm: Request to send empty string. Ignoring.\n";
		return;
	}
	
	if (lyxerr.debugging(Debug::LYXSERVER)) {
		lyxerr << "LyXComm: Sending '" << msg << "'\n";
	}
	
	if (pipename.empty()) return;
	
	if (!ready) {
		lyxerr << "LyXComm: Pipes are closed. Could not send "
		       << msg << '\n';
	} else if (write(outfd, msg.c_str(), msg.length()) < 0) {
		lyxerr << "LyXComm: Error sending message: " << msg << '\n';
		lyxerr << strerror(errno) << '\n';
		lyxerr << "LyXComm: Resetting connection\n";
		closeConnection();
		openConnection();
	}
#ifdef __EMX__
	APIRET rc;
	int errnum;
	rc = DosResetBuffer(outfd);	// To avoid synchronization problems.
	if (rc != NO_ERROR) {
		errnum = TranslateOS2Error(rc);
		lyxerr << "LyXComm: Message could not be flushed: "
		       << msg << '\n';
		lyxerr << strerror(errnum) << '\n';
	}
#endif
}


// LyXServer class

LyXServer::~LyXServer()
{
	// say goodbye to clients so they stop sending messages
	if (clients > 0) pipes.send("LYXSRV:*:bye");
}


/* ---F+------------------------------------------------------------------ *\
   Function  : ServerCallback
    Called by : LyXComm
    Purpose   : handle data gotten from communication
\* ---F------------------------------------------------------------------- */

void LyXServer::callback(LyXServer * serv, string const & msg)
{
	lyxerr << "LyXServer: Received: '" << msg << "'\n";
	
	char const *p = msg.c_str();
	
	// --- parse the string --------------------------------------------
	//
	//  Format: LYXCMD:<client>:<func>:<argstring>\n
	//
	bool server_only = false;
	while(*p) {
		// --- 1. check 'header' ---
		if (compare(p, "LYXSRV:", 7)==0) {
			server_only = true; 
		} else if(0 != compare(p, "LYXCMD:", 7)) {
			lyxerr << "LyXServer: Unknown request\n";
			return;
		}
		p += 7;
		
		// --- 2. for the moment ignore the client name ---
		string client;
		while(*p && *p != ':')
			client += char(*p++);
		if(*p == ':') ++p;
		if(!*p) return;
		
		// --- 3. get function name ---
		string cmd;
		while(*p && *p != ':')
			cmd += char(*p++);
		
		// --- 4. parse the argument ---
		string arg;
		if(!server_only && *p == ':' && *(++p)) {
			while(*p && *p != '\n')
				arg += char(*p++);
			if(*p) ++p;
		}
		
		lyxerr[Debug::LYXSERVER] << "LyXServer: Client: '" << client
					 << "' Command: '" << cmd
					 << "' Argument: '" << arg 
					 << "'\n";
		
		// --- lookup and exec the command ------------------
 
		if (server_only) {
			string buf;
			// return the greeting to inform the client that 
			// we are listening.
			if (cmd == "hello") {
				// One more client
				serv->clients++;
				buf = "LYXSRV:" + client + ":hello";
				lyxerr << "LyXServer: Greeting " 
				       << client << '\n';
				serv->pipes.send(buf);
			} else if (cmd == "bye") {
				// If clients==0 maybe we should reset the pipes
				// to prevent fake callbacks
				serv->clients--;
				lyxerr << "LyXServer: Client " << client
				       << " said goodbye\n";
			} else {
				lyxerr <<"LyXServer: Undefined server command "
				       << cmd << ".\n";
			}
			return;
		}
		
		if (!cmd.empty()) {
			// which lyxfunc should we let it connect to?
			// The correct solution would be to have a
			// specialized (non-gui) BufferView. But how do
			// we do it now? Probably we should just let it
			// connect to the lyxfunc in the single LyXView we
			// support currently. (Lgb)
			
			int action = lyxaction.LookupFunc(cmd.c_str());
			//int action = -1;
			string rval, buf;
			
			if (action>=0) {
				rval = serv->func.Dispatch(action, arg.c_str());
			} else {
				rval = "Unknown command";
			}
			
			if(!rval.empty()) {
				if (action<0 || serv->func.errorStat())
					buf = "ERROR:";
				else
					buf = "INFO:";
				buf += string(client) + ":" + cmd + ":" + rval + "\n";
				serv->pipes.send(buf);
				// !!! we don't do any error checking -
				//  if the client won't listen, the
				//  message is lost and others too
				//  maybe; so the client should empty
				//  the outpipe before issuing a request.
			}
			// not found
		}
	}  /* while *p */
}


/* ---F+------------------------------------------------------------------ *\
   Function  : LyxNotifyClient
   Called by : WorkAreaKeyPress
   Purpose   : send a notify messge to a client
   Parameters: s - string to send
   Returns   : nothing
   \* ---F------------------------------------------------------------------- */

void LyXServer::notifyClient(string const & s)
{
	string buf = string("NOTIFY:") + s + "\n";
	pipes.send(buf);
}
