#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "lyxvc.h"
#include "debug.h"
#include "lyx_gui_misc.h"
#include "bufferlist.h"
#include "syscall.h"
#include "path.h"
#include "filetools.h"
#include "FileInfo.h"
#include "gettext.h"
#include "LyXView.h"
#include "lyxfunc.h"
#include "libgui.h"
#include "lstrings.h"
#include "lyxlex.h"

extern BufferList bufferlist;

LyXVC::LyXVC()
{
	backend = UNKNOWN_VCS;
}


LyXVC::~LyXVC()
{
}


bool LyXVC::file_found_hook(string const & fn)
{
	string tmp(fn);
	FileInfo f;
	// Check if *,v exists.
	tmp += ",v";
	lyxerr[Debug::LYXVC] << "Checking if file is under vc: "
			     << tmp << '\n';
	if (f.newFile(tmp).readable()) {
		lyxerr[Debug::LYXVC] << "Yes it is under vc.\n";
		master = tmp;
		backend = RCS_VCS;
		scanMaster();
		return true;
	} else {
		// Check if RCS/*,v exists.
		tmp = AddName(AddPath(OnlyPath(fn), "RCS"), fn);
		tmp += ",v";
		lyxerr[Debug::LYXVC] << "Checking if file is under vc: "
				     << tmp << '\n';
		if (f.newFile(tmp).readable()) {
			lyxerr[Debug::LYXVC] << "Yes it is under vc.\n";
			master = tmp;
			backend = RCS_VCS;
			scanMaster();
			return true;
		}
	}
	// If either one, return true

	// file is not under any VCS.
	return false;
}

///
void LyXVC::setBuffer(Buffer *buf)
{
	owner_ = buf;
}

bool LyXVC::file_not_found_hook(string const &)
{
	// file is not under any VCS.
	return false;
}


void LyXVC::scanMaster()
{
	lyxerr[Debug::LYXVC] << "LyXVC: This file is a VC file.\n";

	LyXLex lex(0, 0);
	lex.setFile(master);

	string token;
	bool read_enough = false;
	while (lex.IsOK() && !read_enough) {
		lex.next();
		token = lex.GetString();
		
		lyxerr[Debug::LYXVC] <<"LyXVC::scanMaster: current lex text: `"
				     << token << "'\n";

		if (token.empty())
			continue;
		else if (token == "head") {
			// get version here
			lex.next();
			string tmv = strip(lex.GetString(), ';');
			version = tmv;
		} else if (contains(token, "access")
			   || contains(token, "symbols")
			   || contains(token, "strict")) {
			// nothing
		} else if (contains(token, "locks")) {
			// get locker here
			if (contains(token, ";")) {
				locker = "Unlocked";
				vcstat = UNLOCKED;
				continue;
			}
			string tmpt, s1, s2;
			do {
				lex.next();
				tmpt = lex.GetString();
				s1 = strip(tmpt, ';');
				// tmp is now in the format <user>:<version>
				s1 = split(s1, s2, ':');
				// s2 is user, and s1 is version
				if (s1 == version) {
					locker = s2;
					vcstat = LOCKED;
					break;
				}
			} while (!contains(tmpt, ";"));
			
		} else if (token == "comment") {
			// we don't need to read any further than this.
			read_enough = true;
		} else {
			// unexpected
			lyxerr[Debug::LYXVC] 
				<< "LyXVC::scanMaster(): unexpected token\n";
		}
	}
}


//
// I will probably add some backend_xxxx functions later to perform the
// version control system specific commands. Something like:
// void backend_revert(<params>) {
//        if (backend == "RCS") {
//        } else if (backend == "CVS") {
//        } else if (backend == "SCCS") {
//        }
//
// But for 0.12 we will only support RCS.
//

void LyXVC::registrer()
{
	lyxerr[Debug::LYXVC] << "LyXVC: registrer\n";
	string tmp = askForText(_("LyX VC: Initial description"));
	if (tmp.empty()) tmp = "(no initial description)";
	string cmd = "ci -q -u -i -t-\"";
	cmd += tmp;
	cmd += "\" ";
	cmd += OnlyFilename(owner_->getFileName());
	doVCCommand(cmd);
	owner_->getUser()->getOwner()->getLyXFunc().Dispatch("buffer-reload");
}


void LyXVC::checkIn()
{
	lyxerr[Debug::LYXVC] << "LyXVC: checkIn\n";
	owner_->getUser()->getOwner()->getLyXFunc().Dispatch(LFUN_MENUWRITE);
	string tmp = askForText(_("LyX VC: Log Message"));
	if (tmp.empty()) tmp = "(no log msg)";
	doVCCommand("ci -q -u -m\"" + tmp + "\" "
		    + OnlyFilename(owner_->getFileName()));
	owner_->getUser()->getOwner()->getLyXFunc().Dispatch("buffer-reload");
}


void LyXVC::checkOut()
{
	lyxerr[Debug::LYXVC] << "LyXVC: checkOut\n";
	doVCCommand("co -q -l "
		    + OnlyFilename(owner_->getFileName()));
	owner_->getUser()->getOwner()->getLyXFunc().Dispatch("buffer-reload");
}


void LyXVC::revert()
{
	lyxerr[Debug::LYXVC] << "LyXVC: revert\n";
	// Here we should check if the buffer is dirty. And if it is
	// we should warn the user that reverting will discard all
	// changes made since the last check in.
	if (AskQuestion("LyX VC:",
			_("When reverting you will lose all changes made"
			" to the document since the last check-in."),
			_("Do you still want to do it?"))) {
		
		doVCCommand("co -f -u" + getVersion() + ' '
			    + OnlyFilename(owner_->getFileName()));
		owner_->getUser()->getOwner()->
			getLyXFunc().Dispatch("buffer-reload");
	}
}


void LyXVC::undoLast()
{
	lyxerr[Debug::LYXVC] << "LyXVC: undoLast\n";
	doVCCommand("rcs -o" + getVersion() + ' '
		    + OnlyFilename(owner_->getFileName()));
}


void LyXVC::toggleReadOnly()
{
	switch (vcstat) {
	case UNLOCKED:
		checkOut();
		break;
	case LOCKED:
		checkIn();
		break;
	}
}


bool LyXVC::inUse()
{
	if (!master.empty())
		return true;
	return false;
}


string const LyXVC::getVersion() const
{
	return version;
}


string const LyXVC::getLocker() const
{
	return locker;
}


static void unlinkwrapper(string fil)
{
	unlink(fil.c_str());
}


void LyXVC::showLog()
{
	string tmpf = tmpnam(0);
	doVCCommand("rlog "
		    + OnlyFilename(owner_->getFileName())
		    + " > " + tmpf);
	FileBrowser *tmp = showFileBrowser(tmpf, "LyXVC Log");
	connect_to_function(tmp->Closing, unlinkwrapper, tmpf);
}


int LyXVC::doVCCommand(string const & cmd)
{
	lyxerr[Debug::LYXVC] << "doVCCommand: " << cmd << '\n';
	Path p(owner_->getFileDir());
        Systemcalls one;
	int ret = one.Startscript(Systemcalls::System, cmd);
	return ret;
}
