/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *          Copyright (C) 1995-1998 The LyX Team.
 *
 * ======================================================*/

#include <config.h>

// 	$Id: lyx_main.C,v 1.11 1998/10/22 16:56:29 larsbj Exp $	
#ifdef __GNUG__
#pragma implementation "lyx_main.h"
#pragma implementation "path.h"
#endif

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: lyx_main.C,v 1.11 1998/10/22 16:56:29 larsbj Exp $";
#endif /* lint */

#include <stdlib.h>
#include <signal.h>

#include "version.h"
#include "lyx_main.h"
#include "lyx_gui.h"
#include "lyxrc.h"
#include "path.h"
#include "bufferlist.h"
#include "debug.h"
#include "FileInfo.h"
#include "lastfiles.h"
#include "intl.h"
#include "lyxserver.h"
#include "textclasslist.h"

string system_lyxdir;
string build_lyxdir;
string system_tempdir;
string user_lyxdir;	// Default $HOME/.lyx

// Should this be kept global? Asger says Yes.
DebugStream lyxerr;

LastFiles *lastfiles;
LyXRC *lyxrc;

// This is the global bufferlist object
BufferList bufferlist;

LyXServer *lyxserver = 0;
// this should be static, but I need it in buffer.C
bool finished = false;	// flag, that we are quitting the program

// convenient to have it here.
kb_keymap *toplevel_keymap;

LyX::LyX(int *argc, char *argv[])
{
	// Prevent crash with --help
	lyxGUI = 0;
	lastfiles = 0;

	// Here we need to parse the command line. At least
	// we need to parse for "-dbg" and "-help"

	bool gui = easyParse(argc, argv);
	
	// Global bindings (this must be done as early as possible.) (Lgb)
	toplevel_keymap = new kb_keymap;

	lyxerr[Debug::INFO] << "Initializing lyxrc\n";
	lyxrc = new LyXRC;

	// Make the GUI object, and let it take care of the
	// command line arguments that concerns it.
	lyxerr[Debug::INFO] << "Initializing LyXGUI...\n";
	lyxGUI = new LyXGUI(this, argc, argv, gui);
	lyxerr[Debug::INFO] << "Initializing LyXGUI...done\n";

	// Initialization of LyX (reads lyxrc and more)
	lyxerr[Debug::INFO] << "Initializing LyX::init...\n";
	init(argc, argv);
	lyxerr[Debug::INFO] << "Initializing LyX::init...done\n";

	lyxGUI->init();

	// Load the files specified in the command line.
	// Now the GUI and LyX have taken care of their arguments, so
	// the only thing left on the command line should be
	// filenames.
	if ((*argc)==2) 
		lyxerr[Debug::INFO] << "Opening document...\n";
	else if ((*argc)>2)
		lyxerr[Debug::INFO] << "Opening documents...\n";

	Buffer *last_loaded = 0;

	for (int argi = (*argc) - 1; argi >= 1; --argi) {
		Buffer * loadb = bufferlist.loadLyXFile(argv[argi]);
		if (loadb != 0) {
			last_loaded = loadb;
		}
	}

	if (first_start) {
		string splash = i18nLibFileSearch("examples", "splash.lyx");
		lyxerr[Debug::INFO] << "Opening splash document "
				    << splash << "...\n";
		Buffer * loadb = bufferlist.loadLyXFile(splash);
		if (loadb != 0) {
			last_loaded = loadb;
		}
	}

	if (last_loaded != 0) {
		lyxerr[Debug::INFO] << "Yes we loaded some files.\n";
		lyxGUI->regBuf(last_loaded);
	}
	
	// Let the ball begin...
	lyxGUI->runTime();
}


// A destructor is always necessary  (asierra-970604)
LyX::~LyX()
{
	if (lastfiles)
		delete lastfiles;

	if (lyxGUI)
		delete lyxGUI;
}


void error_handler(int err_sig);

void LyX::init(int */*argc*/, char **argv)
{
	// Install the signal handlers
	struct sigaction act;
	act.sa_handler = error_handler;
	//act.sa_mask = 0;
	act.sa_flags = SA_RESETHAND;
	sigaction(SIGHUP, &act, 0);
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGSEGV, &act, 0);
	sigaction(SIGINT, &act, 0);
	sigaction(SIGTERM, &act, 0);

	//
	// Determine path of binary
	//

	string binpath = subst(argv[0], '\\', '/');
	string binname = OnlyFilename(argv[0]);
	// Sorry for system specific code. (SMiyata)
	if (suffixIs(binname, ".exe")) binname.erase(binname.length() - 5, string::npos);
		//binname = binname.substr(0,binname.length()-5);
	
	binpath = ExpandPath(binpath); // This expands ./ and ~/
	
	if (!AbsolutePath(binpath)) {
		string binsearchpath = GetEnvPath("PATH");
		binsearchpath += ";."; // This will make "src/lyx" work always :-)
		binpath = FileOpenSearch(binsearchpath, argv[0]);
	}
	
	binpath = MakeAbsPath(OnlyPath(binpath));

	if (binpath.empty()) {
		lyxerr << _("Warning: could not determine path of binary.")
		       << '\n';
		lyxerr << _("If you have problems, try starting LyX with an absolute path.") << '\n';
	}
	lyxerr[Debug::INIT] << "Path of binary: " << binpath << '\n';

	//
	// Determine system directory.
	//

	// Directories are searched in this order:
	// 1) -sysdir command line parameter
	// 2) LYX_DIR_DEV environment variable
	// 3) Maybe <path of binary>/TOP_SRCDIR/lib
	// 4) <path of binary>/../share/<name of binary>/
	// 5) hardcoded lyx_dir
	// The directory is checked for the presence of the file
	// "chkconfig.ltx", and if that is present, the directory
	// is accepted as the system directory.
	// If no chkconfig.ltx file is found, a warning is given,
	// and the hardcoded lyx_dir is used.

	// If we had a command line switch, system_lyxdir is already set
	string searchpath = MakeAbsPath(system_lyxdir) + ';';

	// LYX_DIR_DEV environment variable
	string lyxdir = GetEnvPath("LYX_DIR_DEV");
	
	if (!lyxdir.empty()) {
		lyxerr[Debug::INIT] << "LYX_DIR_DEV: " << lyxdir << '\n';
		searchpath += lyxdir + ';';
	}

	// <path of binary>/TOP_SRCDIR/lib
	build_lyxdir = MakeAbsPath("../lib", binpath);
	if (!FileSearch(build_lyxdir, "lyxrc.defaults").empty()) {
		string top_srcdir=TOP_SRCDIR;
		// Horrible hack... Since this source file is not in
		// the same directory than the LyX binary, we have to
		// strip the initial "../" if we want to find the LyX
		// support files.
		if (prefixIs(top_srcdir, "../"))
			top_srcdir.erase(0, 3);
			//top_srcdir =
			//	top_srcdir.substr(3, top_srcdir.length() - 1);
		searchpath += MakeAbsPath(AddPath(top_srcdir, "lib"),
					  binpath) + ';';
		lyxerr[Debug::INIT] << "Checking whether LyX is run in "
			"place... yes\n";
	} else {
		lyxerr[Debug::INIT] << "Checking whether LyX is run in place... no\n";
		build_lyxdir.erase();
	}
	  
	// Path of binary/../share/name of binary/
	searchpath += NormalizePath(binpath + "../share/" + 
		      OnlyFilename(binname)) + ';';

	// Hardcoded dir
	searchpath += LYX_DIR;

	// If debugging, show complete search path
	lyxerr[Debug::INIT] << "System directory search path: "
			    << searchpath << '\n';

	string const filename = "chkconfig.ltx";
	string temp = FileOpenSearch(searchpath, filename, string());
	system_lyxdir = OnlyPath(temp);

	// Reduce "path/../path" stuff out of system directory
	system_lyxdir = NormalizePath(system_lyxdir);

	bool path_shown = false;

	// Warn if environment variable is set, but unusable
	if (!lyxdir.empty()) {
		if (system_lyxdir != NormalizePath(lyxdir)) {
			lyxerr << _("LYX_DIR_DEV environment variable no good.") << '\n';
			lyxerr << _("System directory set to: ") 
			       << system_lyxdir << '\n';
			path_shown = true;
		}
	}

	// Warn the user if we couldn't find "chkconfig.ltx"
	if (system_lyxdir.empty()) {
		lyxerr << _("LyX Warning! Couldn't determine system directory.") << '\n';
		lyxerr << _("Try the '-sysdir' command line parameter or") << '\n';
		lyxerr << _("set the environment variable LYX_DIR_DEV to the "
			    "LyX system directory") << '\n';
		lyxerr << _("containing the file `chkconfig.ltx'.") << '\n';
		if (!path_shown)
			lyxerr << _("Using built-in default ") 
			       << LYX_DIR 
			       << _(" but expect problems.") << '\n';
		else
			lyxerr << _("Expect problems.") << '\n';
		system_lyxdir = LYX_DIR;
		path_shown = true;
	}

	// Report the system directory if debugging is on
	if (!path_shown)
		lyxerr[Debug::INIT] << "System directory: '" 
				    << system_lyxdir << "'\n";

	//
	// Determine user lyx-dir
	//
	
	user_lyxdir = AddPath(GetEnvPath("HOME"), string(".")+LYX_NAME);
	lyxerr[Debug::INIT] << "User LyX directory: '" 
			    << user_lyxdir << "'\n";

	// Check that user LyX directory is ok.
	queryUserLyXDir();

	//
	// Load the layouts first
	//

	lyxerr[Debug::INIT] << "Reading textclasses...\n";
	textclasslist.readClasses();

	//
	// Shine up lyxrc defaults
	//

	// Default template path: system_dir/templates
	if (lyxrc->template_path.empty()){
		lyxrc->template_path = AddPath(system_lyxdir, "templates");
	}
   
	// Default lastfiles file: $HOME/.lyx/lastfiles
	if (lyxrc->lastfiles.empty()){
		lyxrc->lastfiles = AddName(user_lyxdir, "lastfiles");
	}

	// Calculate screen dpi as average of x-DPI and y-DPI:
	Screen * scr=(DefaultScreenOfDisplay(fl_get_display()));
	lyxrc->dpi = ((HeightOfScreen(scr)* 25.4 / HeightMMOfScreen(scr)) +
		      (WidthOfScreen(scr)* 25.4 / WidthMMOfScreen(scr))) / 2;
	lyxerr[Debug::INFO] << "DPI setting detected to be "
			    << lyxrc->dpi+0.5 << '\n';

	//
	// Read configuration files
	//

	ReadRcFile("lyxrc.defaults");
	ReadRcFile("lyxrc");

	if (lyxerr.debugging(Debug::LYXRC)) {
		lyxrc->Print();
	}

	// Create temp directory	
	system_tempdir = CreateLyXTmpDir(lyxrc->tempdir_path);
	lyxerr[Debug::INIT] << "LyX tmp dir: `" 
			    << system_tempdir << "'\n";

	// load the lastfiles mini-database
	lyxerr[Debug::INIT] << "Reading lastfiles `"
			    << lyxrc->lastfiles << "'...\n";
	lastfiles = new LastFiles(lyxrc->lastfiles, 
				  lyxrc->check_lastfiles,
				  lyxrc->num_lastfiles);

	// start up the lyxserver. (is this a bit early?) (Lgb)
	// 0.12 this will be way to early, we need the GUI to be initialized
	// first, so move it for now.
	// lyxserver = new LyXServer;
}


// This one is not allowed to use anything on the main form, since that
// one does not exist yet. (Asger)
void LyX::queryUserLyXDir()
{
	// Does user directory exist?
	FileInfo fileInfo(user_lyxdir);
	if (fileInfo.isOK() && fileInfo.isDir()) {
		first_start = false;
		return;
	} else {
		first_start = true;
 	}

	// Nope
	if (!AskQuestion(_("You don't have a personal LyX directory."),
			 _("It is needed to keep your own configuration."),
			 _("Should I try to set it up for you (recommended)?"))) {
		lyxerr << _("Running without personal LyX directory.") << '\n';
		// No, let's use $HOME instead.
		user_lyxdir = GetEnvPath("HOME");
		return;
	}

	// Tell the user what is going on
	lyxerr << _("LyX: Creating directory ")
	       << user_lyxdir 
	       << _(" and running configure...") << '\n';

	// Create directory structure
	if (!createDirectory(user_lyxdir, 0755)) {
		// Failed, let's use $HOME instead.
		user_lyxdir = GetEnvPath("HOME");
		lyxerr << _("Failed. Will use ")
		       << user_lyxdir
		       << _(" instead.") << '\n';
		return;
	}

	// Run configure in user lyx directory
	Path p(user_lyxdir);
	system(AddName(system_lyxdir,"configure").c_str());
	lyxerr << "LyX: " << _("Done!") << '\n';
}


// Read the rc file `name'
void LyX::ReadRcFile(string const & name)
{
	lyxerr[Debug::INIT] << "About to read " << name << "...\n";
	
	string lyxrc_path = LibFileSearch(string(), name);
	if (!lyxrc_path.empty()){
	        lyxerr[Debug::INIT] << "Found " << name 
				    << " in " << lyxrc_path<< '\n';
		if (lyxrc->Read(lyxrc_path) < 0) { 
		        WriteAlert(_("LyX Warning!"), 
				   _("Error while reading ")+lyxrc_path+".",
				   _("Using built-in defaults."));
		}
	} else
	  	lyxerr[Debug::INIT] << "Could not find " << name << '\n';
}


// Give command line help
void commandLineHelp()
{
	cout << _("LyX " LYX_VERSION " of " LYX_RELEASE ".\n"
		  "Usage: lyx [ command line switches ] [ name.lyx ... ]\n"
		  "Command line switches (case sensitive):\n"
		  "   -help           summarize LyX usage\n"
		  "   -sysdir x       try to set system directory to x\n"
		  "   -width x        set the width of the main window\n"
		  "   -height y       set the height of the main window\n"
		  "   -xpos x         set the x position of the main window\n"
		  "   -ypos y         set the y position of the main window\n"
		  "   -dbg feature[,feature]...\n"
		  "                   select the features to debug.\n"
		  "                   Type `lyx -dbg' to see the list of features\n"
		  "   -Reverse        swaps foreground & background colors\n"
		  "   -Mono           runs LyX in black and white mode\n"
		  "   -FastSelection  use a fast routine for drawing selections\n"
		  "Check the LyX man page for more options.\n");
}


bool LyX::easyParse(int *argc, char *argv[])
{
	bool gui = true;
#if 0
	struct option long_options[] =
 	{	
		{ "help", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'V' },
		{ 0, 0, 0, 0}
	};
	int opt;
	bool do_help = false;
        bool do_version = false;
	while ((opt = getopt_long (*argc, argv, "hv", long_options, 0))
!= EOF)
	switch (opt) {
	case '\0': // Long option
		break;
	case 'h':
		cerr << "Help (h)\n";
		do_help = true;
		break;
	case 'V':
		cerr << "Version (V)\n";
		do_version = true;
		break;
	default:
		//usage(EXIT_FAILURE);
		break;
	}

	if (do_help) {
		//usage(EXIT_SUCCESS);
	}
		
	if (do_version) {
		cout << "LyX Version something...\n";
		exit(EXIT_SUCCESS);
	}
#else

	for(int i=1; i < *argc; ++i) {
		string arg = argv[i];
		// Check for -dbg int
		if (arg == "-dbg") {
			if (i+1 < *argc) {
				lyxerr.level(Debug::value(argv[i+1]));
				// Now, remove these two arguments by shifting
				// the following two places down.
				(*argc) -= 2;
				for (int j=i; j < (*argc); ++j)
					argv[j] = argv[j+2];
				--i; // After shift, check this number again.
			} else {
				lyxerr << _("List of supported debug flags:\n");
#warning this is not supported yet
				//lyxerr.showTags();
				exit(0);
			}
		} 
		// Check for "-sysdir"
		else if (arg == "-sysdir") {
			if (i+1 < *argc) {
				system_lyxdir = argv[i+1];

				// Now, remove these two arguments by shifting
				// the following two places down.
				(*argc) -= 2;
				for (int j=i; j < (*argc); ++j)
					argv[j] = argv[j+2];
				--i; // After shift, check this number again.
			} else
				lyxerr << _("Missing directory for -sysdir switch!") << '\n';
		// Check for --help or -help
		} else if (arg == "--help" || arg == "-help") {
			commandLineHelp();
			exit(0);
		} 
		// Check for "-nw": No window
		else if (arg == "-nw") {
			gui = false;
		}
	}
#endif 
	return gui;
}


void error_handler(int err_sig)
{
	switch (err_sig) {
	case SIGHUP:
		cerr << "\nlyx: SIGHUP signal caught\n";
		break;
	case SIGINT:
		// no comments
		break;
	case SIGFPE:
		cerr << "\nlyx: SIGFPE signal caught\n";
		break;
	case SIGSEGV:
		cerr << "\nlyx: SIGSEGV signal caught\n";
		cerr << "Sorry, you have found a bug in LyX."
			" If possible, please read 'Known bugs'\n"
			"under the Help menu and then send us "
			"a full bug report. Thanks!\n";
		break;
	case SIGTERM:
		// no comments
		break;
	}
	
	// Deinstall the signal handlers
	struct sigaction act;
	act.sa_handler = SIG_DFL;
	//act.sa_mask = 0;
	act.sa_flags = 0;
	sigaction(SIGHUP, &act, 0);
	sigaction(SIGINT, &act, 0);
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGSEGV, &act, 0);
	sigaction(SIGTERM, &act, 0);

	bufferlist.emergencyWriteAll();

	lyxerr << "Bye.\n";
	if(err_sig!=SIGHUP && (!GetEnv("LYXDEBUG").empty() || err_sig == SIGSEGV))
		abort();
	exit(0);
}
