/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *          Copyright (C) 1995-1998 The LyX Team.
 *
 *======================================================*/

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "paraparams.h"
#include "layout.h"
#include "tex-strings.h"
#include "lyxlex.h"


LyXParagraphParameters::LyXParagraphParameters()
{
	init();
}


LyXParagraphParameters::LyXParagraphParameters(const LyXParagraphParameters & lpp)
{
	operator=(lpp);
}


LyXParagraphParameters &
LyXParagraphParameters::operator=(const LyXParagraphParameters & lpp)
{
	line_top = lpp.line_top;
	line_bottom = lpp.line_bottom;
	
	pagebreak_top = lpp.pagebreak_top;
	pagebreak_bottom = lpp.pagebreak_bottom;
	
	added_space_top = lpp.added_space_top;
	added_space_bottom = lpp.added_space_bottom;
	
	align = lpp.align;
	noindent = lpp.noindent;
	
	return *this;
}


bool LyXParagraphParameters::operator==(const LyXParagraphParameters &lpp) const
{
	return (line_top == lpp.line_top &&
		line_bottom == lpp.line_bottom &&
		pagebreak_top == lpp.pagebreak_top &&
		pagebreak_bottom == lpp.pagebreak_bottom &&
		added_space_top == lpp.added_space_top &&
		added_space_bottom == lpp.added_space_bottom &&
		align == lpp.align &&
		noindent == lpp.noindent);
}


void LyXParagraphParameters::init()
{
	line_top = false;
	line_bottom = false;
	
	pagebreak_top = false;
	pagebreak_bottom =false;
	
	added_space_top = VSpace::NONE;
	added_space_bottom = VSpace::NONE;
	
	align = LYX_ALIGN_LAYOUT;
	noindent = false;
}


void LyXParagraphParameters::read(LyXLex &lex)
{
	string token;
	int tmpret, tmpret2;
	
	while (lex.IsOK()) {
		lex.nextToken();
		token = lex.GetString();

		if (token.empty())
			continue;
		else if (token == "\\noindent") {
			noindent = true;
		} else if (token == "\\fill_top") {
			added_space_top = VSpace::VFILL;
		} else if (token == "\\fill_bottom") {
			added_space_bottom = VSpace::VFILL;
		} else if (token == "\\line_top") {
			line_top = true;
		} else if (token == "\\line_bottom") {
			line_bottom = true;
		} else if (token == "\\pagebreak_top") {
			pagebreak_top = true;
		} else if (token == "\\pagebreak_bottom") {
			pagebreak_bottom = true;
		} else if (token == "\\align") {
			tmpret = lex.FindToken(string_align);
			if (tmpret == -1) ++tmpret;
			if (tmpret != LYX_LAYOUT_DEFAULT) {
				tmpret2 = 1;
				for (; tmpret>0; --tmpret)
					tmpret2 = tmpret2 * 2;
				align = tmpret2;
			}
		} else if (token == "\\added_space_top"){
			lex.nextToken();
			added_space_top = lex.GetString();
		} else if (token == "\\added_space_bottom") {
			lex.nextToken();
			added_space_bottom = lex.GetString();
                } else if (token == "\\end_paragraph_parameters") {
			break;
		} else {
			lex.printError("Unknown paragraphparameter token: "
				       "`$$Token'");
		}
	}
	if (token != "\\end_paragraph_parameters") {
		lex.printError("Missing \\end_paragraph_parameters");
	}
}


void LyXParagraphParameters::writeFile(FILE *file, BufferParams &,
				       char footflag, char dth) const
{
	string tofile;
	// vertical space above or below the paragraph.
	if (added_space_top.kind() != VSpace::NONE) {
		tofile += "\\added_space_top ";
		tofile += added_space_top.asLyXCommand() + '\n';
	}
	if (added_space_bottom.kind() != VSpace::NONE) {
		tofile += "\\added_space_bottom ";
		tofile += added_space_bottom.asLyXCommand() + '\n';
	}
	
	// lines above or below
	if (line_top)
		tofile += "\\line_top\n";
	if (line_bottom)
		tofile += "\\line_bottom\n";
	
	// Pagebreaks above or below the paragraph
	if (pagebreak_top)
		tofile += "\\pagebreak_top\n";
	if (pagebreak_bottom)
		tofile += "\\pagebreak_bottom\n";
	
	// No indentation of first line of the paragraph
	if (noindent)
		tofile += "\\noindent\n";
	
	// Alignment?
	if (align != LYX_ALIGN_LAYOUT) {
		int h = 0;
		switch (align) {
		case LYX_ALIGN_LEFT: h = 1; break;
		case LYX_ALIGN_RIGHT: h = 2; break;
		case LYX_ALIGN_CENTER: h = 3; break;
		default: h = 0; break;
		}
		tofile += "\\align ";
		tofile += string(string_align[h]) + '\n';
	}
	
	// Write parameters to file.
	if (!tofile.empty()) {
		fprintf(file, "\\begin_paragraph_parameters\n");
		fprintf(file, "%s", tofile.c_str());
		fprintf(file, "\\end_paragraph_parameters\n");
	}
}
