/*
 *  File:        formula.h
 *  Purpose:     Implementation of formula inset
 *  Author:      Alejandro Aguilar Sierra <asierra@servidor.unam.mx> 
 *  Created:     January 1996
 *  Description: Allows the edition of math paragraphs inside Lyx. 
 *
 *  Copyright: (c) 1996, 1997 Alejandro Aguilar Sierra
 *
 *  Version: 0.4, Lyx project.
 *
 *   You are free to use and modify this code under the terms of
 *   the GNU General Public Licence version 2 or later.
 */

#include <config.h>
#include <stdlib.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "formulamacro.h"
#include "commandtags.h"
#include "math_cursor.h"
#include "math_parser.h"
#include "math_macro.h"
#include "lyx_main.h"
#include "bufferlist.h"
#include "lyx_cb.h"
#include "BufferView.h"
#include "lyxscreen.h"
#include "gettext.h"
#include "lyxlex.h"


InsetFormulaMacro::InsetFormulaMacro():
        InsetFormula(true)
{
    tmacro = 0;
    opened = false;
}


InsetFormulaMacro::InsetFormulaMacro(string nm, int na, bool /*e*/):
        InsetFormula(true), name(nm)
{
    tmacro = MathMacroTable::mathMTable.getTemplate(name.c_str());
    if (!tmacro) {
	tmacro = new MathMacroTemplate(name.c_str(), na);
	MathMacroTable::mathMTable.addTemplate(tmacro);
    }
    opened = false;
}


InsetFormulaMacro::~InsetFormulaMacro()
{
    par = 0;
}


Inset* InsetFormulaMacro::Clone()
{
   InsetFormulaMacro* f = new InsetFormulaMacro(name);
   return (Inset*)f;
}


void InsetFormulaMacro::Write(FILE *file)
{
    fprintf(file, "FormulaMacro ");
    Latex(file, 0);
}


int InsetFormulaMacro::Latex(FILE *file, signed char /*fragile*/)
{
    int ret = 1;
    tmacro->WriteDef(file);
    return ret;
}


int InsetFormulaMacro::Latex(string &file, signed char /*fragile*/)
{
    int ret = 1;
    tmacro->WriteDef(file);
    return ret;
}


void InsetFormulaMacro::Read(LyXLex &lex)
{
    FILE *file = lex.getFile();
    mathed_parser_file(file, lex.GetLineNo());   
    mathed_parse(0, 0, (MathParInset **)&tmacro);
    
    // Update line number
    lex.setLineNo(mathed_parser_lineno());
    
    MathMacroTable::mathMTable.addTemplate(tmacro);
    name = tmacro->GetName();
    par = tmacro;
}


int InsetFormulaMacro::ascent(Painter & pain, LyXFont const &f) const
{
    if (opened) {
	tmacro->update();
	return InsetFormula::ascent(pain, f);
    }
    return f.maxAscent()+3;
}


int InsetFormulaMacro::descent(Painter & pain, LyXFont const &f) const
{
    if (opened) {
	tmacro->update();
	return InsetFormula::descent(pain, f);
    }
    return f.maxDescent()+1;
}


int InsetFormulaMacro::width(Painter & pain, LyXFont const &f) const
{
    if (opened) {
	tmacro->update();
	return InsetFormula::width(pain, f);
    }
    string ilabel(_("Macro: "));
    ilabel += name;
    return 6 + f.stringWidth(ilabel);
}


void InsetFormulaMacro::draw(Painter &pain, LyXFont const & f,
			     int baseline, float &x)
{
	tmacro->update();
	if (opened) {
		tmacro->setEditMode(true);
		InsetFormula::draw(pain, f, baseline, x);
		tmacro->setEditMode(false);	
	} else {
		LyXFont font(LyXFont::ALL_SANE);
		font.setColor(LColor::math);
		int width;
		pain.buttonText(int(x), baseline, _("Macro"), font, 
				true, width);
		x +=  width;
	}
}


void InsetFormulaMacro::Edit(int x, int y, unsigned int button, BufferView *bv)
{
    opened = true;
    par = (MathParInset*)tmacro->Clone();
    InsetFormula::Edit(x, y, button, bv);
}

	       
void InsetFormulaMacro::InsetUnlock()
{
    opened = false;
    LyxArrayBase *tarray = tmacro->GetData();
    MathedIter it(tarray);
    it.Clear();
    tmacro->SetData(par->GetData());
    tmacro->setEditMode(false);
    InsetFormula::InsetUnlock();
}


UpdatableInset::RESULT InsetFormulaMacro::LocalDispatch(int action, 
							char const *arg)
{
    if (action==LFUN_MATH_MACROARG) {
	int i = atoi(arg) - 1;
	if (i>=0 && i<tmacro->getNoArgs()) {
	    mathcursor->Insert(tmacro->getMacroPar(i), LM_TC_INSET);
	    InsetFormula::UpdateLocal();
	}
	
	return DISPATCHED;
    }
    tmacro->setEditMode(true);
    tmacro->Metrics();
    UpdatableInset::RESULT result = InsetFormula::LocalDispatch(action, arg);
    tmacro->setEditMode(false);
    
    return result;
}
