// -*- C++ -*-
/*
 * LyX 3 0.0.3 Grammar for ANTLR parser generator.
 *
 * Created by Alejandro Aguilar Sierra (asierra@servidor.unam.mx)
 * Work continued with Asger Alstrup (alstrup@diku.dk)
 * Copyright 1997, LyX Team.
 *
 *  There are two kind of command arguments:
 *      string - Anything inside [] or {} is emited toghether as string.
 *   paragraph - The argument is parsed. It can contain other tokens and
 *               commands.
 *  
 *  LyX metacommands (ignored by LaTeX) are denoted by "%#{lyx}".
 *
 *  Lex classes:
 *    0.  Start. Ignore spaces.
 *    1.  Paragraph. Inside a paragraph, spaces are word separators
 *        and multiple newlines are paragraph separators.
 *    2.  String. Get everything in a single string without checking syntax.
 *    3.  Verbatim, as a plain text editor. Like String but can contain
 *        several lines, including void lines.
 *
 *  Semantic lex classes (LaTeX modes): 
 *    Math .-  Spaces are ignored, every char is a word. 
 *    Tabular .- Like paragraph but tab tokens are valid.
 *
 *
 */

/*
Asger recommends to use these flags in the makefile:

CFLAGS = -ansi -Wall -Wno-unused -pedantic -I. -I$(ANTLR_H)
AFLAGS = -CC -e3
DFLAGS = -C2 -i -CC
*/

#header <<
#include "AToken.h"
#include "ATokPtr.h"
#include "LString.h"

typedef ANTLRCommonToken ANTLRToken;
>>

<<
#include "PBlackBox.h"
#include "DLGLexer.h"

int main() {

    ParserBlackBox<DLGLexer, LyXParser, ANTLRToken> p(stdin);
    p.parser()->prog();
    return 0;
}
>>


/************************************************************************/
/* Lexical specification                                                */
/************************************************************************/
  


// Groups and arguments
#token CurlyBegin "\{"                          //  {
#token CurlyEnd "\}"                            //  }
#token SharpBegin "\["                          //  [
#token SharpEnd "\]"                            //  ]


//*** Use either word based tokens

// A word us compossed by ASCII characters 
#token Word "[A-Z a-z ]+"

//*** or character based tokens, never both at the same time.

// Normal characters
//#token Character "[A-Z a-z 0-9]"
// To admit these, the doc should use either an encoding package or a special font
//#token Iso8859 "[]+"


#token Number "[0-9]+"

// Non-alphanumeric
#token Punctuation "[,;:.\(\)\?=\-\+']"

// Scaped characters
#token Special "\\[#$%&_\{\}]"

// Accents
#token Accent "\\[\"  ' ^ H \~ \\c \= b . d u v t]"

// Tabulators
#token Tab "&"

#token WordSep "[\ \t]+"
#token WordSep "[\r\n]"                 <<newline();>>

#token ParSep "[\n][\n]"                <<newline();newline();>>

// Comments
#token "\%" << mode(LYXCOMM); skip(); >>

// command
#token Command "\\[A-Z a-z]+"

#token BeginEnv "\\begin\{" <<mode(ENVIRON);>>
#token EndEnv "\\end\{" <<mode(ENVIRON);>>



#lexclass LYXCOMM

#token LyxCommand "\#\{lyx\} ~[\n]+" 
    << if (strncmp("begin.preamble",lextext(),14)==0) mode(IGNORE);
       else mode(START); >>
       
#token " ~[\n]+ "      << skip(); >>
#token "[\n]" << mode(START); skip(); newline();>>



// Things interesting to LaTeX but not to LyX
#lexclass IGNORE

#token " ~[\n]+ "      
  << skip(); if(strncmp("end.preamble",lextext(),12)==0) mode(START); >>
  
#token "[\n]" << mode(START); skip(); newline();>>



#lexclass VERBATIM

#token String "~[\n]+"    
    <<if (strcmp(lextext(),"\\end{verbatim}")==0) {
         mode(ENVIRON);
	 return EndEnv;
      }>>
      
#token String "[\n][\n]"      <<newline();newline();>>
#token "[\n]"      <<skip(); newline();>>


#lexclass ENVIRON

#token Argument "~[\}]+"  <<if (strcmp(lextext(),"verbatim")==0) mode(VERBATIM);>>
#token "[\}]"      <<mode(START); skip();>>


#lexclass STRING

#token Argument "~[\}]+"
#token "[\n]"  << mode(START); >>


/************************************************************************/

class LyXParser {
<<
/* Parser Members */
public:
	void init()
	{
            ANTLRParser::init();
	}
    
protected:
	void emit(const char *t, const char *s) {
		fprintf(stderr, "%s: %s\n", t, s);
	        fflush(stderr);
	}
    
        int format;
	
>>

/************************************************************************/
/* Syntaxtical specification						*/
/************************************************************************/

prog:   << ANTLRTokenPtr opts; >> 
	body
	<<fprintf(stderr, "Done!\n");>>
	;

body:	(atom | command | lyxcommand | argument| environ)*
	;

atom: 	
      	  Character <<emit("word", LT(1)->getText());>>
      	| Accent <<emit("accent", LT(1)->getText());>>
	| Iso8859 <<emit("iso-8859-x", LT(1)->getText());>>
	| Word <<emit("Word", LT(1)->getText());>>
	| WordSep <<emit("Space", "");>>
	| ParSep <<emit("ParSep","");>>
	| Number <<emit("Number", LT(1)->getText());>>
	| Punctuation <<emit("Punctuation", LT(1)->getText());>>
	| Special <<emit("Special", LT(1)->getText());>>
	| String <<emit("String", LT(1)->getText());>>
	| Tab <<emit("Tabulator", LT(1)->getText());>>
	;

command:  
        Command <<emit("Command",  &LT(1)->getText()[1]);>>
	;

argument:
	CurlyBegin
	<<	
		emit("Open argument", "");
	>> 
	
	body 
	
	CurlyEnd
	<<
	        emit("Close argument", "");
	>>     
	|
	
	SharpBegin
	<<	
		emit("Open optional argument", "");
	>> 
	
	body 
	
	SharpEnd
	<<
	        emit("Close optional argument", "");
	>>
	;

string_argument:
	CurlyBegin
	<<	
		emit("Open argument", "");
	>> 
	
	body 
	
	CurlyEnd
	<<
	        emit("Close argument", "");
	>>     
	|
	
	SharpBegin
	<<	
		emit("Open optional argument", "");
	>> 
	
	body 
	
	SharpEnd
	<<
	        emit("Close optional argument", "");
	>>
	;

environ:
	BeginEnv Argument
	<<	
		emit("Begin",LT(1)->getText());
	>> 
	
	body 
	
	EndEnv Argument
	<<
	        emit("End", LT(1)->getText());
	>>     
	;

lyxcommand: 
        cmd:LyxCommand
	<<emit("LyXCommand", &cmd->getText()[6]);>>
	;

}
