/* This file is part of
 * ======================================================
 * 
 *           LyX, The Document Processor
 * 	 
 *	    Copyright (C) 1995 Matthias Ettrich
 *          Copyright (C) 1995-1998 The LyX Team.
 *
 * ======================================================*/

#include <config.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include <algorithm>

#include "textclasslist.h"
#include "filetools.h"
#include "lyx_gui_misc.h"
#include "debug.h"
#include "lyxlex.h"

// 	$Id: textclasslist.C,v 1.7 1998/10/17 14:46:29 larsbj Exp $

#if !defined(lint) && !defined(WITH_WARNINGS)
static char vcid[] = "$Id: textclasslist.C,v 1.7 1998/10/17 14:46:29 larsbj Exp $";
#endif /* lint */

// Global variable: textclass table.
LyXTextClassList textclasslist;


// Reads the style files
void LyXTextClassList::readClasses()
{
	lyxerr[Debug::INFO] << "LyXSetStyle: parsing configuration...\n";
	
	if (!read()) {
		lyxerr << "LyXSetStyle: an error occured during parsing.\n"
		       << "             Exiting.\n";
		// This is a pretty tough thing to do. Perhaps de should
		// have a hardcoded simple textclass that could be used
		//if this happes?
		exit(1);
	}

	lyxerr[Debug::INFO] << "LyXSetStyle: configuration parsed.\n";
}


// Gets textclass number from name
signed char LyXTextClassList::numberOfClass(string const &textclass) const
{
	TextClassList::const_iterator tit = classList.begin();
	for(; tit != classList.end(); ++tit) {
		if (tit->name == textclass) {
			return tit - classList.begin();
		}
	}
	return -1;
}


// Gets layout structure from style number and textclass number
LyXLayout const *LyXTextClassList::style(unsigned char textclass,
				   unsigned char layout) const
{
	classList[textclass].load();

	if (layout < classList[textclass].numLayouts()) {
		return classList[textclass].getLayout(layout);
	} else {
		return classList[textclass].getLayout(0); // default style
	}
}


// Gets layout number from name and textclass number
signed char LyXTextClassList::numberOfLayout(unsigned char textclass,
					     string const &name) const
{
	classList[textclass].load();

	return classList[textclass].getNumberOfLayout(name);
}


// Gets a layout (style) name from layout number and textclass number
string LyXTextClassList::nameOfLayout(unsigned char textclass,
				       unsigned char layout) const
{
	classList[textclass].load();

	return classList[textclass].getNameOfLayout(layout);
}


// Gets a textclass name from number
string LyXTextClassList::nameOfClass(unsigned char number) const
{
	if (classList.size() == 0) {
		if (number == 0) return "dummy";
		else return "@@end@@";
	}

	if (number < classList.size())
		return classList[number].name;
	else
		return "@@end@@";
}


// Gets a textclass latexname from number
string LyXTextClassList::latexNameOfClass(unsigned char number) const
{
	classList[number].load();

	if (classList.size() == 0) {
		if (number == 0) return "dummy";
		else return "@@end@@";
	}

	if (number < classList.size())
		return classList[number].latexname;
	else
		return "@@end@@";
}


// Gets a textclass description from number
string LyXTextClassList::descOfClass(unsigned char number) const 
{
	if (classList.size() == 0) {
		if (number == 0) return "dummy";
		else return "@@end@@";
	}

	if (number < classList.size())
		return classList[number].description;
	else
		return "@@end@@";
}


// Gets a textclass structure from number
LyXTextClass const * LyXTextClassList::textClass(unsigned char textclass) const
{
	classList[textclass].load();

	if (textclass < classList.size())
		return &classList[textclass];
	else
		return &classList[0]; 
}


// Reads LyX textclass definitions according to textclass config file
bool LyXTextClassList::read()
{
	LyXLex lex(0, 0);
	string real_file = LibFileSearch("", "textclass.lst");
	lyxerr[Debug::TCLASS] << "Reading textclasses from "
			      << real_file << '\n';

	if (real_file.empty()) {
		lyxerr << "LyXTextClassList::Read: unable to find "
			"textclass file  `"
		       << MakeDisplayPath(real_file, 1000) 
		       << "'. Exiting.\n";

		WriteAlert(_("LyX wasn't able to find its layout descriptions!"),
			   _("Check that the file \"textclass.lst\""),
			   _("is installed correctly. Sorry, has to exit :-("));
		return false;
		// This causes LyX to end... Not a desirable behaviour. Lgb
		// What do you propose? That the user gets a file dialog
		// and is allowed to hunt for the file? (Asger)
	}

	lex.setFile(real_file);
	
	if (!lex.IsOK()) {
		lyxerr << "LyXTextClassList::Read: unable to open "
			"textclass file  `"
		       << MakeDisplayPath(real_file, 1000) << "'\n";
		lyxerr << "Check your installation. LyX can't continue.\n";
 		return false;
	}
	bool finished = false;
	string fname, clname, desc;

	// Parse config-file
	while (lex.IsOK() && !finished) {
		switch (lex.lex()) {
		case LyXLex::LEX_FEOF:
			finished = true;
			break;
		default:
			fname = lex.GetString();
			lyxerr[Debug::TCLASS] << "Fname: " << fname << '\n';
			if (lex.next()) {
				clname = lex.GetString();
				lyxerr[Debug::TCLASS] << "Clname: "
						      << clname << '\n';
				if (lex.next()) {
					      desc = lex.GetString();
					      lyxerr[Debug::TCLASS] 
						      << "Desc: "
						      << desc << '\n';
					      // This code is run when we have
					      // fname, clname and desc
					      LyXTextClass tmpl(fname,
								clname,
								desc);
					      if (lyxerr.
						  debugging(Debug::TCLASS)) {
					            tmpl.load();
					      }
					      classList.push_back(tmpl);
				}
			}
		}
	}

	// We sort the list on LyXTextClass' description.
	sort(classList.begin(), classList.end(),
	     LyXTextClass::compare_desc());

	return classList.size() != 0;
}


// Load textclass
/* Returns false if this fails */
bool LyXTextClassList::load(unsigned char const number)
{
	bool result = true;

	if (number < classList.size()) {
		classList[number].load();
		if (classList[number].numLayouts() == 0)
			result = false;
	} else
		result = false;

	return result;
}
